/*
    \brief Deklarationen der Funktionen für das Übertragen und das Bearbeiten von Dateien.

    \author Bernd Ottmann
    \date   21.02.2003
*/
#pragma once
#ifndef DATAFOX_FILES_H
#ifndef DOXYGEN_SHOULD_SKIP_THIS
#define DATAFOX_FILES_H
#endif

#include "df_defines.h"

/*!
    \brief Schreiben einer Setupdatei (*.aes; *.ael).

    Die Setupdatei (*.aes; *.ael) wird auf das Gerät übertragen. Danach müssen ggf. die Listen neu übertragen werden. Ebenfalls ggf. die Zutrittslisten oder Timeboylisten.

    \note Es dürfen sich keine abrufbare Datensätze mehr auf dem Gerät befinden. Rufen Sie diese zuvor mit DFCReadRecord() ab oder
    wenn sie nicht mehr benötigt werden, dann können Sie diese mit DFCComClearData() löschen.

    \ingroup files

    \export{_DFCSetupLaden@16, DFCSetupLaden}

    \param[in] connectionId Zu verwendende Verbindungs-Id.
    \param[in] deviceAddress Anzusprechende Geräteadresse.
    \param[in] filePath Dateiname der zu übertragenden Setupdatei mit einer der Endungen *.aes oder *.ael.
    \param[out] errorNumber Fehlernummer bei fehlerhafter Ausführung.

    \retval 0 Fehler bei Ausführung.
    \retval 1 Ausführung war erfolgreich.
*/
DF_API_EXTERN_DECORATE int DF_API_CALL DFCSetupLaden( int connectionId, int deviceAddress, const char *filePath, int *errorNumber );

/*!
    \brief Lesen einer Setupdatei (*.aes; *.ael).

    Die im Gerät enthaltene Setupdatei wird ausgelesen und in der angegebenen Datei gespreichert.

    \ingroup files

    \export{_DFCDownload@16, DFCDownload}

    \warning Eine ggf. vorhandene Datei kann durch die auszulesende überschrieben werden.

    \param[in] connectionId Zu verwendende Verbindungs-Id.
    \param[in] deviceAddress Anzusprechende Geräteadresse.
    \param[in] filePath Dateiname der zu speichernden Setupdatei mit einer der Endungen *.aes oder *.ael.
    \param[out] errorNumber Fehlernummer bei fehlerhafter Ausführung.

    \retval 0 Fehler bei Ausführung.
    \retval 1 Ausführung war erfolgreich.
*/
DF_API_EXTERN_DECORATE int DF_API_CALL DFCDownload( int connectionId,int deviceAddress, const char *filePath, int *errorNumber );

/*!
    \brief Verändern von Setupdateien aufgrund einer Vorlagedatei (*.aes; *.ael).

    Übertragen der notwendigen Gerätedaten die aus Zeichensatz-, Grafik-, Textdaten sowie Firmware bestehen und zusammen
    in einer Datei der Endung *.hex oder in einem Archiv der Endung *.dfz vorliegen.

    \ingroup files

    \export{_DFCModifyStudioFile@20, DFCModifyStudioFile}

    \note Zur Funktion liegt ein eigenes Dokument auf der Produkt-DVD.

    \param[in] sourceFilePath Dateiname der Setupdatei die als Template dient.
    \param[in] destFilePath Dateiname der modifizierten Setupdatei, welche dann in das Gerät übertragen werden soll.
    \param[in] xmlData Datenfragment oder Dateiname einer Datei welche die Informationen zum Modifizieren enthält.
    \param[in] flags Reservierter Parameter. Standardwert 0.
    \param[out] errorNumber Fehlernummer bei fehlerhafter Ausführung.

    \retval 0 Fehler bei Ausführung.
    \retval 1 Ausführung war erfolgreich.

    \par Beispiel für \a xmlData.
    \code{.xml}
    <?xml version="1.0" ?>
    <modify>
        <password>
          <system>
              <value>1234</value>
          </system>
          <user>
              <value>2345</value>
          </user>
          <menu_item>
              <name>MEID_003</name>
              <value>4711</value>
          </menu_item>
      </password>

      <fingerprint>
          <security_level>20</security_level>
          <image_quality>70</image_quality>
          <minutiae_count>20</minutiae_count>
      </fingerprint>

      <replace>
          <name>
              <search>MID_001</search>
              <replacement>PZE Setup</replacement>
          </name>
          <name>
              <search>MEID_001</search>
              <replacement>Kommt:</replacement>
          </name>
          <name>
              <search>MEID_002</search>
              <replacement>Geht:</replacement>
          </name>
          <name>
              <search>MEID_003</search>
              <replacement>Abwesend</replacement>
          </name>
      </replace>
    </modify>
    \endcode
*/
DF_API_EXTERN_DECORATE int DF_API_CALL DFCModifyStudioFile( const char *sourceFilePath, const char *destFilePath, const char *xmlData, int flags, int *errorNumber );

/*!
    \brief Schreiben von Gerätedateien (Firmwareupdates) zum Gerät.

    Übertragen der notwendigen Gerätedaten die aus Zeichensatz-, Grafik-, Textdaten sowie Firmware bestehen und zusammen
    in einer Datei der Endung *.hex oder in einem Archiv der Endung *.dfz vorliegen.

    \ingroup files

    \export{_DFCUpload@16, DFCUpload}

    \note Unter Unix-Systemen ist bei Verwendung eines Archives *.dfz dieses vor der Übertragung zu entpacken und beim
    Aufruf der Funktion das Verzeichnis anzugeben.

    \param[in] connectionId Zu verwendende Verbindungs-Id.
    \param[in] deviceAddress Anzusprechende Geräteadresse.
    \param[in] filePath Dateiname der Gerätedatei (*.dfz; *.hex).
    \param[out] errorNumber Fehlernummer bei fehlerhafter Ausführung.

    \retval 0 Fehler bei Ausführung.
    \retval 1 Ausführung war erfolgreich.
*/
DF_API_EXTERN_DECORATE int DF_API_CALL DFCUpload( int connectionId, int deviceAddress, const char *filePath, int *errorNumber );

#if (DF_API_VERSION >= 0x04020300)

/*!
    \brief Schreiben von Dateidaten zum Gerät.

    Schreibt die Daten der angegebenen Datei unter dem angegebenen Dateinamen in das Gerät. Je nach Dateityp kann diese
    ggf. später mit DFCFileDownload() wieder ausgelesen werden.

    \ingroup files

    \export{_DFCFileUpload@24, DFCFileUpload}

    \warning Diese Funktion ist nicht dafür gedacht Anwenderdateien auf dem Gerät abzulegen. Unbekannte Dateien werden
    bei Bedarf von der Firmware gelöscht. Ggf. kann es durch eine Unsachgemäße Verwendung zu Problemen bei der
    Übernahme von Listendaten u. a. kommen.

    \param[in] connectionId Zu verwendende Verbindungs-Id.
    \param[in] deviceAddress Anzusprechende Geräteadresse.
    \param[in] fileType Typ der zu schreibenden Dateidaten.
    <table class="paramvalues">
        <tr><th>Wert</th><th>Bedeutung</th></tr>
        <tr><td>-1</td><td>Dateityp wird an der Namensendung erkannt.</td></tr>
        <tr><td>0</td><td>Datei mit Textdaten der Firmware. Endung *.dfl</td></tr>
        <tr><td>1</td><td>Datei mit Touchtastendefinitionen des Touch. Endung *.dfk</td></tr>
        <tr><td>2</td><td>Datei mit Farbeinstellungen für TimeboyIV. Endung *.dfc</td></tr>
        <tr><td>3</td><td>Datei mit LAN / WLAN Konfigurationsdaten. Endung *.df0</td></tr>
        <tr><td>4</td><td>Dateiangabe. Endung (*.*)</td></tr>
    </table>
    \param[in] filePath Dateiname.
    \param[in] flags Reservierter Parameter. Standardwert 0.
    \param[out] errorNumber Fehlernummer bei fehlerhafter Ausführung.

    \retval 0 Fehler bei Ausführung.
    \retval 1 Ausführung war erfolgreich.
*/
DF_API_EXTERN_DECORATE int DF_API_CALL DFCFileUpload( int connectionId, int deviceAddress, int fileType, const char *filePath, int flags, int *errorNumber );

/*!
    \brief Lesen von Dateidaten vom Gerät.

    Liest die Daten der angegebenen Datei unter dem angegebenen Dateinamen aus dem Gerät. Schreiben von Dateidaten
    erfolgt mit der Funktion DFCFileUpload().

    \ingroup files

    \export{_DFCFileDownload@24, DFCFileDownload}

    \param[in] connectionId Zu verwendende Verbindungs-Id.
    \param[in] deviceAddress Anzusprechende Geräteadresse.
    \param[in] fileType Typ der zu lesenden Dateidaten.
    <table class="paramvalues">
        <tr><th>Wert</th><th>Bedeutung</th></tr>
        <tr><td>-1</td><td>Dateityp wird an der Namensendung erkannt.</td></tr>
        <tr><td>0</td><td>Nur Schreiben möglich. Datei mit Textdaten der Firmware. Endung *.dfl</td></tr>
        <tr><td>1</td><td>Nur Schreiben möglich. Datei mit Touchtastendefinitionen des Touch. Endung *.dfk</td></tr>
        <tr><td>2</td><td>Nur Schreiben möglich. Datei mit Farbeinstellungen für TimeboyIV. Endung *.dfc</td></tr>
        <tr><td>3</td><td>Datei mit LAN / WLAN Konfigurationsdaten. Endung *.df0</td></tr>
        <tr><td>4</td><td>Dateiangabe. Endung (*.*)</td></tr>
    </table>
    \param[in] filePath Dateiname.
    \param[in] flags Reservierter Parameter. Standardwert 0.
    \param[out] errorNumber Fehlernummer bei fehlerhafter Ausführung.

    \retval 0 Fehler bei Ausführung.
    \retval 1 Ausführung war erfolgreich.
*/
DF_API_EXTERN_DECORATE int DF_API_CALL DFCFileDownload( int connectionId, int deviceAddress, int fileType, const char *filePath, int flags, int *errorNumber );

#endif

#if (DF_API_VERSION >= 0x04010500)

/*!
    \brief Schreiben von Modulupdates zum Gerät.

    Schreibt die Updatedaten zu verbauten Modulen auf das Gerät. Im Gerät verbaute Module haben teilweise eine eigene
    updatefähige Firmware. Durch diese Funktion können entsprechende Updatedateien übertragen werden, die dann durch
    die Gerätefirmware auf die verbauten Module übertragen wird.

    \ingroup files

    \export{_DFCUploadModule@20, DFCUploadModule}

    \note Ein Updatevorgang kann durch Verwendung von DFCGetInfo() überprüft werden.

    \param[in] connectionId Zu verwendende Verbindungs-Id.
    \param[in] deviceAddress Anzusprechende Geräteadresse.
    \param[in] fileType Typ der zu schreibenden Dateidaten.
    <table class="paramvalues">
        <tr><th>Wert</th><th>Bedeutung</th></tr>
        <tr><td>0</td><td>Firmwareupdate für Biokey von Idencom. Endungen *.up3, *.up4, *.up5</td></tr>
    </table>
    \param[in] filePath Dateiname.
    \param[out] errorNumber Fehlernummer bei fehlerhafter Ausführung.

    \retval 0 Fehler bei Ausführung.
    \retval 1 Ausführung war erfolgreich.
*/
DF_API_EXTERN_DECORATE int DF_API_CALL DFCUploadModule( int connectionId, int deviceAddress, int fileType, const char *filePath, int *errorNumber );

#endif

#endif
