#include "api/df_active_mode.h"
#include "datafox_library_iv.h"

DF_API_EXTERN_DECORATE int DF_API_CALL DFCStartActiveConnection( const char *networkAddress, int listenPort, int deviceAddress, int timeout, int aliveTimeout, int maxConnections, int infoFlags, int unused_setTo0, int *errorNumber )
{
    int id = LOG_CHANNEL_ID;
    ChannelBridge *channelBridge = theApp.channelBridge( id );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS );

    channelBridge->logBegin( "%s", __func__ );

    // zulaessige Bereiche pruefen
    if ( networkAddress == nullptr || inet_addr( networkAddress ) == INADDR_NONE
        || errorNumber == nullptr || maxConnections < 1 || maxConnections > 140 ) 
    {
        channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER );

        if ( errorNumber )
        {
            *errorNumber = channelBridge->getLastErrorNumber( 0 );
        }

        channelBridge->logEnd( "\t%s\t%i\t%i\t%i\t%i\t%i\t%i\t%i\t%i\t(%i, %i)", 
            (networkAddress != nullptr) ? networkAddress : "", listenPort, deviceAddress, timeout, aliveTimeout, maxConnections, infoFlags, unused_setTo0,
            (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( 0 ) );

        return 0;
    }

    // Fehlerobjekt initalisieren.
    *errorNumber = IV_ERROR_SUCCESS;

    // 2012_07_09 (BO) V04.02.03: deviceID fix auf 0 gestellt.
    int iRes = theApp.startActiveMode( networkAddress, listenPort, 0, timeout, aliveTimeout, maxConnections, infoFlags, unused_setTo0, errorNumber );

    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->setLastErrorNumber( 0, *errorNumber );
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( 0 ) );
    }

    channelBridge->logEnd( "\t%s\t%i\t%i\t%i\t%i\t%i\t%i\t%i\t%i\t(%i, %i)", 
        networkAddress, listenPort, deviceAddress, timeout, aliveTimeout, maxConnections, infoFlags, unused_setTo0, *errorNumber, iRes, channelBridge->getLastErrorNumber( 0 ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCStopActiveConnection( int *errorNumber )
{
    int id = LOG_CHANNEL_ID;
    ChannelBridge *channelBridge = theApp.channelBridge( id );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr ) 
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS );

    channelBridge->logBegin( "%s", __func__ );

    // zulaessige Bereiche pruefen
    if ( errorNumber == nullptr )
    {
        channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER );

        channelBridge->logEnd( "\t%i\t(%i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( 0 ) );
        return 0;
    }

    // Fehlerobjekt initalisieren.
    *errorNumber = IV_ERROR_SUCCESS;

    int iRes = theApp.stopActiveMode();

    channelBridge->logEnd( "\t%i\t(%i, %i)", *errorNumber, iRes, channelBridge->getLastErrorNumber( 0 ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCGetFirstActiveConnectionId()
{
    int id = LOG_CHANNEL_ID;
    ChannelBridge *channelBridge = theApp.channelBridge( id );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return -1;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS );

    channelBridge->logBegin( "%s", __func__ );

    int iRes = -1;
    // Besteht die Aktive Verbindung
    if ( theApp.isActiveMode() ) 
    {
        std::lock_guard<std::recursive_mutex> guard( theApp._mutexMapChannelBridge );
        // Ersten Kanal im Active-Mode ermitteln.
        for ( auto iter = theApp._mapChannelBridge.begin(); iter != theApp._mapChannelBridge.end(); ++iter ) 
        {
            if ( iter->first < ACTIVE_CHANNEL_UPPER_BOUND )
            {
                if ( iter->second->isActive() )
                {
                    iRes = iter->first;
                    break;
                }
            }
            else
            {
                // Key liegt ausserhalb der Suchgrenze.
                break;
            }
        }
    }

    channelBridge->logEnd( "\t(%i, %i)", iRes, channelBridge->getLastErrorNumber( 0 ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCGetFirstActiveChannelID()
{
    return DFCGetFirstActiveConnectionId();
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCGetNextActiveConnectionId( int previousId )
{
    int id = LOG_CHANNEL_ID;
    ChannelBridge *channelBridge = theApp.channelBridge( id );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return -1;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS );

    channelBridge->logBegin( "%s\t%i", __func__, previousId );

    // Naechste aktive Verbindung ermitteln
    int iRes = -1;
    // Besteht die Aktive Verbindung
    if ( theApp.isActiveMode() ) 
    {
        std::lock_guard<std::recursive_mutex> guard( theApp._mutexMapChannelBridge );
        // Ersten Kanal im Active-Mode ermitteln.
        for ( auto iter = theApp._mapChannelBridge.begin(); iter != theApp._mapChannelBridge.end(); ++iter ) 
        {
            if ( iter->first < ACTIVE_CHANNEL_UPPER_BOUND )
            {
                if ( iter->first > previousId && iter->second->isActive() )
                {
                    iRes = iter->first;
                    break;
                }
            }
            else
            {
                // Key liegt ausserhalb der Suchgrenze.
                break;
            }
        }
    }

    channelBridge->logEnd( "\t(%i, %i)", iRes, channelBridge->getLastErrorNumber( 0 ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCGetNextActiveChannelID( int previousId )
{
    return DFCGetNextActiveConnectionId( previousId );
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCGetInfoActiveChannel( int connectionId, char *value, int *valueLength )
{
    int id = LOG_CHANNEL_ID;
    ChannelBridge *channelBridge = theApp.channelBridge( id );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    int iRes = 0;

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS );

    channelBridge->logBegin( "%s\t%i", __func__, connectionId );

    // zulaessige Bereiche pruefen
    if ( value == nullptr || valueLength == nullptr ) 
    {
        channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER );

        channelBridge->logEnd( "\t\t(%i, %i)", 0, channelBridge->getLastErrorNumber( 0 ) );
        return 0;
    }

    // Besteht die Aktive Verbindung
    if ( theApp.isActiveMode() ) 
    {
        std::lock_guard<std::recursive_mutex> guard( theApp._mutexMapChannelBridge );
        if ( theApp._mapChannelBridge.count( connectionId ) && theApp._mapChannelBridge[connectionId]->isActive() ) 
        {
            Channel4 &channel = theApp._mapChannelBridge[connectionId]->channel4_;
            // Aktuellen Infostring aus Kanal übernehmen.
            if ( channel.get_active_connection_info( value, valueLength ) )
            {
                iRes = 1;
            }
        }
    }

    channelBridge->logEnd( "\t%s\t%i\t(%i, %i)", value, *valueLength, iRes, channelBridge->getLastErrorNumber( 0 ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCSetRecordAvailable( int enabled )
{
    int id = LOG_CHANNEL_ID;
    ChannelBridge *channelBridge = theApp.channelBridge( id );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS );

    channelBridge->logBegin( "%s\t%i", __func__, enabled );

    int iRes = theApp.SetRecordAvailable( enabled );

    channelBridge->logEnd( "\t(%i, %i)", iRes, channelBridge->getLastErrorNumber( 0 ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCRecordAvailable( int *connectionId, int *deviceAddress, char *infoText, int *infoTextLength, int *errorNumber )
{
    int id = LOG_CHANNEL_ID;
    ChannelBridge *channelBridge = theApp.channelBridge( id );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS );

    channelBridge->logBegin( "%s", __func__ );

    // zulaessige Bereiche pruefen
    if ( connectionId == nullptr || deviceAddress == nullptr 
        || infoTextLength == nullptr || (infoText == nullptr && *infoTextLength > 0) || errorNumber == nullptr ) 
    {
        channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER );

        if ( errorNumber )
        {
            *errorNumber = channelBridge->getLastErrorNumber( 0 );
        }

        channelBridge->logEnd( "\t\t\t\t\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( 0 ) );
        return 0;
    }

    // Fehlerobjekt initalisieren.
    *errorNumber = IV_ERROR_SUCCESS;

    int iRes = 0;
    // Ist das Melden von Datenaetzen aktiviert?
    if ( theApp.IsNotifyRecord() ) 
    {
        do 
        {        
            if ( (iRes = theApp.RecordAvailable( connectionId, deviceAddress )) == IV_ERROR_SUCCESS ) 
            {
                std::lock_guard<std::recursive_mutex> guard( theApp._mutexMapChannelBridge );
                if ( theApp._mapChannelBridge.count( *connectionId ) )
                {
                    Channel4 &channel = theApp._mapChannelBridge[*connectionId]->channel4_;
                    // Aktuellen Infostring aus Kanal übernehmen.
                    if ( *infoTextLength > 0 && channel.get_active_connection_info( infoText, infoTextLength ) == false )
                    {
                        channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER );
                        iRes = 0;
                    }
                    else
                    {
                        iRes = 1;
                    }
                }
                else
                {
                    // Wiederholung, da diese Kanalnummer nicht verfuegbar.
                    continue;
                }
            } 
            else if ( iRes > IV_ERROR_SUCCESS ) 
            {
                channelBridge->setLastErrorNumber( 0, IV_ERROR_PROCESSING );
                iRes = 0;
            }
            else 
            {
                channelBridge->setLastErrorNumber( 0, -iRes );
                iRes = 0;
            }
        } while ( false );
    } 
    else 
    {
        channelBridge->setLastErrorNumber( 0, IV_ERROR_AC_NO_NOTIFY_RECORD );
    }

    if ( channelBridge->getLastErrorNumber( 0 ) != IV_ERROR_SUCCESS )
    {
        *errorNumber = channelBridge->getLastErrorNumber( 0 );
    }

    channelBridge->logEnd( "\t%i\t%i\t%s\t%i\t%i\t(%i, %i)", *connectionId, *deviceAddress, infoText, *infoTextLength, *errorNumber, iRes, channelBridge->getLastErrorNumber( 0 ) );
    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCBindDeviceToChannel( int connectionId, int deviceType, int deviceSerial, const char *deviceIp, int *errorNumber )
{
    int id = LOG_CHANNEL_ID;
    ChannelBridge *channelBridge = theApp.channelBridge( id );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    int iRes = 0;

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS );

    channelBridge->logBegin( "%s\t%i\t%i\t%i", __func__, connectionId, deviceType, deviceSerial );

    // Liegen die Werte im gueltigen Bereich.
    if ( connectionId < 1 || connectionId >= ACTIVE_CHANNEL_LOWER_BOUND 
        || (deviceType != -1 && (deviceType < 0 || deviceType >= IV_VALUE_TYPE_DEVICE_COUNT)) 
        || (deviceIp != nullptr && strlen( deviceIp ) && inet_addr( deviceIp ) == INADDR_NONE) 
        || errorNumber == nullptr )
    {
        channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER );

        if ( errorNumber )
        {
            *errorNumber = channelBridge->getLastErrorNumber( 0 );
        }

        channelBridge->logEnd( "\t%s\t%i\t(%i, %i)", (deviceIp != nullptr) ? deviceIp : "", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( 0 ) );
        return 0;
    }

    // Fehlerobjekt initalisieren.
    *errorNumber = IV_ERROR_SUCCESS;

    // Es gibt 4 moegliche mappings
    // 1. Loeschen eines bestehenden mappings
    // 2. map channelID => deviceType + deviceSerial
    // 3. map channelID => deviceType + deviceSerial + deviceIP
    // 4. map channelID => deviceIP
    ChannelBridge *channel = theApp.channelBridge( connectionId );
    if ( channel->isBind() && deviceType == -1 && (deviceIp == 0 || !strlen( deviceIp )) ) 
    {
        // Bindung wird entfernt.

        // Bindungsstruktur aufbauen und zuweisen
        StructBindingInfo bindingInfo;
        memset( &bindingInfo, 0, sizeof( StructBindingInfo ) );
        bindingInfo.channelId = connectionId;

        // Hinzufuegen zur Liste der zu bearbeitenden Eintraege.
        {
            std::lock_guard<std::mutex> guard( theApp._mutexListBindingInfo );
            theApp._listBindPending.push_back( bindingInfo );  
        }

        // Wenn der Active-Mode-Thread läuft, dann darauf warten, dass die BindingInfo verarbeitet wurde.
        if ( theApp.isActiveMode() )
        {
            std::unique_lock<std::mutex> lk( theApp._mutexConditionBindingInfo );
            theApp._conditionListBindingInfo.wait_for( lk, std::chrono::seconds( 10 ) );
        }

        iRes = 1;
    } 
    else if ( (deviceType > -1 && deviceSerial < 1) || (deviceType == -1 && (deviceIp == nullptr || !strlen( deviceIp ))) ) 
    {
        channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER );
    } 
    else if ( channel->isBind() || channel->isOpen() )
    {
        // Der Kanal darf nicht bereits gebunden oder geoeffnet sein.
        channelBridge->setLastErrorNumber( 0, IV_ERROR_CHANNEL_USED );
    }
    else
    {
        // Bindungsstruktur aufbauen und zuweisen
        StructBindingInfo bindingInfo;
        memset( &bindingInfo, 0, sizeof( StructBindingInfo ) );
        bindingInfo.channelId = connectionId;
        bindingInfo.bi.state = 1;
        bindingInfo.bi.typeDevice = deviceType;
        bindingInfo.bi.serial = deviceSerial;
        if ( deviceIp != 0 )
        {
            bindingInfo.bi.ip = deviceIp;
        }

        // Hinzufuegen zur Liste der zu bearbeitenden Eintraege.
        {
            std::lock_guard<std::mutex> guard( theApp._mutexListBindingInfo );
            theApp._listBindPending.push_back( bindingInfo );
        }

        // Wenn der Active-Mode-Thread läuft, dann darauf warten, dass die BindingInfo verarbeitet wurde.
        if ( theApp.isActiveMode() )
        {
            std::unique_lock<std::mutex> lk( theApp._mutexConditionBindingInfo );
            theApp._conditionListBindingInfo.wait_for( lk, std::chrono::seconds( 10 ) );
        }

        iRes = 1;
    }

    if ( iRes == 0 )
    {
        *errorNumber = channelBridge->getLastErrorNumber( 0 );
    }

    channelBridge->logEnd( "\t%s\t%i\t(%i, %i)", (deviceIp != nullptr) ? deviceIp : "", *errorNumber, iRes, channelBridge->getLastErrorNumber( 0 ) );

    return iRes;
}
