#ifndef FILEINTELHEX_H
#define FILEINTELHEX_H

// Grunddaten und Grundoffsets im Datenarray.
#define INTELHEX_DEFAULT_DATA_LENGTH  20  // Anzahl Datenbyte pro Datensatz
#define INTELHEX_RECORD_MARK          ':' // Startzeichen eines Datensatzes
#define INTELHEX_OFFSET_START          0  // start of record
#define INTELHEX_OFFSET_LENGTH         1  // byte count
#define INTELHEX_OFFSET_OFFSET         3  // load address
#define INTELHEX_OFFSET_TYPE           7  // record type
#define INTELHEX_OFFSET_DATA           9  // record data
#define INTELHEX_OFFSET_CHECKSUM       7  // record checksum

// Groessen fuer Textpuffer der Intelhex-Records.
#define INTELHEX_STRING_LENGTH_HEAD    9 // Anzahl Zeichen fuer Recordbeginn zu Daten
#define INTELHEX_STRING_LENGTH_FRAME   (INTELHEX_STRING_LENGTH_HEAD + 2) // Anzahl Zeichen fuer Recordbeginn zu Daten + CRC
#define INTELHEX_STRING_LENGTH_MAX     (INTELHEX_STRING_LENGTH_FRAME + 510) // Anzahl Zeichen fuer Recordframe + Daten
#define INTELHEX_STRING_LENGTH_DATA    (INTELHEX_STRING_LENGTH_MAX - INTELHEX_STRING_LENGTH_FRAME) // Maximale Anzahl Datenbytes

#include <string>
#include <boost/format.hpp>
#include <locale>

#include <vector>
#include <boost/dynamic_bitset.hpp>

class FileIntelHex
{
public:
    FileIntelHex();
    virtual ~FileIntelHex();

public:
    enum class Error_state
    {
        none,                       // Kein Fehler im Zustand.
        file_open,                  // Oeffnen der Datei gescheitert.
        // Error_state der IntelHex-Verarbeitung
        record_empty,               // Die uebergebene Datenzeile ist leer.
        record_crc,                 // Der Berechnete CRC stimmt nicht mit dem im Datensatz ueberein.
        record_format,              // Der Datensatz weist fehler auf. Puffer zu kurz, lang oder fehlerhafte Laengenabgabe.
        record_buffer_overflow,     // Der Datensatz geht ueber die Grenze des uebergebenen Puffers hinweg.
        file_write,                 // Es ist ein Dateifehler beim Schreiben der Daten aufgetreten.
        file_read,                  // Es ist ein Dateifehler beim Lesen der Daten aufgetreten.
        file_empty,                 // In der Datei liegt kein Datensatz vor.
        // Error_state der Dateiverarbeitung
        crc_fail,                   // CRC von Daten falsch.
        format,                     // Bereichsfehler bei eingelesenen Daten usw..
        no_data,                    // Keine Daten beim laden einer Datei.
        load_display_data,          // Laden der Anzeigedaten gescheitert.
        load_device_text,           // Laden der Anzeigetexte gescheitert.
        device_unknown,             // Unbekannter Gerätetyp der Gerätedatei.
        program_overflow,           // Programmspeicher Ueberlauf.
        record_make,                // Erzeugen eines IntelHex Datensatzes ist fehlgeschlagen.
        crypt_fix_fail,             // Verschluesselungswert ungueltig.
        crypt_none,                 // Entschluesselungsversuch ohne vorherige Verschluesselung.
        crypt_is,                   // Verschluesselungsversuch obwohl bereits verschluesselt.
        ini_key,                    // Schluesselwert aus Inidatei konnte nicht geladen werden.
        ini_key_value,              // Schluesselwert aus Inidatei liegt ausserhalb des gueltigen Berreiches.
        size,                       // Groesse der moeglichen Datenmenge ueberschritten.
        version,                    // Version der Geraetedatei ist aktueller als erwartet.
        record_end_make,            // Daten oder Endzeile fuer Hexdatei konnte nicht erstellt werden.
        head_crc,                   // Der CRC fuer den Dateikopf schlug fehl.
        firmware_version_length,    // Zeichenkette fuer die Firmwareversion ist zu kurz (< 8) oder zu lang (> 32).
        program_crc,                // Der CRC fuer den Programm-Flash schlug fehl.
        function_unavailable,       // Funktion wird nicht unterstuetzt.
        load_raw_firmware,          // Laden der Rohen Firmwaredaten gescheitert.
        version_file,               // Keine gueltige Dateiversionsangabe vorhanden.
        as_text,                    // Fehlertext steht ueber string zur Verfuegung.
        file_format,                // Das Dateiformat ist nicht wie erwartet Intel-Hex.
        line_format,                // Das Format einer Datenzeile ist falsch.
        alloc_mem,                  // Speicherallokation konnte nicht durchgefuehrt werden.
        no_file_match,              // Es wurde keine zutreffende Geraetedatei in diesem Archiv gefunden.
        serialize_file_data         // Verarbeitung der Dateidaten ist gescheitert.
    };

    enum class Record_type 
    {
        // Standard - Datensatztypen bei Intel Hex 32
        data                     = 0x00, // Data record
        end_of_file              = 0x01, // End-of-file record
        extended_segment_address = 0x02, // Extended Segment Address Record
        start_segment_address    = 0x03, // Start Segment Address Record
        extended_linear_address  = 0x04, // Extended Linear Address Record
        start_linear_address     = 0x05, // Start Linear Address Record
        // Datei der Anzeigedaten
        display_font             = 0x21,
        display_grafic           = 0x22,
        // Datei der Geraetedaten
        device_header            = 0x30,
        // Datei der Textdaten
        text                     = 0x51,
        // Datei der Touchkeydaten
        touch_key                = 0x61,
        // Datei der Displayfarben
        display_color            = 0x71,
        // Datei der WlanConfig
        wlan_config              = 0x81,
        // Datei der HardwareConfig
        hardware_config          = 0x91,
        // Datei der Hauptmenues
        main_menu                = 0xA1
    };

    // helper
public:
    virtual void resetData();

    virtual bool load( const std::string &filePath );
    virtual bool save( const std::string &filePath, bool appendToFile = false, bool appendEndLine = true );

public:
    inline unsigned char *data()
    {
        return &m_buffer.front();
    }
    inline const unsigned char *dataPointer( unsigned int offs )
    {
        return &m_buffer[offs];
    }

    inline void setSizeData( unsigned int size )
    {
        m_buffer.resize( size );
    }

    inline std::vector<unsigned char> content()
    {
        return m_buffer;
    }
    inline void setContent( std::vector<unsigned char> data )
    {
        m_buffer = data;
    }
    inline std::size_t sizeData()
    {
        return m_buffer.size();
    }
    unsigned int firstAddress()
    {
        return m_base_address;
    }
    unsigned int lastAddress()
    {
        unsigned int addr = m_base_address + static_cast<unsigned int>( m_buffer.size() );
        if ( m_buffer.size() )
        {
            addr--;
        }

        return addr;
    }
    unsigned int checksum32bit();

    unsigned int crc32( const unsigned char *data, unsigned int size, unsigned int crc = 0xFFFFFFFF, bool closure = true );
    unsigned short crc16( const unsigned char *data, unsigned int size, unsigned short crc = 0xFFFF );
protected:
    std::string m_file_path;

public:
    inline std::string filePath() const
    { 
        return m_file_path;
    }
    void setFilePath( std::string filePath )
    {
        m_file_path = filePath;
    }

protected:
    Record_type m_record_type;
    inline Record_type recordType() const
    { 
        return m_record_type;
    }
    void setRecordType( Record_type type )
    {
        m_record_type = type;
    }

protected:
    enum class Address_mode
    {
        unknown,
        segment,
        linear
    } m_address_mode;
    struct {
        unsigned short csRegister;
        unsigned short ipRegister;
        bool           exists;
    } m_start_segment_address;

    struct {
        unsigned int eipRegister;
        bool         exists;
    } m_start_linear_address;

protected:
    unsigned int m_base_address;
    boost::dynamic_bitset<> m_byte_marker;
    std::vector<unsigned char> m_buffer;

    const unsigned char m_fill_byte;
    const unsigned int m_block_size;

public:
    inline unsigned int blockSize()
    {
        return m_block_size;
    }
    inline unsigned char fillByte()
    {
        return m_fill_byte;
    }
    inline Address_mode addressMode()
    {
        return m_address_mode;
    }
    inline void setAddressMode( Address_mode mode )
    {
        m_address_mode = mode;
    }

public:
    void UINT8ToHex( unsigned char val, char *data );
    unsigned char HexToUINT8( const char *data );

    void UINT16ToHex( unsigned short val, char *data );
    unsigned short HexToUINT16( const char *data );

    bool appendCrc( std::string &data );
    bool checkCrc( const char *data );

public:
    std::string nextDataLine( std::istream &s );
    unsigned char GetCount( const char *data );
    unsigned short GetAddress( const char *data );
    unsigned char GetType( const char *data );
    unsigned char GetData( const char *data, unsigned char *dest );
    unsigned char GetData8( const char *data, unsigned char num );
    unsigned short GetData16( const char *data, unsigned char num );
    virtual std::string typedDataLine( unsigned short address, unsigned char type, unsigned char *data, unsigned char length );
    virtual std::string dataLine( unsigned short address, unsigned char *data, unsigned char length );
    virtual std::string endLine();

public:
    inline unsigned char maxDataSize()
    {
        return m_max_data_size;
    }

private:
    std::locale m_locale;
    const unsigned char m_max_data_size;

    // Fehlerbehandlung
protected:
    virtual void resetError();
public:
    virtual std::string errorString( Error_state state );

protected:
    inline void setErrorState( Error_state state )
    {
        m_error_state = state;
        if ( state != Error_state::as_text ) // Fehlertext zur Nummer zuweisen.
            m_error_text = errorString( m_error_state );
    }
public:
    inline Error_state errorState()
    {
        return m_error_state;
    }

protected:
    inline void setErrorText( const std::string &text )
    {
        m_error_state = Error_state::as_text; // Eigenen Fehlertext gesetzt.
        m_error_text = text;
    }
public:
    inline std::string errorText()
    {
        return m_error_text;
    }

protected:
    inline void setErrorDetail( const std::string &detail )
    {
        m_error_detail = detail;
    }
public:
    inline std::string errorDetail()
    {
        return m_error_detail;
    }

private:
    static Error_state m_error_state;
    static std::string m_error_text;
    static std::string m_error_detail;
};

#endif
