// file_studio_setup.cpp: Implementierung der Klasse FileStudio.
//
//////////////////////////////////////////////////////////////////////

#include "file/file_studio_setup.h"
#include <sys/stat.h>
#include <time.h>
#include <algorithm>

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

FileStudio::FileStudio() 
    : m_eep( m_eep_mem.data() )
    , m_flash_header( *reinterpret_cast<StructFlashHeader*>( m_eep_mem.data() ) )
    , m_file_header( *reinterpret_cast<StructFileStudioHeader*>( m_file_header_mem.data() ) )
    , m_flash( m_flash_mem.data() )
    , m_config( *reinterpret_cast<StructConfig*>( m_config_mem.data() ) )
    , m_flashExtended( m_extended_mem.data() )
    , m_extended_header( *reinterpret_cast<StructExtendedFlash*>( m_extended_mem.data() ) )
{
    // Initialisieren der 32bit-CRC Umgebung.
    InitCRC32Table();

    // Dokument aufsetzen
    SetDefault();
}

FileStudio::~FileStudio()
{
}

void FileStudio::SetDefault( int flags )
{
    // Eeprom 
    m_eep_mem.fill( 0xFF );

    // Header auf Standard setzen.
    SetDefaultHeader();

    // Zusaetlicher Setupspeicher
    m_flashSizeSU2 = FLASH_SIZE_SU2;
    // Flashbereiche
    m_flash_mem.fill( 0xFF );
    // Konfigbereich
    m_config_mem.fill( 0 );
    // Extended Flash wird jedoch mit 0 iniatlisiert
    m_extended_mem.fill( 0 );

    // Der erweiterte Speicherbereich wird mit einer dynamischen Struktur begonnen,
    // deren erster Parameter ein Check - UINT32 - Wert enthaelt.

    // Alles mit Null initialisieren.
    memset( &m_extended_header, 0, sizeof(StructExtendedFlash) );
    // Struktur fuer erweiterten Bereich aufsetzen
    m_extended_header.check = EXTENDED_CHECK;
    m_extended_header.size  = sizeof(StructExtendedFlash);
    m_extended_header.offsetFreeScope = m_extended_header.size;
    // Flag fuer Verarbeitung
    m_extended_header.flags = EXTF_NEW;

    // Die Listenangaben werden bei gesetzten flag 1 beibehalten?
    if ( (flags & 0x01) != 0x01 )
    {
        // Listenpfade löschen
        for ( auto& s : m_listFileName )
        {
            s.clear();
        }
        for ( auto& s : m_listEntranceFileName )
        {
            s.clear();
        }
        for ( auto& s : m_listEntrance2FileName )
        {
            s.clear();
        }
        for ( auto& s : m_strExtendedFilePath )
        {
            s.clear();
        }
    }
}

void FileStudio::SetDefaultHeader()
{
    // Dateiheader
    m_file_header_mem.fill( 0 );
    
    // Dateikopf aufsetzen
    m_file_header.check = magic_header_number();
    m_file_header.bugfixSize = sizeof_StructFileStudioHeader_1;
    m_file_header.size = sizeof(StructFileStudioHeader);

    // Ggf. vorher gesetzte Werte wieder eintragen.
    m_file_header.studioVersion = header_version( m_studioVersion.master(), m_studioVersion.major(), m_studioVersion.minor(), m_studioVersion.build() );

    // Es wird mindestens diese Version benoetigt.
    // Die Version wird ggf. auf die benoetigte in der jeweiligen MakeImage-Funktion angepasst.
    NeedStudioVersion( "04.02.02.08" );
    NeedLibraryVersion( "04.02.02.08" );
}

/*
return:
0: Fehler beim Erstellen der Headerdaten eingetreten.
1: Dateiheaderdaten erfolgreich erstellt und eingetragen.
*/
int FileStudio::buildHeader()
{
    // CRC der Dateibereiche
    m_file_header.crcEep = GetCrcEep();
    m_file_header.crcEepExtended = 0; // War in Version 1, CRC von Restheaderbereich berechnen und eintragen
    m_file_header.crcFlash = GetCrcFlash();
    m_file_header.crcListFile = GetCrcListFile();
    m_file_header.crcConfig = GetCrcConfig();
    m_file_header.crcFlashExtended = GetCrcFlashExtended();
    m_file_header.crcListEntranceFile = GetCrcListEntranceFile();
    m_file_header.crcListEntrance2File = GetCrcListEntrance2File();
    m_file_header.crcListExtendedFile = GetCrcListExtendedFile();
    
    // Size und CRC der Ersten Version immer eintragen.
    m_file_header.bugfixSize = sizeof_StructFileStudioHeader_1;
    m_file_header.bugfixCRC = GetCRC32( m_file_header_mem.data(), m_file_header.bugfixSize - sizeof(unsigned int) );
    
    // Size dieser Version eintragen, wenn noch nicht aktuell.
    if ( m_file_header.size < sizeof(StructFileStudioHeader) )
        m_file_header.size = sizeof(StructFileStudioHeader);

    // Gesamt-CRC fuer Header berechnen und eintragen.
    *(unsigned int*)&m_file_header_mem[m_file_header.size - sizeof(unsigned int)] = GetCRC32( m_file_header_mem.data(), m_file_header.size - sizeof(unsigned int) );

    // Wenn eine aktueller Version als hier aktuell vorhanden geladen wurde:
    // Die Groessenangabe und Daten werden beibehalten.
    // Es kann davon ausgegangen werden, dass im vorangegangenen Ladeverlauf die Mindestversion des Studio oder Library geprueft wurde.
    // Wenn das Laden erfolgreich durchgefuehrt werden konnte handelt es sich um hinzugefuegte Daten die nicht so relevant sind oder ggf. gar nicht verwendet werden.
    // Es wird in einem solchen Fall mit der Struktur gearbeitet wie sie ist.

    return 1;
}

/*
return:
0: Laengenangabe oder Pruefsumme des Headers falsch.
1: Pruefungen erfolgreich und ohne Fehler durchgefuehrt.
*/
int FileStudio::checkHeader()
{
    if ( IsHeader() )
    {
        // Version 1, Angaben sind bei allen Versionen verfuegbar.
        // Laengenangabe liegt im moeglichen Bereich?
        if ( m_file_header.bugfixSize != sizeof_StructFileStudioHeader_1 )
        {
            return 0;
        }

        // Pruefen ob die Headerdaten ggf. korrupt sind?
        if ( *(unsigned int*)&m_file_header_mem[m_file_header.bugfixSize - sizeof(unsigned int)] != GetCRC32( m_file_header_mem.data(), m_file_header.bugfixSize - sizeof(unsigned int) ) )
            return 0;


        // Version >= 2
        // Laengenangabe liegt im moeglichen Bereich?
        if ( m_file_header.size != 0 && (m_file_header.size < sizeof_StructFileStudioHeader_2 || m_file_header.size > m_file_header_mem.size()) )
        {
            return 0;
        }
        else if ( m_file_header.size != 0 )
        {
            // Pruefen ob die Headerdaten ggf. korrupt sind?
            if ( *(unsigned int*)&m_file_header_mem[m_file_header.size - sizeof(unsigned int)] != GetCRC32( m_file_header_mem.data(), m_file_header.size - sizeof(unsigned int) ) )
                return 0;

            // CRC loeschen, da geprueft und sich ggf. mit Werten dieser Stukturversion ueberlagert.
            *(unsigned int*)&m_file_header_mem[m_file_header.size - sizeof(unsigned int)] = 0;
        }
        else // size == 0
        {
            // Dahinterliegende Daten als ungueltig behandeln.
            memset( &m_file_header_mem[m_file_header.bugfixSize], 0, m_file_header_mem.size() -  m_file_header.bugfixSize );
        }

        // Version 1
        // CRC loeschen, da geprueft.
        *(unsigned int*)&m_file_header_mem[m_file_header.bugfixSize - sizeof(unsigned int)] = 0;

        // Ist Version 2?
        //if ( m_file_header.size == sizeof_StructFileStudioHeader_2 )

        // Wenn eine aktueller Version als hier aktuell vorhanden geladen wuerde:
        // Es kann davon ausgegangen werden, dass im nachfolgenden Ladeverlauf die Mindestversion des Studio oder Library nicht erreicht wird,
        // wenn nicht, dann handelt es sich um hinzugefuegte Daten die nicht so relevant sind oder ggf. gar nicht verwendet werden.
        // Es wird in einem solchen Fall mit der Struktur geargeitet wie sie ist.
    }

    return 1;
}

/**
 * @brief Lesen eines Strings, mit führender Längenangabe, aus einer Datei.
 *
 * @param[out] s Gelesener String.
 * @param[in] fp Datei aus der ab der aktuellen Position gelesen wird.
 *
 * @returns Anzahl gelesener Bytes inkl. Bytes für die Längenangabe.
 */
std::size_t FileStudio::read_cstring( std::string& s, std::FILE* fp )
{
    // String nach MFC - CString Konvention lesen.
    std::size_t result = 0;
    bool fail = false;
    std::size_t size = 0;
    unsigned char len[2];
    s.clear();

    // Stringlänge ermitteln. Die Längenangabe liegt entweder mit 1 Byte vor dem String solange
    // sie kleiner als 255 ist, ansonsten folgt einem Byte mit dem Wert 255 ein 16bit Wert.
    result = std::fread( len, sizeof(unsigned char), 1, fp ); // ein Byte Länge
    if ( result == sizeof( unsigned char ) ) 
    {
        if ( len[0] == 0xFF )
        {
            result += std::fread( len, sizeof( unsigned char ), 2, fp ); // zwei Byte Länge
            if ( result == 3 * sizeof( unsigned char ) ) 
            {
                size = *reinterpret_cast<unsigned short*>( len );
            }
            else
            {
                // Fehler beim Ermitteln der Länge.
                fail = true;
            }
        }
        else
        {
            size = len[0];
        }

        if ( !fail  )
        {
            // String für die zu lesende Anzahl Zeichen anpassen.
            s.resize( size, 0 );
        }
    }
    else
    {
        // Fehler beim Ermitteln der Länge.
        fail = true;
    }

    if ( !fail )
    {
        if ( (size = std::fread( &s[0], sizeof( unsigned char ), s.size(), fp )) != s.size() )
        {
            // String an die wirklich gelesene Anzahl anpassen.
            s.resize( size, 0 );
        }

        // Gelesene Länge anpassen.
        result += size;
    }

    return result;
}

/**
 * @brief Schreiben eines Strings, mit führender Längenangabe, in eine Datei.
 *
 * @param[in] s Zu schreibender String.
 * @param[in] fp Datei in die ab der aktuellen Position geschrieben wird.
 *
 * @returns Anzahl geschriebener Bytes inkl. Bytes für die Längenangabe.
 */
std::size_t FileStudio::write_cstring( const std::string& s, FILE* fp )
{
    // String nach MFC - CString Konvention schreiben.
    std::size_t result = 0;
    std::size_t size;
    bool fail = false;

    if ( s.size() > 0xFF )
    {
        unsigned char len[3];
        len[0] = 0xFF; // Markierung das eine 16bit Laengenangabe folgt.
        *reinterpret_cast<unsigned short*>( &len[1] ) = static_cast<unsigned short>( s.size() );
        result = std::fwrite( len, sizeof( unsigned char ), sizeof( len ), fp );
        if ( result != sizeof( len ) )
        {
            fail = true;
        }
    }
    else
    {
        unsigned char len = static_cast<unsigned char>( s.size() );
        result = fwrite( &len, sizeof( unsigned char ), sizeof( len ), fp );
        if ( result != sizeof( len ) )
        {
            fail = true;
        }
    }

    // String rausschreiben
    if ( !fail )
    {
        if ( (size = std::fwrite( s.c_str(), sizeof( unsigned char ), s.size(), fp )) != s.size() )
        {
            // Fehler beim Schreiben des gesamten Strings.
        }

        // Geschriebene Länge anpassen.
        result += size;
    }

    return result;
}

/**
 * @brief Laden der Setupdaten aus der angegebenen Datei.
 *
 * @param[in] file_path Angabe der zu ladenden Datei.
 * @param[out] error Erhält bei Fehler die entsprechende Fehlernummer.
                     <table class="paramvalues">
                         <tr><th>Wert</th><th>Bedeutung</th></tr>
                         <tr><td>-1</td><td> </td></tr>
                         <tr><td>-2</td><td>Öffnen der Datei gescheitert. </td></tr>
                         <tr><td>-3</td><td>Lesen aus der Datei gescheitert. </td></tr>
                         <tr><td>-4</td><td>Prüfsummenfehler im Dateiinformationblock. </td></tr>
                         <tr><td>-5</td><td>Zum Öffnen der Setupdatei wird mindestens das DatafoxStudioIV - Version 
                                            m_file.m_file_header.minStudioVersion benötigt.
                                            In der Setupdatei werden Einstellungen und Erweiterungen verwendet, 
                                            die durch dieses DatafoxStudioIV noch nicht unterstützt werden. 
                                            Bitte aktualisiseren Sie die Anwendung auf die angegebene Mindestversion.
                                    </td></tr>
                        <tr><td>-6</td><td>Warnung: Um alle in der Setupdatei vorliegenden Informationen zu Unterstützen 
                                           wird mindestens das DatafoxStudioIV - Version m_file.m_file_header.shouldStudioVersion 
                                           benötigt. Ein Abspeichern der Setupdatei mit dem vorliegenden DatafoxStudioIV 
                                           führt zu einem Verlust von Verwaltungsspezifischen Dateidaten. 
                        </td></tr>
                        <tr><td>-7</td><td>Zum Öffnen der Setupdatei wird mindestens die DFComDLL - Version m_file.m_file_header.minLibraryVersion 
                                           benötigt. In der Setupdatei werden Einstellungen und Erweiterungen verwendet, 
                                           die durch diese DFComDLL noch nicht unterstützt werden. Bitte aktualisiseren 
                                           Sie die Bibliothek auf die angegebene Mindestversion.
                        </td></tr>
                        <tr><td>-8</td><td>Warnung: Um alle in der Setupdatei vorliegenden Informationen zu Unterstützen 
                                           wird mindestens die DFComDLL - Version m_file.m_file_header.shouldLibraryVersion 
                                           benötigt. Ein Abspeichern der Setupdatei mit dem vorliegenden DatafoxStudioIV 
                                           führt zu einem Verlust von Verwaltungsspezifischen Dateidaten.
                        </td></tr>
                        <tr><td>-9</td><td>Unerwartetes Setupdateiformat.</td></tr>
                        <tr><td>-10</td><td> </td></tr>
                        <tr><td>-11</td><td> </td></tr>
                        <tr><td>-12</td><td> </td></tr>
                        <tr><td>-13</td><td> </td></tr>
                        <tr><td>-14</td><td> </td></tr>
                        <tr><td>-15</td><td> </td></tr>
                        <tr><td>-16</td><td> </td></tr>
                        <tr><td>-17</td><td> </td></tr>
                        <tr><td>-18</td><td>Prüfsummenfehler im Dateninhalt.</td></tr>
                     </table>

 * @param[in] flags Zusätzliche Angabe über durchzuführende Prüfungen während dem Ladevorgang.
                     <table class="paramvalues">
                         <tr><th>Wert</th><th>Bedeutung</th></tr>
                         <tr><td>1</td><td>Ignoriere die benötigte Mindestversionsangabe des DatafoxStudioIV.</td></tr>
                         <tr><td>2</td><td>Ignoriere die Angabe der sollte-Version des DatafoxStudioIV.</td></tr>
                         <tr><td>4</td><td>Ignoriere die benötigte Mindestversionsangabe des DatafoxLibraryIV.</td></tr>
                         <tr><td>8</td><td>Ignoriere die Angabe der sollte-Version des DatafoxLibraryIV.</td></tr>
                     </table>
 *
 * @retval 0 Fehler bei Ausführung.
 * @retval 1 Ausführung war erfolgreich.
 */
int FileStudio::load( const std::string& file_path, int *error, Enum_flag_ignore_version flags )
{
    int res = 1;
    struct stat st;
    unsigned int sizeSetup;
    int listFileSize = 0;
    int entranceListFileSize = 0;
    int entrance2ListFileSize = 0;
    int extendedFilePathSize = 0;

    if ( stat( file_path.c_str(), &st ) != 0 )
    {
        *error = -1;
        return 0;
    }

    FILE *fp = fopen( file_path.c_str(), "rb" );
    if ( fp == NULL )
    {
        *error = -2;
        return 0;
    }

    // EEPROM - lesen
    if ( fread( m_eep_mem.data(), sizeof(unsigned char), m_eep_mem.size(), fp ) != m_eep_mem.size() )
    {
        *error = -3;
        res = 0;
        goto error_label;
    }

    // Dateiheader lesen.
    if ( fread( m_file_header_mem.data(), sizeof(unsigned char), m_file_header_mem.size(), fp ) != m_file_header_mem.size() )
    {
        *error = -3;
        res = 0;
        goto error_label;
    }

    // Headerdaten pruefen, ggf. Anpassungen durchfuehren.
    if ( !checkHeader() )
    {
        // Header hat CRC Fehler.
        *error = -4;
        res = 0;
        goto error_label;
    }

    if ( IsHeader() )
    {
        // Mindestversion des Studio pruefen.
        if ( !(flags & fiv_min_studio) && m_file_header.minStudioVersion )
        {
            SoftwarebuildVersion minVersion( m_file_header.minStudioVersion );
            if ( minVersion > m_studioVersion )
            {
                *error = -5;
                res = 0;
                goto error_label;
            }
        }
        // Sollteversion des Studio pruefen?
        if ( !(flags & fiv_should_studio) && !(m_file_header.flags & hfiv_studio) && m_file_header.shouldStudioVersion )
        {
            SoftwarebuildVersion shouldVersion( m_file_header.shouldStudioVersion );
            if ( shouldVersion > m_studioVersion )
            {
                *error = -6;
                res = 0;
                goto error_label;
            }
        }

        // Mindestversion der Library pruefen.
        if ( !(flags & fiv_min_library) && m_file_header.minLibraryVersion )
        {
            SoftwarebuildVersion minVersion( m_file_header.minLibraryVersion );
            if ( minVersion > m_libraryVersion )
            {
                *error = -7;
                res = 0;
                goto error_label;
            }
        }
        // Sollteversion der Library pruefen?
        if ( !(flags & fiv_should_library) && !(m_file_header.flags & hfiv_library) && m_file_header.shouldLibraryVersion )
        {
            SoftwarebuildVersion shouldVersion( m_file_header.shouldLibraryVersion );
            if ( shouldVersion > m_libraryVersion )
            {
                *error = -8;
                res = 0;
                goto error_label;
            }
        }
    }
    else
    {
        // Den Header im Standard wiederherstellen.
        SetDefaultHeader();
    }

    // Setupgroesse
    sizeSetup = endian_reverse( m_flash_header.FLASH_SIZE_SUS );
    m_flashSizeSU2 = 0;
    // Ermittlung der Groesse des zu lesenden Speicherblocks fuer Flash.
    if ( m_flash_header.FILE_VERSION == 0xDF02 && sizeSetup > FLASH_SIZE_SU )
    {
        m_flashSizeSU2 = FLASH_SIZE_SU2;
    }

    // Flashbereich - lesen
    if ( fread( m_flash_mem.data(), sizeof(unsigned char), FLASH_SIZE_SU + m_flashSizeSU2, fp ) != (FLASH_SIZE_SU + m_flashSizeSU2) )
    {
        *error = -3;
        res = 0;
        goto error_label;
    }

    // ListFile - lesen
    for ( auto i = 0u; i < m_listFileName.size(); i++ )
    {
        // String inkl. Längenangabe aus Datei lesen.
        auto size = read_cstring( m_listFileName[i], fp );
        if ( size == 0 )
        {
            *error = -3;
            res = 0;
            goto error_label;
        }

        // Gesamtanzahl gelesener Zeichen halten.
        listFileSize += size;
    }

    // Konfiguration - lesen
    if ( fread( m_config_mem.data(), sizeof(unsigned char), m_config_mem.size(), fp ) != m_config_mem.size() )
    {
        *error = -3;
        res = 0;
        goto error_label;
    }

    // Sind ab aktueller Postition noch 16kByte erweiterter Speicher vorhanden?
    if ( (unsigned int)st.st_size >= 
        (m_eep_mem.size() + m_file_header_mem.size() + FLASH_SIZE_SU + m_flashSizeSU2 + listFileSize + m_config_mem.size() 
            + m_extended_mem.size()) )
    {
        // Erweiterter Bereich - lesen: 16kByte Extended Flash
        if ( fread( m_extended_mem.data(), sizeof(unsigned char), m_extended_mem.size(), fp ) != m_extended_mem.size() )
        {
            *error = -3;
            res = 0;
            goto error_label;
        }

        // existiert?
        if ( m_extended_header.check != EXTENDED_CHECK )
        {
            *error = -9;
            res = 0;
            goto error_label;
        }
    } 
    else
    {
        *error = -9;
        res = 0;
        goto error_label;
    }

    // Sind noch ListFile fuer Zutrittskontrolle vorhanden?
    if ( (unsigned int)st.st_size >= 
        (m_eep_mem.size() + m_file_header_mem.size() + FLASH_SIZE_SU + m_flashSizeSU2 + listFileSize + m_config_mem.size() 
            + m_extended_mem.size() + m_listEntranceFileName.size()) )
    {
        // ListFile - lesen
        for ( auto i = 0u; i < m_listEntranceFileName.size(); i++ )
        {
            // String inkl. Längenangabe aus Datei lesen.
            auto size = read_cstring( m_listEntranceFileName[i], fp );
            if ( size == 0 )
            {
                *error = -3;
                res = 0;
                goto error_label;
            }

            // Gesamtanzahl gelesener Zeichen halten.
            entranceListFileSize += size;
        }

        // Loeschen, wenn ZK nicht aktiviert.
        if ( m_config.ENABLE_ENTRANCE == 0 )
        {
            for ( auto& s : m_listEntranceFileName )
            {
                s.clear();
            }
            // CRC von Pruefung ausschliessen.
            m_file_header.crcListEntranceFile = 0;
            // CRC ungueltig.
            m_file_header.crc = 0;
        }
    }
    else
    {
        *error = -9;
        res = 0;
        goto error_label;
    }

    // Sind noch ListFile fuer Zutrittskontrolle 2 vorhanden?
    if ( (unsigned int)st.st_size >= 
        (m_eep_mem.size() + m_file_header_mem.size() + FLASH_SIZE_SU + m_flashSizeSU2 + listFileSize + m_config_mem.size() 
            + m_extended_mem.size() + entranceListFileSize + m_listEntrance2FileName.size()) )
    {
        // ListFile - lesen
        for ( auto i = 0u; i < m_listEntrance2FileName.size(); i++ )
        {
            // String inkl. Längenangabe aus Datei lesen.
            auto size = read_cstring( m_listEntrance2FileName[i], fp );
            if ( size == 0 )
            {
                *error = -3;
                res = 0;
                goto error_label;
            }

            // Gesamtanzahl gelesener Zeichen halten.
            entrance2ListFileSize += size;
        }

        // Loeschen, wenn ZK nicht aktiviert.
        if ( m_config.ENABLE_ENTRANCE2 == 0 )
        {
            for ( auto& s : m_listEntrance2FileName )
            {
                s.clear();
            }
            // CRC von Pruefung ausschliessen.
            m_file_header.crcListEntrance2File = 0;
            // CRC ungueltig.
            m_file_header.crc = 0;
        }
    }
    else
    {
        /*
        *error = -9;
        res = 0;
        goto error_label;
        */
        // (BO)Bugfix: 2012_03_19
        // Diese Dateiangaben muessen noch nicht verfuegbar sein.
        // Es koennten Dateien fuer AEIII+ ueber die Library geladen werden!
    }

    // Sind noch ExtendedFilePath fuer Zusatzdateien vorhanden?
    if ( (unsigned int)st.st_size >= 
        (m_eep_mem.size() + m_file_header_mem.size() + FLASH_SIZE_SU + m_flashSizeSU2 + listFileSize + m_config_mem.size() 
            + m_extended_mem.size() + entranceListFileSize + entrance2ListFileSize + m_strExtendedFilePath.size()) )
    {
        // ListFile - lesen
        for ( auto i = 0u; i < m_strExtendedFilePath.size(); i++ )
        {
            // String inkl. Längenangabe aus Datei lesen.
            auto size = read_cstring( m_strExtendedFilePath[i], fp );
            if ( size == 0 )
            {
                *error = -3;
                res = 0;
                goto error_label;
            }

            // Gesamtanzahl gelesener Zeichen halten.
            extendedFilePathSize += size;
        }
    }
    else
    {
        // Diese Dateiangaben muessen noch nicht verfuegbar sein.
        // Es koennen Dateien ab DatafoxStudioIV - Version 04.01.01.10 geladen werden!
    }

    if ( IsHeader() )
    {
        // CRC der Dateibereiche pruefen
        if ( m_file_header.crcEep && m_file_header.crcEep != GetCrcEep() )
        {
            *error = -10;
            res = 0;
            goto error_label;
        }
        m_file_header.crcEep = 0;

        if ( m_file_header.crcEepExtended && m_file_header.crcEepExtended != GetCrcEepExtended() )
        {
            *error = -11;
            res = 0;
            goto error_label;
        }
        m_file_header.crcEepExtended = 0;

        if ( m_file_header.crcFlash && m_file_header.crcFlash != GetCrcFlash() )
        {
            *error = -12;
            res = 0;
            goto error_label;
        }
        m_file_header.crcFlash = 0;

        if ( m_file_header.crcListFile && m_file_header.crcListFile != GetCrcListFile() )
        {
            *error = -13;
            res = 0;
            goto error_label;
        }
        m_file_header.crcListFile = 0;

        if ( m_file_header.crcConfig && m_file_header.crcConfig != GetCrcConfig() )
        {
            *error = -14;
            res = 0;
            goto error_label;
        }
        m_file_header.crcConfig = 0;

        if ( m_file_header.crcFlashExtended && m_file_header.crcFlashExtended != GetCrcFlashExtended() )
        {
            *error = -15;
            res = 0;
            goto error_label;
        }
        m_file_header.crcFlashExtended = 0;

        if ( m_file_header.crcListEntranceFile && m_file_header.crcListEntranceFile != GetCrcListEntranceFile() )
        {
            *error = -16;
            res = 0;
            goto error_label;
        }
        m_file_header.crcListEntranceFile = 0;

        if ( m_file_header.crcListEntrance2File && m_file_header.crcListEntrance2File != GetCrcListEntrance2File() )
        {
            *error = -17;
            res = 0;
            goto error_label;
        }
        m_file_header.crcListEntrance2File = 0;

        if ( m_file_header.crcListExtendedFile && m_file_header.crcListExtendedFile != GetCrcListExtendedFile() )
        {
            *error = -18;
            res = 0;
            goto error_label;
        }
        m_file_header.crcListExtendedFile = 0;
    }

	m_extended_header = *(StructExtendedFlash*)m_flashExtended;

    // Eintragen in der Extended - Struktur
    if ( m_extended_header.offsetExtendedStruct )
    {
        // Die Werte fuer Dateinamen und Zeitstempel der Modifizierung eintragen.
        // Dateinamen extrahieren
        char *fName = (char*)&file_path[file_path.size()-1];
        while ( fName > &file_path[0] )
        {
            // Maskierte Leerzeichen (unix)
            if ( *(fName - 1) == '\\' && *fName == ' ' )
            {

            }
            else if( *(fName - 1) == '\\' )
            {
                // Pfadbestandteil vor Dateinamen (windows)
                break;
            }
            else if ( *(fName - 1) == '/' )
            {
                // Pfadbestandteil vor Dateinamen (unix)
                break;
            } 

            fName--;
        }

        char mTimeStamp[20]; // 2009-05-19 09:16:22
        struct tm *ptm = localtime( &st.st_mtime );
        sprintf( mTimeStamp, "%02i%02i-%02i-%02i %02i:%02i:%02i", 
            (ptm->tm_year + 1900) / 100, (ptm->tm_year + 1900) % 100, ptm->tm_mon + 1, ptm->tm_mday,
            ptm->tm_hour, ptm->tm_min, ptm->tm_sec );

        StructExtended *extended = (StructExtended*)&m_flashExtended[m_extended_header.offsetExtendedStruct];
        strncpy( extended->fileName, fName, sizeof(extended->fileName) );
        // Nullterminierung sicherstellen
        extended->fileName[sizeof(extended->fileName) - 1] = 0;
        strcpy( extended->fileTimeStamp, mTimeStamp );
    }

error_label:
    if ( fp )
    {
        fclose( fp );
    }

    return res;
}

int FileStudio::save( const std::string& file_path, int *error )
{
    int res = 1;
    unsigned int sizeSetup;

    *error = 0;
    FILE *fp = fopen( file_path.c_str(), "wb" );
    if ( fp == NULL )
    {
        return 0;
    }

    // Groesse der Setupdaten eintragen
    sizeSetup = GetFlashSetupSize();
    m_flash_header.FLASH_SIZE_SUS = endian_reverse( sizeSetup );

    // Ermittlung der Groesse des zu schreibenden Speicherblocks fuer Flash.
    if ( sizeSetup > FLASH_SIZE_SU )
    {
        m_flashSizeSU2 = FLASH_SIZE_SU2;
        m_flash_header.FILE_VERSION = 0xDF02;
    }
    else
    {
        m_flashSizeSU2 = 0;
        m_flash_header.FILE_VERSION = 0xDF01;
    }

    // EEPROM - schreiben
    if ( fwrite( m_eep_mem.data(), sizeof(unsigned char), m_eep_mem.size(), fp ) != m_eep_mem.size() )
    {
        res = 0;
        goto error_label;
    }

    // Headerdaten erstellen, ggf. Anpassungen durchfuehren.
    if ( !buildHeader() )
    {
        res = 0;
        goto error_label;
    }

    // Dateiheaderbereich schreiben
    if ( fwrite( m_file_header_mem.data(), sizeof(unsigned char), m_file_header_mem.size(), fp ) != m_file_header_mem.size() )
    {
        res = 0;
        goto error_label;
    }

    // Flashbereich - schreiben
    if ( fwrite( m_flash_mem.data(), sizeof(unsigned char), FLASH_SIZE_SU + m_flashSizeSU2, fp ) != (FLASH_SIZE_SU + m_flashSizeSU2) )
    {
        res = 0;
        goto error_label;
    }

    // ListFile - schreiben
    for ( auto i = 0u; i < m_listFileName.size(); i++ )
    {
        auto size = write_cstring( m_listFileName[i], fp );
        if ( size == 0 )
        {
            res = 0;
            goto error_label;
        }
    }

    // Konfiguration - schreiben
    if ( fwrite( m_config_mem.data(), sizeof(unsigned char), m_config_mem.size(), fp ) != m_config_mem.size() )
    {
        res = 0;
        goto error_label;
    }

    // Erweiterter Speicher - schreiben
    if ( fwrite( m_extended_mem.data(), sizeof(unsigned char), m_extended_mem.size(), fp ) != m_extended_mem.size() )
    {
        res = 0;
        goto error_label;
    }

    // Zutrittskontrolle - ListFile - schreiben
    for ( auto i = 0u; i < m_listEntranceFileName.size(); i++ )
    {
        auto size = write_cstring( m_listEntranceFileName[i], fp );
        if ( size == 0 )
        {
            res = 0;
            goto error_label;
        }
    }
    // Zutrittskontrolle - ListFile - schreiben
    for ( auto i = 0u; i < m_listEntrance2FileName.size(); i++ )
    {
        auto size = write_cstring( m_listEntrance2FileName[i], fp );
        if ( size == 0 )
        {
            res = 0;
            goto error_label;
        }
    }

    // Zusatzpfadangaben - schreiben
    for ( auto i = 0u; i < m_strExtendedFilePath.size(); i++ )
    {
        auto size = write_cstring( m_strExtendedFilePath[i], fp );
        if ( size == 0 )
        {
            res = 0;
            goto error_label;
        }
    }

error_label:
    if ( fp )
    {
        fclose( fp );
    }

    return res;
}

//########################
// CRC der Dateibereiche
//########################

unsigned int FileStudio::get_crc_cstring( const std::string& s, unsigned int crc, bool closure )
{
    // Nach MFC - CString Konvention Berechnung durchführen.
    if ( s.size() > 0xFF )
    {
        unsigned char len[3];
        len[0] = 0xFF;
        *(unsigned short *)&len[1] = (unsigned short)s.size();
        crc = GetCRC32( len, sizeof( len ), crc, false );
    }
    else
    {
        unsigned char len = (unsigned char)s.size();
        crc = GetCRC32( &len, sizeof( len ), crc, false );
    }

    // String rausschreiben
    return GetCRC32( (const unsigned char *)s.data(), s.size(), crc, closure );
}

///< CRC ueber Eep-Bereich.
unsigned int FileStudio::GetCrcEep()
{
    return GetCRC32( m_eep_mem.data(), m_eep_mem.size() );
}

///< CRC ueber EepExtended-Bereich.
unsigned int FileStudio::GetCrcEepExtended()
{
    return GetCRC32( &m_file_header_mem[m_file_header.bugfixSize], m_file_header_mem.size() - m_file_header.bugfixSize );
}

///< CRC ueber Flash-Bereich.
unsigned int FileStudio::GetCrcFlash()
{
    // Groesse der Setupdaten
    unsigned int sizeSetup = GetFlashSetupSize();

    // Ermittlung der Groesse des zu beachtenden Speicherblocks fuer Flash.
    if ( sizeSetup > FLASH_SIZE_SU )
        m_flashSizeSU2 = FLASH_SIZE_SU2;

    return GetCRC32( m_flash_mem.data(), FLASH_SIZE_SU + m_flashSizeSU2 );
}

///< CRC ueber die Listendateiangaben.
unsigned int FileStudio::GetCrcListFile()
{
    unsigned int crc = 0;
    for ( auto i = 0u; i < m_listFileName.size(); i++ )
    {
        // CRC über die einzelnen Strings, so wie sie in die Datei geschrieben würden, berechnen.
        crc = get_crc_cstring( m_listFileName[i], crc, (i == (m_listFileName.size() - 1)) );
    }

    return crc;
}

///< CRC ueber die Config-Struktur.
unsigned int FileStudio::GetCrcConfig()
{
    return GetCRC32( m_config_mem.data(), m_config_mem.size() );
}

///< CRC ueber FlashExtended-Bereich.
unsigned int FileStudio::GetCrcFlashExtended()
{
    return GetCRC32( m_extended_mem.data(), m_extended_mem.size() );
}

///< CRC ueber die Listendateiangaben der ZK1.
unsigned int FileStudio::GetCrcListEntranceFile()
{
    unsigned int crc = 0;
    for ( auto i = 0u; i < m_listEntranceFileName.size(); i++ )
    {
        // CRC über die einzelnen Strings, so wie sie in die Datei geschrieben würden, berechnen.
        crc = get_crc_cstring( m_listEntranceFileName[i], crc, (i == (m_listEntranceFileName.size() - 1)) );
    }

    return crc;
}

///< CRC ueber die Listendateiangaben der ZK2.
unsigned int FileStudio::GetCrcListEntrance2File()
{
    unsigned int crc = 0;
    for ( auto i = 0u; i < m_listEntrance2FileName.size(); i++ )
    {
        // CRC über die einzelnen Strings, so wie sie in die Datei geschrieben würden, berechnen.
        crc = get_crc_cstring( m_listEntrance2FileName[i], crc, (i == (m_listEntrance2FileName.size() - 1)) );
    }

    return crc;
}

///< CRC ueber die Zusaetzlichen Listendateiangaben.
unsigned int FileStudio::GetCrcListExtendedFile()
{
    unsigned int crc = 0;
    for ( auto i = 0u; i < m_strExtendedFilePath.size(); i++ )
    {
        // CRC über die einzelnen Strings, so wie sie in die Datei geschrieben würden, berechnen.
        crc = get_crc_cstring( m_strExtendedFilePath[i], crc, (i == (m_strExtendedFilePath.size() - 1)) );
    }

    return crc;
}

// Die zu verwendende Version der Datei
int FileStudio::NeededVersion()
{
    if ( GetFlashSetupSize() > FLASH_SIZE_SU )
        return 2;

    return 1;
}

unsigned short FileStudio::endian_reverse( unsigned short val )
{
    return ((val & 0xFF) << 8)
        | ((val >> 8) & 0xFF);
}

unsigned int FileStudio::endian_reverse( unsigned int val )
{
    return ((val & 0xFF) << 24)
        | ((val & 0xFF00) << 8)
        | ((val >> 8) & 0xFF00)
        | ((val >> 24) & 0xFF);
}

// Groesse der Setupdaten im Flash ermitteln
unsigned int FileStudio::GetFlashSetupSize()
{
    // ##############################################################
    // Pointer sind direkte Adressen die im AEIII+:
    //   1. Im Motorolla - Format vorliegen.
    //   2. Mit einem Basisoffset -FLASH_KONFIG_START- berechnet sind.

    // Pointer auf ListenBeschreibungen ermitteln
    auto pLB = endian_reverse( m_flash_header.LIST_STRUCT );
    pLB -= FLASH_KONFIG_START;

    // Pointerarray der Listenbeschreibungen durchschreiten
    int i;
    for ( i = 0; i < MAX_DATENSATZ_BESCHREIBUNG; i++ )
    {
        // Nullpointer suchen fuer Listenanzahl
        if ( *(unsigned int*)&m_flash[pLB + i * sizeof(unsigned int)] == 0 )
            break;
    }

    // Liste gefunden?
    if ( i > 0 )
    {
        // Pointer auf letzte definierte Liste
        pLB = endian_reverse( *(unsigned int*)&m_flash[pLB + (i - 1) * sizeof(unsigned int)] );
        pLB -= FLASH_KONFIG_START;

        StructDatensatzBeschreibung *pLBStruct = (StructDatensatzBeschreibung*)&m_flash[pLB];

        // Pointer auf Feldarray der Liste
        unsigned int pFeldArray = endian_reverse( pLBStruct->P );
        pFeldArray -= FLASH_KONFIG_START;

        // Felder vorhanden?
        if ( pLBStruct->Felder )
        {
            // Pointer auf letztes Feld der Listenbeschreibung ermitteln.
            unsigned int pFeld = endian_reverse( *(unsigned int*)&m_flash[pFeldArray + (pLBStruct->Felder - 1) * sizeof(unsigned int)] );
            pFeld -= FLASH_KONFIG_START;

            // Feldadresse + eigene Groesse bildet FlashSetupSize
            return pFeld + sizeof(StructDatensatzFeld);
        }


        // Keine Felder definiert!
        return pFeldArray;
    }

    // Keine Listenbeschreibung definiert!
    // Die Anzahl fuer Listenbeschreibungspointer muss mitberechnet werden,
    // um bei einer Uebertragung diese zu loeschen.
    return (pLB + MAX_DATENSATZ_BESCHREIBUNG * sizeof(unsigned int));
}

// Groesse des Erweiterten Setupbereichs ermitteln
unsigned int FileStudio::GetExtendedFlashSetupSize()
{
    // Erweiterte Struktur vorhanden?
    if ( m_extended_header.check != EXTENDED_CHECK )
        return 0;

    return m_extended_header.offsetFreeScope;
}

// Groesse der Listen- oder Datensatzbeschreibungen im Flash ermitteln
// \param lists[in] Gibt an ob die Groesse der Listen- oder Datensatzbeschreibungen ermittelt werden sollen.
unsigned int FileStudio::GetTablesSize( bool lists )
{
    // ##############################################################
    // Pointer sind direkte Adressen die im AEIII+:
    //   1. Im Motorolla - Format vorliegen.
    //   2. Mit einem Basisoffset -FLASH_KONFIG_START- berechnet sind.

    // Pointer auf Listen- oder DatensatzBeschreibungen ermitteln
    unsigned int pTables = endian_reverse( (lists) ? m_flash_header.LIST_STRUCT : m_flash_header.DAT_STRUCT );
    pTables -= FLASH_KONFIG_START;

    // Pointerarray der Listenbeschreibungen durchschreiten
    int i;
    for ( i = 0; i < MAX_DATENSATZ_BESCHREIBUNG; i++ )
    {
        // Nullpointer suchen fuer letzte Tabelle
        if ( *(unsigned int*)&m_flash[pTables + i * sizeof(unsigned int)] == 0 )
            break;
    }

    // Tabelle gefunden?
    if ( i > 0 )
    {
        // Pointer auf die letzte definierte Tabelle
        pTables = endian_reverse( *(unsigned int*)(&m_flash[pTables + (i - 1) * sizeof(unsigned int)]) );
        pTables -= FLASH_KONFIG_START;

        StructDatensatzBeschreibung *pTableStruct = (StructDatensatzBeschreibung*)&m_flash[pTables];

        // Pointer auf Feldarray der Tabelle
        unsigned int pFeldArray = endian_reverse( pTableStruct->P );
        pFeldArray -= FLASH_KONFIG_START;

        // Felder vorhanden?
        if ( pTableStruct->Felder )
        {
            // Pointer auf letztes Feld der Listenbeschreibung ermitteln.
            unsigned int pFeld = endian_reverse( *(unsigned int*)&m_flash[pFeldArray + (pTableStruct->Felder - 1) * sizeof(unsigned int)] );
            pFeld -= FLASH_KONFIG_START;

            // Feldadresse + eigene Groesse bilden Endadresse
            return ((pFeld + sizeof(StructDatensatzFeld)) - pTables);
        }


        // Keine Felder definiert!
        return (pFeldArray - pTables);
    }

    // Keine Tabellen definiert!
    return 0;
}

// MD5 ueber den Dateiheader berechnen.
char *FileStudio::GetMD5Header()
{
    if ( !IsHeader() )
    {
        return 0;
    }

    return MD5Bin( reinterpret_cast<unsigned char*>( &m_file_header ), m_file_header.size );
}

// MD5 ueber Listen- oder Datensatzbeschreibungen im Flash berechnen.
// \param lists[in] Gibt an ob der MD5 der Listen- oder Datensatzbeschreibungen ermittelt werden soll.
char *FileStudio::GetMD5Tables( bool lists )
{
    // ##############################################################
    // Pointer sind direkte Adressen die im AEIII+:
    //   1. Im Motorolla - Format vorliegen.
    //   2. Mit einem Basisoffset -FLASH_KONFIG_START- berechnet sind.

    // Pointer auf Listen- oder DatensatzBeschreibungen ermitteln
    unsigned int pTables = endian_reverse( (lists) ? m_flash_header.LIST_STRUCT : m_flash_header.DAT_STRUCT );;
    pTables -= FLASH_KONFIG_START;

    // Pointerarray der Listenbeschreibungen durchschreiten
    int size = GetTablesSize( lists );

    // Tabelle gefunden?
    if ( size > 0 )
    {
        return MD5Bin( &m_flash[pTables], size );
    }

    // Keine Tabellen definiert!
    return 0;
}

int FileStudio::ValidFirmwareVersion( SoftwarebuildVersion &fwVersion )
{
    if ( IsHeader() )
    {
        // Pruefung auf 04.01.xx
        if ( fwVersion >= "04.01.01.00" && fwVersion < "04.02.00.00" )
        {
            // Unsupported der Firmware 04.01.xx.yy pruefen.
            if ( is_invalid_header_version( m_file_header.minFirmware1Version ) )
                return -1;

            // Mindestversion der Firmware 04.01.xx.yy pruefen.
            if ( m_file_header.minFirmware1Version )
            {
                SoftwarebuildVersion minVersion( m_file_header.minFirmware1Version );
                if ( minVersion > fwVersion )
                    return -2;
            }
            // Sollteversion des der Firmware 04.01.xx.yy pruefen?
            if ( !(m_file_header.flags & hfiv_firmware1) && m_file_header.shouldFirmware1Version )
            {
                SoftwarebuildVersion shouldVersion( m_file_header.shouldFirmware1Version );
                if ( shouldVersion > fwVersion )
                    return -3;
            }
        }
        else if ( fwVersion >= "04.02.00.00" && fwVersion < "04.03.00.00" )
        {

            // Unsupported der Firmware 04.02.xx.yy pruefen.
            if ( is_invalid_header_version( m_file_header.minFirmware2Version ) )
                return -4;

            // Mindestversion der der Firmware 04.02.xx.yy pruefen.
            if ( m_file_header.minFirmware2Version )
            {
                SoftwarebuildVersion minVersion( m_file_header.minFirmware2Version );
                if ( minVersion > fwVersion )
                    return -5;
            }
            // Sollteversion der der Firmware 04.02.xx.yy pruefen?
            if ( !(m_file_header.flags & hfiv_firmware2) && m_file_header.shouldFirmware2Version )
            {
                SoftwarebuildVersion shouldVersion( m_file_header.shouldFirmware2Version );
                if ( shouldVersion > fwVersion )
                    return -6;
            }
        }
        else if ( fwVersion >= "04.03.00.00" && fwVersion < "04.04.00.00" )
        {
            // Unsupported der Firmware 04.03.xx.yy pruefen.
            if ( is_invalid_header_version( m_file_header.minFirmware3Version ) )
                return -8;

            // Mindestversion der der Firmware 04.03.xx.yy pruefen.
            if ( m_file_header.minFirmware3Version )
            {
                SoftwarebuildVersion minVersion( m_file_header.minFirmware3Version );
                if ( minVersion > fwVersion )
                    return -9;
            }
            // Sollteversion der der Firmware 04.03.xx.yy pruefen?
            if ( !(m_file_header.flags & hfiv_firmware3) && m_file_header.shouldFirmware3Version )
            {
                SoftwarebuildVersion shouldVersion( m_file_header.shouldFirmware3Version );
                if ( shouldVersion > fwVersion )
                    return -10;
            }
        }
        else
        {
            // Die Firmwareversion wird nicht unterstützt.
            return -7;
        }

        // Firmwareversion wird unterstuetzt.
        return 1;
    }

    // Es liegt kein Dateiheader vor!
    return 0;
}
