#include "channel/channel_4.h"
using namespace Datafox::Communication;

#include <stdio.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <array>
#include <bitset>
#include <boost/assign/list_of.hpp>
#include <boost/algorithm/string.hpp>
#include <cstring>
#include <algorithm>

#if defined(_MSC_VER) || defined(__MINGW32__)
#define stat _stat
#endif

// Versionspruefung
#include "common/dfcom_utils.h"
#include "DatafoxLibraryIV/include/library_version.h"

#include "file/file_device_text.h"
#include "file/file_touch_keys.h"
#include "file/file_display_colors.h"
#include "file/file_wlan_config.h"

#if defined(_MSC_VER) || defined(__MINGW32__)
#define snprintf	    _snprintf
#endif

// ##################################################################
/**
* @brief Initialisiert die Membervariablen und 
*        benoetigte Umgebung.
*
* @param id Kanalnummer die fuer Log usw. verwendet wird. Sie identifiziert 
*           die Klasse.
*/
Channel4::Channel4( int connection_id, unsigned char device_address )
    : Module_table( connection_id, device_address ) 
{
    trace( "%s(%i):INFO(3): %s [ %i, %i ]\n", __FILE__, __LINE__, __func__, connection_id, device_address );

    /*	//#####################################################
    // SOCKETS
    // Benutzung der Socket - Funktionalitaet initialisieren
    WSADATA WindowsData;
    WORD    Version;

    Version = MAKEWORD(2, 2);
    WSAStartup(Version, &WindowsData);
    //#####################################################
    */

    // Membervariablen setzen
    set_communication_thread_stop_signal( false );
    set_active_mode_state( ams_unused );
    set_communication_thread_error( IV_ERROR_SUCCESS );

    // Sockets initialisieren
    m_listen_socket  = nullptr;

    // Callback fuer Nachricht auf Datensaetze
    m_call_back_notify_record_available = nullptr;

    // Melden von Datensaetzen aktiviert?
    m_notify = 0;

    // Hardwareinfo aufsetzen
    memset(&m_hardware_info, 0, sizeof(StructHardwareInfo));
    m_hardware_info.typeDevice           = IV_VALUE_TYPE_DEVICE_PZE;
    m_hardware_info.typeBoard            = IV_VALUE_TYPE_BOARD_2_1;
    m_hardware_info.componentRS485_0     = 0;
    m_hardware_info.typeTCPIP            = IV_VALUE_TYPE_TCPIP_XPORT;
    m_hardware_info.typeModemMC          = IV_VALUE_TYPE_MODEM_MC_NONE;
    m_hardware_info.componentRS485_1     = 0;
    m_hardware_info.typeTransponder      = IV_VALUE_TYPE_TRANSPONDER_TSR32;
    m_hardware_info.typeDisplay          = IV_VALUE_TYPE_DISPLAY_SMALL;
    m_hardware_info.typeKeypad           = IV_VALUE_TYPE_KEYPAD_9;
    m_hardware_info.typeMMC              = IV_VALUE_TYPE_MMC_NONE;
    m_hardware_info.typeGPS              = IV_VALUE_TYPE_GPS_NONE;
    m_hardware_info.componentUSB         = 0;
    m_hardware_info.typeRadio            = IV_VALUE_TYPE_RADIO_NONE;
    m_hardware_info.typeDCF77            = IV_VALUE_TYPE_DCF77_NONE;
    m_hardware_info.componentInAnalog    = IV_VALUE_COMPONENT_IN_ANALOG_NONE;
    m_hardware_info.componentInDigital   = IV_VALUE_COMPONENT_IN_DIGITAL_1;
    m_hardware_info.componentOutDigital  = IV_VALUE_COMPONENT_OUT_DIGITAL_1;
    m_hardware_info.typeFingerprint      = IV_VALUE_TYPE_FINGERPRINT_NONE;
    m_hardware_info.typeExternalReader   = IV_VALUE_TYPE_EXTERNAL_READER_UNDEFINED;
    m_hardware_info.optionIOModule       = 0;
    m_hardware_info.optionTBAdaption     = 0;

    // Textdaten als Default uebertragen
    m_load_as_default_language = 0;
}

// ##################################################################
/**
* @brief Destruktor bereinigt die verwendete Umgebung.
*/
Channel4::~Channel4()
{
    // Beendigung sicherstellen, auch wenn er bereits beendet wurde.
    if ( m_communication_thread.joinable() )
    {
        set_communication_thread_stop_signal();
        m_communication_thread.join();
    }

    //#####################################################
    // SOCKETS

    // Socket - Funktionalitaet exit
    //	WSACleanup();
    //#####################################################
}

// ##################################################################
/**
* @brief Entgegennehmen der Verbindung und starten eines Verarbeitungsthreads.
*/
int Channel4::connect( SOCKET *listen_socket, int connection_id, unsigned char device_address, int timeout, int alive_timeout, int info_flags ) 
{
    trace( "%s(%i):INFO(3): %s [ 0x08X, %i, %i, %i, %i ]\n", __FILE__, __LINE__, __func__, listen_socket, connection_id, device_address, timeout, alive_timeout );

    // Hauptthread starten
    if ( get_active_mode_state() == ams_unused )
    {
        // Kanalobjekt ist nun in Verwendung.
        set_active_mode_state( ams_starting );
        trace( "%s(%i):INFO(1): %s [ setActiveModeState(%i) is_starting ]\n", __FILE__, __LINE__, __func__, device_address );

        // Wenn das Kanalobjekt bereits verwendet wurde, dann einen join ausführen damit das Thread-Objekt ungueltig wird.
        if ( m_communication_thread.joinable() )
        {
            set_communication_thread_stop_signal();
            m_communication_thread.join();
        }

        // Kanalnummer und DeviceID uebernehmen.
        set_connection_id( connection_id );
        set_device_address_sender( device_address );

        // Listensocket von dem die Verbindungsanforderung kommt wird 
        // ueber lokale Membervariable uebergeben.
        m_listen_socket = listen_socket;
        if (timeout < 3000)
        {
            trace( "%s(%i):WARNING(1): %s(%i) [ Timeout von %i auf mindestens 3000ms gesetzt. ]\n", __FILE__, __LINE__, __func__, device_address, timeout );
            timeout = 3000;
        }
        else if ( timeout > 60000 )
        {
            trace( "%s(%i):WARNING(1): %s(%i) [ Timeout von %i auf maximal 60000ms gesetzt. ]\n", __FILE__, __LINE__, __func__, device_address, timeout );
            timeout = 60000;
        }
        get_connection().set_read_timeout( timeout );
        m_alive_timeout = alive_timeout;
        m_info_flags = info_flags;

        // Kommunikationsworker
        set_communication_thread_start_signal();
        m_communication_thread = std::thread( [=] { get_communication_thread(); } );
        // Thread umschalten
        std::this_thread::sleep_for( std::chrono::milliseconds( 10 ) );

        int loops = 0;
        // Start des Threads pruefen.
        while ( get_communication_thread_start_signal() == false )
        {
            if ( loops == 30 )
            {
                // Nach 3 Sekunden sollte der Threadstart und die Verbindungsannahme beendet sein.
                if ( m_communication_thread.joinable() )
                {
                    set_communication_thread_stop_signal();
                    m_communication_thread.join();
                }

                // Fehler beim Starten des Threads da immer noch im Zustand is_starting?
                if ( get_active_mode_state() == ams_starting )
                    set_active_mode_state( ams_unused );

                break;
            }

            std::this_thread::sleep_for( std::chrono::milliseconds( 100 ) );
            loops++;
        }
    }

    return (get_active_mode_state() != ams_unused);
}

// ##################################################################
/**
* @brief Gewünschte Info über die Active Verbindung erstellen.
*/
bool Channel4::get_active_connection_info( char *info, int* info_length )
{
    trace("%s(%i):INFO(3): %s [ 0x%p, 0x%p ]\n", __FILE__, __LINE__, __func__, info, info_length );

    if ( get_active_mode_state() == ams_active )
    {
        char info_text[255];
        int infoLength = 0;
        int infoLengthMax = 0;
        memset(info_text, 0, sizeof(info_text));
        std::bitset<32> flags( m_info_flags );

        for ( int paramNumber = 0; infoLength != -1 && infoLength < *info_length; ++paramNumber )
        {
            infoLengthMax = infoLength;

            if ( paramNumber == 0 ) // Gerätetyp
            {
                infoLength = snprintf( info_text, sizeof(info_text), "%u\r", m_hardware_info.typeDevice );
            }
            else if ( paramNumber == 1 ) // Seriennummer
            {
                infoLength += snprintf( &info_text[infoLength], sizeof(info_text) - infoLength, "%u\r", (unsigned int)get_property( "device.serial", -1 ) );
            }
            else if ( paramNumber == 2 ) // Firmwareversion optional
            {
                // Wenn keine Flags gesetzt wurden, dann Ende.
                if ( !flags.count() )
                    break;

                if ( flags.test( 0 ) )
                    infoLength += snprintf( &info_text[infoLength], sizeof(info_text) - infoLength, "%s\r", m_version_firmware.c_str() );
                else
                    infoLength += snprintf( &info_text[infoLength], sizeof(info_text) - infoLength, "\r" );
            }
            else if ( paramNumber == 3 ) // TCPIP Adresse und Port optional
            {
                // Wenn keine weiteren Flags gesetzt wurden, dann Ende.
                if ( !flags.test( 1 ) && !flags.test( 2 ) )
                    break;

                if ( flags.test( 1 ) )
                    infoLength += snprintf( &info_text[infoLength], sizeof(info_text) - infoLength, "%s:%i\r", m_tcpip_info.ip.c_str(), m_tcpip_info.port );
                else
                    infoLength += snprintf( &info_text[infoLength], sizeof(info_text) - infoLength, "\r" );
            }
            else if ( paramNumber == 4 ) // Mac optional
            {
                if ( flags.test( 2 ) )
                    infoLength += snprintf( &info_text[infoLength], sizeof(info_text) - infoLength, "%s\r", m_tcpip_info.mac.c_str() );
            }
            else
                break;
        }
        // Pruefen ob die Werte kopiert werden koennen; die Puffergroesse muss passen!
        if ( infoLengthMax < *info_length )
        {
            snprintf( info, infoLengthMax, "%s", info_text );
            info[infoLengthMax] = 0;
            *info_length = infoLengthMax;
            return true;
        }
        else
        {
            *info_length = 0;
        }
    }

    return false;
}

bool Channel4::get_communication_thread_start_signal()
{
    return m_communication_thread_start_signal;
}

void Channel4::set_communication_thread_start_signal( bool signal )
{
    m_communication_thread_start_signal = signal;
}

bool Channel4::get_communication_thread_stop_signal()
{
    return m_communication_thread_stop_signal;
}

void Channel4::set_communication_thread_stop_signal( bool signal )
{
    m_communication_thread_stop_signal = signal;
}

int Channel4::get_communication_thread_error()
{
    return m_communication_thread_error;
}

void Channel4::set_communication_thread_error( int error )
{
    m_communication_thread_error = error;
}

// ##################################################################
/**
* @brief Workerthread fuer die Verarbeitung der Kommunikationsaufgaben.
*/
void Channel4::get_communication_thread()
{
    trace("%s(%i):INFO(3): %s\n", __FILE__, __LINE__, __func__);

#if defined(_MSC_VER)
#ifndef MYCONST
#define MYCONST const
#endif
#else
#define MYCONST
#endif

    // Zustand auf Startet.
    set_communication_thread_stop_signal( false );
    set_communication_thread_error( IV_ERROR_SUCCESS );

    // Hauptschleife ausfuehren?
    bool run = false;

    int res;
    int failCount = 0;
    int resumeCount = 0;

    SOCKADDR_IN sa;
    struct timeval seltime;
    fd_set readfds;

    uint16_t count;
    uint8_t data[IV_PACKET_LENGTH_DATA]; // 21 Byte Paketframe

    int len;

    // Timeout fuer Abfrage auf Daten
    seltime.tv_sec  = 0;
    seltime.tv_usec = 10000;

    // Zustand setzen.
    set_active_mode_state( ams_accepting );
    trace( "%s(%i):INFO(1): %s [ setActiveModeState(%i) is_accepting ]\n", __FILE__, __LINE__, __func__, get_connection_id() );

    // Verbindungsgesuch annehmen.
    len = sizeof(sa);
    get_connection().set_socket( accept( *m_listen_socket, (SOCKADDR *)&sa, (socklen_t*)&len ) );

    // Signal fuer den Erfolgreichen Start und den Annahmeversuch setzen.
    set_communication_thread_start_signal( true );

    if( get_connection().get_socket() == INVALID_SOCKET )
    {
        trace( "%s(%i):ERROR(1): %s [ accept fail (%i) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError() );
        // Fehler setzen.
        set_communication_thread_error( IV_ERROR_FAIL );
    }
    else
    {
        // Verbindungsmeldung ausgeben
        trace( "%s(%i):INFO(1): %s [ (%i) Connected to %s, port %d ]\n", __FILE__, __LINE__, __func__, get_connection_id(), inet_ntoa( sa.sin_addr ), htons( sa.sin_port ) );

        // Mindesttimeout sicherstellen
        if ( get_connection().get_read_timeout() < 3000 )
            get_connection().set_read_timeout( 3000 );

        // Socket zugewiesen
        get_connection().set_type( Connection_base::Connection_type::socket );

        // Verbindungsentgegennahme laeuft.
        set_active_mode_state( ams_detecting );
        trace( "%s(%i):INFO(1): %s [ setActiveModeState(%i) is_detecting ]\n", __FILE__, __LINE__, __func__, get_connection_id() );

        // Grunddaten zur Verbindung abfragen
        for ( failCount = 0; failCount < 4; ++failCount )
        {
            // Paketabarbeitung einleiten
            if ( (res = ping( 254, nullptr, nullptr )) != IV_ERROR_SUCCESS )
            {
                if ( (res == -IV_ERROR_CHANNEL_OPEN_REMOTE) || (failCount > 2) )
                {
                    // Wenn Daten gemeldet werden,
                    // die Anzahl aber 0 ist, so wurde die Verbindung beendet.
                    Module_table::close_connection();

                    trace("%s(%i):INFO(1): %s [ Disconnect (%i) ]\n", __FILE__, __LINE__, __func__, get_connection_id());

                    // Fehler setzen.
                    set_communication_thread_error( IV_ERROR_FAIL );

                    // Abbruch, zu viele Fehlversuche oder Gegenstelle hat geschlossen.
                    break;
                }
            }
            else
            {
                // Atwort auf Ping erhalten. Verbindung OK...
                break;
            }
        }

        if ( get_communication_thread_error() == IV_ERROR_SUCCESS && get_communication_thread_stop_signal() == false )
        {
            // com.                  Praefix fuer Kommunikationsvariablen
            // active                Aktivieren der Aktiven Verbindung
            // notify                notify => Aktivieren der Datensatzmeldung
            // prio                  Prioritaet des Geraetes
            // host                  Hostname des Remoterechners
            // port                  Portnummer des Remoterechners
            // retry                 Anzahl Versuche fuer Verbindungsaufbau
            // timeout               Auszeit zwischen den Versuchsbloecken
            // repeat                Auszeit zur wiederholten Datensatzmeldung, wenn diese nicht abgerufen wurden.
            // alive                 Pinganfrage wenn fuer diese Zeit keine Kommunikation zustande kam.
            // Alle Zeitangaben in Sekunden und 32bit

            std::string val;
            std::string notify;
            std::string status;
            unsigned int serial;

            val = std::to_string( 3 + static_cast<int>( ceil( static_cast<double>( m_alive_timeout ) * 1.05 ) ) ); // 5% fuer Sync. dazugeben
            // Aliveparameter berechnen und setzen.
            // Geraetetyp und Seriennummer als GeraeteID abfragen
            if ((res = write_system_variable(254, "COM.ALIVE", val)) != IV_ERROR_SUCCESS ||
                (res = read_serial_number(254, serial)) != IV_ERROR_SUCCESS ||
                (res = read_system_variable(254, "COM.NOTIFY", notify)) != IV_ERROR_SUCCESS ||
                (res = read_firmware_version(254, m_version_firmware)) != IV_ERROR_SUCCESS ||
                // Der Systemstatus wird nur bei aktivem Trace gelesen. Bei aktivem Produktschutz wird er nicht geliefert.
                (get_connection().get_trace_mode() >= Trace_base::Trace_mode::on && (res = read_system_variable(254, "SYSTEM.STATUS", status)) != IV_ERROR_SUCCESS && res != 28))
            {
                // Verbindung schliessen
                Module_table::close_connection();

                trace("%s(%i):INFO(1): %s [ Disconnect (%i) ]\n", __FILE__, __LINE__, __func__, get_connection_id());

                // Fehler setzen.
                set_communication_thread_error( IV_ERROR_FAIL );
            }
            else 
            {
                // Gelesene Seriennummer setzen.
                set_property( "device.serial", serial );

                // Geraetetyp ermitteln.
                SoftwareVersion versionFirmware(m_version_firmware);
                m_hardware_info.typeDevice = versionFirmware.typeDevice();

                // Melden von Datensaetzen
                try
                {
                    m_notify = std::stoi( notify );
                }
                catch ( ... )
                {
                    m_notify = 0;
                }

                // Bereits ermittelte Info ausgeben
                trace("%s(%i):INFO(1): %s [ getTypeDevice( %i ), getSerial( %u ) ]\n", __FILE__, __LINE__, __func__, m_hardware_info.typeDevice, get_property( "device.serial", -1 ));
                // Systemstatus auswerten
                if ( get_connection().get_trace_mode() >= Trace_base::Trace_mode::on && status.size() )
                {
                    char *token = strtok(&status[0], "\r");
                    while (token != nullptr)
                    {
                        trace("%s(%i):INFO(1): %s [ SystemStatusLine( %s ) ]\n", __FILE__, __LINE__, __func__, token);
                        token = strtok(nullptr, "\r");
                    }
                }

                // Daten fuer das TCPIP-Modul ermitteln
                m_tcpip_info.reset();

                // Informationen zum TCP/IP abrufen
                // Erst nachtraeglich hinzugefuegt, deswegen separat ausfuehren.
                if ((res = read_hardware_module_info(254, "COM.TCPIP", "", m_tcpip_info.data)) != IV_ERROR_SUCCESS)
                {
                    // Nur der Fehler "Majorname unbekannt" ist hier gueltig. Dann wird der Befehl noch nicht unterstuetzt.
                    if ((res & 0xFF) != 1)
                    {
                        // Wenn Daten gemeldet werden,
                        // die Anzahl aber 0 ist, so wurde die Verbindung beendet.
                        Module_table::close_connection();

                        trace("%s(%i):INFO(1): %s [ Disconnect (%i) ]\n", __FILE__, __LINE__, __func__, get_connection_id());

                        // Fehler setzen.
                        set_communication_thread_error( IV_ERROR_FAIL );
                    }
                }
                else
                {
                    // Werte aufschluesseln
                    m_tcpip_info.parse();

                    // Bereits ermittelte Info ausgeben
                    trace("%s(%i):INFO(1): %s [ GetInfoTCPIP( %i ), ModulVersion( %s ), IP( %s ), Port( %i ), MAC( %s ) ]\n", __FILE__, __LINE__, __func__, get_connection_id(), m_tcpip_info.version.c_str(), m_tcpip_info.ip.c_str(), m_tcpip_info.port, m_tcpip_info.mac.c_str());
                }

                // Gewuenschte Zusatzinfos abrufen
                if (m_info_flags & 0x01)
                {

                }
            }

            // Bedingung fuer Start der Hauptschleife erfuellt.
            if ( get_communication_thread_error() == IV_ERROR_SUCCESS && get_communication_thread_stop_signal() == false )
            {
                // Ermittelbare Daten zur Verbindung erfolgreich ermittelt.
                set_active_mode_state( ams_usable );
                trace( "%s(%i):INFO(1): %s [ setActiveModeState(%i) is_usable ]\n", __FILE__, __LINE__, __func__, get_connection_id() );

                // Hauptschleife ausfuehren.
                run = true;
            }
        }
    }

    // Fehlerzaehler neu initialisieren
    failCount = 0;

    // Hauptschleife des Worker
    while ( run )
    {
        // Beendigungssignal abfragen
        if ( get_communication_thread_stop_signal() )
        {
            // Verbindung herunterfahren
            if ( get_connection().get_socket() != INVALID_SOCKET )
            {
                // Befehlsausfuehrung geschuetzt durch Mutex.
                lock_guard_call guard( m_mutex_call );

                // Senden am Socket verhindern
                if ( shutdown( get_connection().get_socket(), 0x01/*SD_SEND*/) == SOCKET_ERROR )
                {
                    trace("%s(%i):ERROR(1): %s [ shutdown fail (%i) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError());
                }
                else 
                {
                    // Alle anstehenden Daten lesen
                    do {

                        FD_ZERO((fd_set FAR *) &readfds);
                        FD_SET(get_connection().get_socket(), (fd_set FAR *) &readfds);

                        if ((res = select(FD_SETSIZE,
                            (fd_set FAR *)&readfds, (fd_set FAR *) 0, (fd_set FAR *) 0,
                            (MYCONST struct timeval FAR *)&seltime)) == SOCKET_ERROR)
                        {
                            trace("%s(%i):ERROR(1): %s [ select fail (%i) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError());
                        }
                        else if (res)
                        {
                            char packet_receive[IV_PACKET_LENGTH_MAX];
                            // Anstehende Daten lesen
                            if ((res = recv(get_connection().get_socket(), packet_receive, sizeof(packet_receive), 0 )) == SOCKET_ERROR)
                            {
                                trace("%s(%i):ERROR(1): %s [ recv fail (%i) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError());
                            }
                            else if (res == 0)
                            {
                                // Disconnect von Gegenstelle
                            }
                        }
                    } while((res > 0) && (res != SOCKET_ERROR));
                }

                // Verbindung schliessen
                Module_table::close_connection();
            }

            // Hauptschleife verlassen
            run = false;
            continue;
        }

        // Eigener Block wegen MutexCall
        {
            // Befehlsausfuehrung geschuetzt durch Mutex.
            lock_guard_call guard( m_mutex_call );

            // Socketdaten initialisieren
            FD_ZERO((fd_set FAR *) &readfds);
            FD_SET(get_connection().get_socket(), (fd_set FAR *) &readfds);

            // Auf eintreffende Daten warten
            if ((res = select(FD_SETSIZE,
                (fd_set FAR *)&readfds, (fd_set FAR *) 0, (fd_set FAR *) 0,
                (MYCONST struct timeval FAR *)&seltime)) != SOCKET_ERROR)
            {
                // Stehen Daten an? (ansonsten Timeout)
                if (res)
                {
                    // Paketabarbeitung einleiten
                    if ((res = call((short)IV_CONTEXT_SLAVE, (short)IV_MODULE_BASE, (short)IV_FUNCTION_BASE, data, count, 1)) != IV_ERROR_SUCCESS)
                    {
                        if ((res == -IV_ERROR_CHANNEL_OPEN_REMOTE) || (failCount > 2))
                        {
                            // Wenn Daten gemeldet werden,
                            // die Anzahl aber 0 ist, so wurde die Verbindung beendet.
                            Module_table::close_connection();

                            trace("%s(%i):INFO(1): %s [ Disconnect ]\n", __FILE__, __LINE__, __func__, get_connection_id(), inet_ntoa(sa.sin_addr), htons(sa.sin_port));

                            // Fehler setzen.
                            set_communication_thread_error( IV_ERROR_FAIL );

                            // Hauptschleife verlassen
                            run = false;
                            continue;
                        }

                        // Fehlerzaehler erhoehen
                        failCount++;
                    }
                    else
                    {
                        // Fehlerzaehler zuruecksetzen
                        failCount = 0;
                    }
                }
            }
            else
            {
                trace("%s(%i):ERROR(1): %s [ select fail (%i) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError());

                // Verbindung schliessen
                Module_table::close_connection();

                trace("%s(%i):INFO(1): %s [ Disconnect ]\n", __FILE__, __LINE__, __func__, get_connection_id(), inet_ntoa(sa.sin_addr), htons(sa.sin_port));

                // Fehler setzen.
                set_communication_thread_error( IV_ERROR_FAIL );

                // Hauptschleife verlassen
                run = false;
                continue;
            }

            // Aliveping senden?
            if ( (get_protocol().get_time_read_packet() + m_alive_timeout) < time(nullptr) )
            {
                // Ping
                if ( (res = ping(254, nullptr, nullptr)) != IV_ERROR_SUCCESS )
                {
                    if ( (res == -IV_ERROR_CHANNEL_OPEN_REMOTE) || (failCount > 2) )
                    {
                        // Wenn Daten gemeldet werden,
                        // die Anzahl aber 0 ist, so wurde die Verbindung beendet.
                        Module_table::close_connection();

                        trace("%s(%i):INFO(1): %s [ Disconnect (%i) ]\n", __FILE__, __LINE__, __func__, get_connection_id());

                        // Fehler setzen.
                        set_communication_thread_error( IV_ERROR_FAIL );

                        // Hauptschleife verlassen
                        run = FALSE;
                        continue;
                    }

                    // Fehlerzaehler erhoehen
                    failCount++;
                }
                else
                {
                    // Fehlerzaehler zuruecksetzen
                    failCount = 0;
                }
                
                trace("%s(%i):INFO(1): %s [ Alive - Ping (%i) %ds ]\n", __FILE__, __LINE__, __func__, get_connection_id(), int(time(nullptr) - get_protocol().get_time_read_packet()));
            }
        }

        // Kurze Ausszeit um die Systemlast zu reduzieren.
        std::this_thread::sleep_for( std::chrono::milliseconds( 10 ) );
    }

    // Eventuellen Eintrag aus Benachrichtungsschlange entfernen.
    if ( m_call_back_notify_record_available )
    {
        CNotifyRecord nr;

        // Meldungen dieses Kanals loeschen
        nr.erase = 1;

        // Aktuelle Kanalnummer eintragen
        nr.channelID = get_connection_id();

        // Meldung aus Warteschlange entfernen.
        m_call_back_notify_record_available(nr);
    }

    // Kanalobjekt wieder frei
    set_active_mode_state( ams_unused );
    trace("%s(%i):INFO(1): %s [ setActiveModeState(%i) is_none ]\n", __FILE__, __LINE__, __func__, get_connection_id());
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*
* @param device_address Anzusprechende Geraetenummer.
*/
int Channel4::call( uint16_t context, uint16_t module, uint16_t func, uint8_t *buffer, uint16_t &count, int context_value )
    // Modulbefehl senden und Antwort zurueckgeben
    //  device_address -> Adresse (Geraetenummer) des Empfaengers
    //  module        -> Nummer des anzusprechenden Modules
    //  func          -> Nummer der Funktion im Module
    //  buf           -> Buffer fuer Funktionsparameter.
    //  count         -> Anzahl der Funktionsparameter in Byte. Rueckgabewert bei Parameterempfang.
    //
    //  return -> <0 -> Kommunikationsfehler
    //             0 -> senden und empfangen und ausfuehren erfolgreich
    //            >0 -> module-/Funktionsfehler
{
    trace("%s(%i):INFO(3): %s [ %u, %u, %u, %p, %u, %d ]\n", __FILE__, __LINE__, __func__, context, module, func, buffer, count, context_value);

    return Module_system::call(context, module, func, buffer, count, context_value);
}

int Channel4::generate_fingerprint_id_list( unsigned char device_address, int flags, unsigned int &current_count )
{
    int res;
    unsigned char val = 0;
    std::string name( "FINGERPRINT.IDLIST" );
    std::string reserved;
    HardwareInfo hardwareInfo;

    if ( (res = read_hardware_info( device_address, hardwareInfo )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    listItemBoard tags;
    // Nur wenn Biokey verbaut ist.
    tags = hardwareInfo.moduleBiokey();
    if ( tags.size() == 0 )
    {
        return -IV_ERROR_NO_MODULE;
    }

    if ( (res = generate_table( device_address, name, reserved )) != IV_ERROR_SUCCESS )
    {
        // Liegen im Modul keine Daten vor?
        if ( res == 133 )
        {
            return -IV_ERROR_NO_DATA_AVAILABLE;
        }
    }

    // Flag fuer anschliessende Ermittlung der Datensatzanzahl gesetzt?
    if ( flags & 0x01 )
    {
        int handle;
        unsigned int unsortedCount = 0;
        unsigned int deletedCount = 0;
        current_count = 0;

        // Erstellte Tabelle oeffnen.
        if ( (res = open_table( device_address, name, 0, handle, reserved )) != IV_ERROR_SUCCESS )
        {
            return res;
        }

        // Anzahl enthaltener Datensaetze abrufen.
        if ( (res = read_table_row_count( device_address, handle, current_count, unsortedCount, deletedCount )) != IV_ERROR_SUCCESS )
        {
            return res;
        }

        // Tabelle schliessen.
        if ( (res = close_table( device_address, handle )) != IV_ERROR_SUCCESS )
        {
            return res;
        }
    }

    return res;
}

// Backup der Fingerprints erstellen
int Channel4::fingerprintBackup( unsigned char device_address, const std::string &file_path, int flags )
{
    (void)flags;
    int res = IV_ERROR_SUCCESS;
    std::string name("FINGERPRINT.IDLIST");
    std::string reserved;
    unsigned char val = 0;
    HardwareInfo hardwareInfo;

    if ( (res = read_hardware_info( device_address, hardwareInfo )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    listItemBoard tags;
    // Nur wenn Biokey verbaut ist.
    tags = hardwareInfo.moduleBiokey();
    if ( tags.size() == 0 )
        return -IV_ERROR_NO_MODULE;

    // Tabelle mit Fingerprintdaten erzeugen
    if ( (res = generate_table(device_address, name, reserved)) != IV_ERROR_SUCCESS )
    {
        return res;
    }
    int hTable;
    // Tabelle mit Fingerprintdaten oeffnen
    if ((res = open_table(device_address, name, 0, hTable, reserved)) != IV_ERROR_SUCCESS)
    {
        return res;
    }
    unsigned int recordCount = 0;
    unsigned int unsortedCount = 0;
    unsigned int deletedCount = 0;
    // Pruefen ob Datensaetze vorliegen
    if ((res = read_table_row_count(device_address, hTable, recordCount, unsortedCount, deletedCount)) != IV_ERROR_SUCCESS)
    {
        // Modul oder Funktionsfehler?
        if (res > 0) close_table(device_address, hTable);
        return res;
    }
    if (recordCount == 0)
    {
        // Es liegen keine Daten vor
        return -IV_ERROR_NO_DATA_AVAILABLE;
    }
    FILE *fp;
    StructFingertemplateId Id;
    unsigned char type = 3; // Saturn01 zuerst verwenden.
    char data[IV_PACKET_LENGTH_DATA];
    int size;
    unsigned short crc;

    // Datei oeffnen
    if ((fp = fopen(file_path.c_str(), "wb")) == nullptr)
    {
        close_table(device_address, hTable);
        return -IV_ERROR_FILE_OPEN;
    }

    // Kopfstruktur vorbereiten
    StructFTFileHeader structHeader;
    memset(&structHeader, 0, sizeof(StructFTFileHeader));
    // Kopfdaten reservieren
    if (fwrite(&structHeader, 1, sizeof(StructFTFileHeader), fp) < sizeof(StructFTFileHeader))
    {
        // Daten wurden nicht geschrieben
        close_table(device_address, hTable);
        fclose(fp);
        return -IV_ERROR_FILE_WRITE;
    }
    // Kopfdaten aufsetzen. Die Datei ist erst gueltig, wenn die Kopfdaten geschrieben wurden.
    structHeader.Id = 0xDFDF;
    structHeader.offs = sizeof(StructFTFileHeader);
    structHeader.count = recordCount;

    // Fingertemplates abrufen.
    do
    {
        // Datenspeicher loeschen
        size = sizeof(data);
        memset(data, 0, sizeof(data));

        // Datensatz aus Liste abrufen
        if ((res = read_current_table_row_data_ext(device_address, hTable, 0, 0, data, (unsigned short&)size)) != IV_ERROR_SUCCESS)
        {
            // Modul oder Funktionsfehler? Dann ist die Kommunikation noch moeglich.
            if (res > 0) close_table(device_address, hTable);
            fclose(fp);
            return res;
        }
        // bekannte Daten in struktur uebernehmen
        memcpy((unsigned char*)&Id, data, sizeof(StructFingertemplateId));

        // Datenspeicher loeschen
        size = sizeof(data);
        memset(data, 0, sizeof(data));

        // Das dazugehoerige Fingertemplate abfragen
        while ((res = read_fingerprint_record(device_address, type, (unsigned int)atof(Id.pid), atoi(Id.fid), reinterpret_cast<unsigned char*>(data), size)) != IV_ERROR_SUCCESS)
        {
            // Wenn das Templateformat unbekannt oder nicht dem verwendeten entspricht, dann ggf. noch anderes versuchen.
            // Funktionsfehler 1 und 3
            if ((res == 1 || res == 3) && type != 0)
            {
                type--; // Nächstes Format prüfen.
                continue;
            }

            // Modul oder Funktionsfehler? Dann ist die Kommunikation noch moeglich.
            if (res > 0) close_table(device_address, hTable);
            fclose(fp);
            return res;
        }

        // Groesse ist abhaengig von verwendetem Templatetyp.
        switch ( type )
        {
        case 0:
            size = sizeof(StructFingertemplateDIN);
            break;

        case 1:
            size = sizeof(StructFingertemplateIdencom);
            break;

        case 2:
            size = sizeof(StructFingertemplateIdencomStandard);
            break;

        case 3:
            size = sizeof(StructFingertemplateSaturn01);
            break;

        default:
            size = 0;
            break;
        }
        // CRC16 ueber das Fingertemplate erstellen
        crc = GetCRC16(reinterpret_cast<unsigned char*>(data), size);
        // Datensatz in die Datei wegschreiben
        if (fwrite(data, 1, (size_t)size, fp) < (size_t)size)
        {
            // Daten wurden nicht geschrieben
            close_table(device_address, hTable);
            fclose(fp);
            return -IV_ERROR_FILE_WRITE;
        }
        // CRC schreiben
        if (fwrite(&crc, 1, sizeof(unsigned short), fp) < sizeof(unsigned short))
        {
            // Daten wurden nicht geschrieben
            close_table(device_address, hTable);
            fclose(fp);
            return -IV_ERROR_FILE_WRITE;
        }

        // Naechsten Datensatz abrufen
    } while ((res = write_current_table_row(device_address, hTable, 1, 1)) == IV_ERROR_SUCCESS);

    // Nicht: Sprung fuehrte zu EOF?
    // Funktionsfehler: 4
    if (res != 4)
    {
        // Modul oder Funktionsfehler?
        if (res > 0) close_table(device_address, hTable);
        fclose(fp);
        return res;
    }

    // Headerstruktur abschliessen und schreiben
    structHeader.type = type;
    structHeader.crc = GetCRC16((unsigned char*)&structHeader, sizeof(structHeader) - 2);

    // Auf Dateianfang und Headerstruktur schreiben
    if (fseek(fp, 0, 0) != 0 || fwrite(&structHeader, 1, sizeof(structHeader), fp) < sizeof(structHeader))
    {
        // Daten wurden nicht geschrieben
        close_table(device_address, hTable);
        fclose(fp);
        return -IV_ERROR_FILE_WRITE;
    }

    // Tabelle schliessen
    close_table(device_address, hTable);

    // Datei schliessen
    fclose(fp);

    return IV_ERROR_SUCCESS;
}

// Flags:
// Bit 1: gesetzt bedeutet: Enthaltene Fingertemplates beibehalten.
// Wiederherstellen der Fingerprints
int Channel4::fingerprintRestore( unsigned char device_address, const std::string &file_path, int flags )
{
    int res = IV_ERROR_SUCCESS;
    FILE *fp;
    StructFingertemplateId Id;
    unsigned char data[IV_PACKET_LENGTH_DATA];
    size_t size, sizeCheck;
    unsigned short crc, crcSegment;
    int recordSkiped = 0;
    std::string name("FINGERPRINT.IDLIST");
    std::string reserved;
    unsigned int currentCountAfter = 0;
    unsigned int unorderedCountAfter = 0;
    unsigned int deletedCountAfter = 0;
    unsigned char val = 0;
    HardwareInfo hardwareInfo;

    if ( (res = read_hardware_info( device_address, hardwareInfo )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    listItemBoard tags;
    // Nur wenn Biokey verbaut ist.
    tags = hardwareInfo.moduleBiokey();
    if ( tags.size() == 0 )
        return -IV_ERROR_NO_MODULE;

    // Datei oeffnen
    if ((fp = fopen(file_path.c_str(), "rb")) == nullptr)
    {
        return -IV_ERROR_FILE_OPEN;
    }

    // Kopfstruktur vorbereiten
    StructFTFileHeader structHeader;
    memset(&structHeader, 0, sizeof(StructFTFileHeader));
    // Kopfdaten lesen
    if (fread(&structHeader, 1, sizeof(StructFTFileHeader), fp) < sizeof(StructFTFileHeader))
    {
        // Daten wurden nicht gelesen
        fclose(fp);
        return -IV_ERROR_FILE_READ;
    }
    // Kopfdaten pruefen.
    crc = GetCRC16((unsigned char*)&structHeader, sizeof(StructFTFileHeader));
    if (structHeader.Id != 0xDFDF || 
        crc != 0 ||
        structHeader.type > 3)
    {
        fclose(fp);
        return -IV_ERROR_FILE_FORMAT;
    }
    // Auf Datenanfang wechseln
    fseek(fp, structHeader.offs, 0);

    // Groesse ist abhaengig von verwendetem Templatetyp.
    switch ( structHeader.type )
    {
    case 0:
        size = sizeof(StructFingertemplateDIN);
        break;

    case 1:
        size = sizeof(StructFingertemplateIdencom);
        break;

    case 2:
        size = sizeof(StructFingertemplateIdencomStandard);
        break;

    case 3:
        size = sizeof(StructFingertemplateSaturn01);
        break;

    default:
        size = 0;
    }

    // Datensaetze loeschen?
    if ( (flags & 1) == 1 )
    {
        // Alle Datensaetze vom Modul loeschen
        if ((res = delete_fingerprint_record(device_address, 0, 0)) != IV_ERROR_SUCCESS)
        {
            fclose(fp);
            return res;
        }
    }

    // Fingertemplates einspielen.
    while (!feof(fp) && !ferror(fp))
    {
        // Datenspeicher loeschen
        memset(data, 0, sizeof(data));
        memset(&Id, 0, sizeof(StructFingertemplateId));

        // Template aus Datei laden
        if ((sizeCheck = fread(data, 1, size, fp)) < size)
        {
            if (sizeCheck == 0 && feof(fp))
            {
                continue;
            }
            else if (sizeCheck && !ferror(fp))
            {
                fclose(fp);
                return -IV_ERROR_FILE_CRC; // Daten sind inkonsistent
            }
            // Daten wurden nicht gelesen
            fclose(fp);
            return -IV_ERROR_FILE_READ;
        }
        // CRC16 des Fingertemplates erstellen
        crc = GetCRC16(data, (unsigned int)size);
        // CRC lesen
        if ((sizeCheck = fread(&crcSegment, 1, sizeof(unsigned short), fp)) < sizeof(unsigned short)) {
            if ((sizeCheck == 0 && feof(fp)) || (sizeCheck && !ferror(fp)))
            {
                fclose(fp);
                return -IV_ERROR_FILE_CRC; // Daten sind inkonsistent
            }
            // Daten wurden nicht gelesen
            fclose(fp);
            return -IV_ERROR_FILE_READ;
        }
        // CRC pruefen
        if (crc != crcSegment)
        {
            fclose(fp);
            return -IV_ERROR_FILE_CRC;
        }

        // Das Fingertemplate anfuegen
        if ((res = append_fingerprint_record(device_address, structHeader.type, data, (unsigned int)size)) != IV_ERROR_SUCCESS)
        {
            // Wenn das Template bereits vorhanden war, dann ggf. uebergehen.
            // Funktionsfehler 134
            if (res == 134)
            {
                recordSkiped++;
                continue;
            }
            else if (res == 132)
            {
                // Speicher des Moduls ist voll
                res = -IV_ERROR_MEMORY_FULL;
            }
            else if (res == 1 || res == 3)
            {
                // Falsches Templateformat
                res = -IV_ERROR_FILE_DATA_FORMAT;

            }
            else if (res == 2)
            {
                int failLoop = 0;
                // Bei aelterer Hardware des Biokey3000 kann eine Verzoegerung nach dem loeschen der Datensaetze
                // vorkommen. Hier wird versucht das Template nochmals zu uebertragen.
                for (; failLoop < 2; failLoop++)
                {
                    if ((res = append_fingerprint_record(device_address, structHeader.type, data, (int)size)) == 2)
                    {
                        // Kleine Verzoegerung
                        //m_channel4.Trace("ERROR: GetInfoVerzoegerung: %i\n", failLoop);
                        std::this_thread::sleep_for( std::chrono::seconds( 2 ) );
                        continue;
                    }
                    break;
                }

                // Konnte der Befehl ausgefuehrt werden?
                // Wenn das Template bereits vorhanden war, dann ggf. uebergehen.
                // Funktionsfehler 134
                if (res == 134)
                {
                    recordSkiped++;
                    continue;
                }
                else if (res == 132)
                {
                    // Speicher des Moduls ist voll
                    res = -IV_ERROR_MEMORY_FULL;
                }
                else if (res == 1 || res == 3)
                {
                    // Falsches Templateformat
                    res = -IV_ERROR_FILE_DATA_FORMAT;
                }
                else if (res == IV_ERROR_SUCCESS)
                {
                    //m_channel4.Trace("OK: Verzoegerung: %i\n", failLoop);
                    continue;
                }
            }

            fclose(fp);
            return res;
        }
    }

    // Nicht: Sprung fuehrte zu EOF?
    if (ferror(fp))
    {
        fclose(fp);
        return -IV_ERROR_FILE_READ;
    }

    // Datei schliessen
    fclose(fp);

    // Tabelle mit Fingerprintdaten erzeugen
    if ((res = generate_table(device_address, name, reserved)) != IV_ERROR_SUCCESS)
    {
        // Liegen im Modul keine Daten vor?
        if (res == 133)
        {
            res = -IV_ERROR_NO_DATA_AVAILABLE;
        }
        return res;
    }
    int hTable;
    // Tabelle mit Fingerprintdaten oeffnen
    if ((res = open_table(device_address, name, 0, hTable, reserved)) != IV_ERROR_SUCCESS)
    {
        return res;
    }
    // Pruefen ob Datensaetze vorliegen
    if ((res = read_table_row_count(device_address, hTable, currentCountAfter, unorderedCountAfter, deletedCountAfter)) != IV_ERROR_SUCCESS)
    {
        // Modul oder Funktionsfehler?
        if (res > 0) close_table(device_address, hTable);
        return res;
    }
    // Tabelle schliessen
    if ((res = close_table(device_address, hTable)) != IV_ERROR_SUCCESS)
    {
        return res;
    }

    // Pruefen ob die Datensatzanzahl passt.
    if (structHeader.count > currentCountAfter)
    {
        // Es muessen mindestens die Anzahl geforderter Datensaetze vorhanden sein.		
    }

    return IV_ERROR_SUCCESS;
}

int Channel4::setOptionFirmware( unsigned char device_address, int mask, int options )
{
    int res = IV_ERROR_SUCCESS;

    // Externe Leserserie
    if (mask & OFF_EXTERNAL_READER)
    {
        if ((res = write_external_transponder_reader_type(device_address, (options & OFF_EXTERNAL_READER) ? 2 : 1)) != IV_ERROR_SUCCESS)
        {
            return res;
        }
    }

    // IOModule
    if (mask & OFF_IOMODULE)
    {
        if ((res = write_io_module_option(device_address, (options & OFF_IOMODULE) ? 1 : 0)) != IV_ERROR_SUCCESS)
        {
            // Erst ab Version 2 verfuegbar
            if (((res >> 8) & 0xFF) == 2)
            {
                // Die weitere Verarbeitung abbrechen.
                return IV_ERROR_SUCCESS;
            }

            return res;
        }
    }

    // Timeboy anbindung
    if (mask & OFF_TBADAPTION)
    {
        if ((res = write_timeboy_adaption_option(device_address, (options & OFF_TBADAPTION) ? 1 : 0)) != IV_ERROR_SUCCESS)
        {
            // Erst ab Version 2 verfuegbar
            if (((res >> 8) & 0xFF) == 2)
            {
                // Die weitere Verarbeitung abbrechen.
                return IV_ERROR_SUCCESS;
            }

            return res;
        }
    }

    // ComServer
    if (mask & OFF_COMSERVER)
    {
        if ((res = write_com_server_option(device_address, (options & OFF_COMSERVER) ? 1 : 0)) != IV_ERROR_SUCCESS)
        {
            // Erst ab Version 2 verfuegbar
            if (((res >> 8) & 0xFF) == 2)
            {
                // Die weitere Verarbeitung abbrechen.
                return IV_ERROR_SUCCESS;
            }

            return res;
        }
    }

    // Transponder schreiben
    if (mask & OFF_TPWRITE)
    {
        if ((res = write_transponder_write_option(device_address, (options & OFF_TPWRITE) ? 1 : 0)) != IV_ERROR_SUCCESS)
        {
            // Erst ab Version 2 verfuegbar
            if (((res >> 8) & 0xFF) == 2)
            {
                // Die weitere Verarbeitung abbrechen.
                return IV_ERROR_SUCCESS;
            }

            return res;
        }
    }

    // HTTP-Protokoll
    if (mask & OFF_HTTP)
    {
        if ((res = write_http_option(device_address, (options & OFF_HTTP) ? 1 : 0)) != IV_ERROR_SUCCESS)
        {
            // Erst ab Version 2 verfuegbar
            if (((res >> 8) & 0xFF) == 2)
            {
                // Die weitere Verarbeitung abbrechen.
                return IV_ERROR_SUCCESS;
            }

            return res;
        }
    }

    // Biokey update
    if (mask & OFF_BIOKEY_UPDATE)
    {
        if ((res = write_fingerprint_scanner_update_option(device_address, (options & OFF_BIOKEY_UPDATE) ? 1 : 0)) != IV_ERROR_SUCCESS)
        {
            // Erst ab Version 2 verfuegbar
            if (((res >> 8) & 0xFF) == 2)
            {
                // Die weitere Verarbeitung abbrechen.
                return IV_ERROR_SUCCESS;
            }

            return res;
        }
    }

    // SetupOption: Signalverarebeitung
    if (mask & OFF_SO_IOCONTROL)
    {
        if ((res = write_setup_options(device_address, 0x0001, (options & OFF_SO_IOCONTROL) ? 0x0001 : 0x0000)) != IV_ERROR_SUCCESS)
        {
            // Erst ab Version 3 verfuegbar
            if (((res >> 8) & 0xFF) == 2)
            {
                // Die weitere Verarbeitung abbrechen.
                return IV_ERROR_SUCCESS;
            }

            return res;
        }
    }

    // Systemflag fuer Setup und Listen beibehalten
    if (mask & OFF_KEEP_SETUP_AFTER_UPDATE)
    {
        unsigned int flags = 0;
        // Systemflags
        if ( (res = read_system_flags( device_address, flags )) != IV_ERROR_SUCCESS )
        {
            // Erst ab Version 2 verfuegbar
            if (((res >> 8) & 0xFF) == 2)
            {
                // Die weitere Verarbeitung abbrechen.
                return IV_ERROR_SUCCESS;
            }

            return res;
        }

        bool should_keep_setup_after_update = (options & OFF_KEEP_SETUP_AFTER_UPDATE) == OFF_KEEP_SETUP_AFTER_UPDATE;
        bool is_keep_setup_after_update = (flags & IV_VALUE_SYSTEM_FLAG_DEL_SETUP_AND_LISTS) != IV_VALUE_SYSTEM_FLAG_DEL_SETUP_AND_LISTS;

        if ( should_keep_setup_after_update != is_keep_setup_after_update )
        {
            flags &= ~IV_VALUE_SYSTEM_FLAG_DEL_SETUP_AND_LISTS;
            flags |= should_keep_setup_after_update ? 0 : IV_VALUE_SYSTEM_FLAG_DEL_SETUP_AND_LISTS;
            if ( (res = write_system_flags(device_address, flags )) != IV_ERROR_SUCCESS )
            {
                // Erst ab Version 2 verfuegbar
                if (((res >> 8) & 0xFF) == 2)
                {
                    // Die weitere Verarbeitung abbrechen.
                    return IV_ERROR_SUCCESS;
                }

                return res;
            }
        }
    }

    return res;
}

int Channel4::getOptionFirmware( unsigned char device_address, int &mask, int &options )
{
    int res = IV_ERROR_SUCCESS;
    unsigned char val;

    // Zuruecksetzen der Werte
    mask = options = 0;

    // Externe Leserserie
    if ((res = read_external_transponder_reader_type(device_address, val)) != IV_ERROR_SUCCESS)
    {
        return res;
    }
    mask |= OFF_EXTERNAL_READER;
    options |= (val == 2) ? OFF_EXTERNAL_READER : 0;

    // IOModule
    if ((res = read_io_module_option(device_address, val)) != IV_ERROR_SUCCESS)
    {
        return res;
    }
    mask |= OFF_IOMODULE;
    options |= (val == 1) ? OFF_IOMODULE : 0;

    // Timeboy anbindung
    if ((res = read_timeboy_adaption_option(device_address, val)) != IV_ERROR_SUCCESS)
    {
        return res;
    }
    mask |= OFF_TBADAPTION;
    options |= (val == 1) ? OFF_TBADAPTION : 0;

    // ComServer
    if ((res = read_com_server_option(device_address, val)) != IV_ERROR_SUCCESS)
    {	
        // Erst ab Version 2 verfuegbar
        if (((res >> 8) & 0xFF) == 2)
        {
            // Die weitere Verarbeitung abbrechen.
            return IV_ERROR_SUCCESS;
        }

        return res;
    }
    mask |= OFF_COMSERVER;
    options |= (val == 1) ? OFF_COMSERVER : 0;

    // Transponder schreiben
    if ((res = read_transponder_write_option(device_address, val)) != IV_ERROR_SUCCESS)
    {
        // Erst ab Version 2 verfuegbar
        if (((res >> 8) & 0xFF) == 2)
        {
            // Die weitere Verarbeitung abbrechen.
            return IV_ERROR_SUCCESS;
        }

        return res;
    }
    mask |= OFF_TPWRITE;
    options |= (val == 1) ? OFF_TPWRITE: 0;

    // Http-Protokoll
    if ((res = read_http_option(device_address, val)) != IV_ERROR_SUCCESS)
    {
        // Erst ab Version 2 verfuegbar
        if (((res >> 8) & 0xFF) == 2)
        {
            // Die weitere Verarbeitung abbrechen.
            return IV_ERROR_SUCCESS;
        }

        return res;
    }
    mask |= OFF_HTTP;
    options |= (val == 1) ? OFF_HTTP : 0;

    // Biokey Update
    if ((res = read_fingerprint_scanner_update_option(device_address, val)) != IV_ERROR_SUCCESS)
    {
        // Erst ab Version 2 verfuegbar
        if (((res >> 8) & 0xFF) == 2)
        {
            // Die weitere Verarbeitung abbrechen.
            return IV_ERROR_SUCCESS;
        }

        return res;
    }
    mask |= OFF_BIOKEY_UPDATE;
    options |= (val == 1) ? OFF_BIOKEY_UPDATE : 0;

    unsigned short soMask, soValue;
    // Setup Options: Signalverarbeitung
    if ((res = read_setup_options(device_address, soMask, soValue)) != IV_ERROR_SUCCESS)
    {
        // Erst ab Version 2 verfuegbar
        if (((res >> 8) & 0xFF) == 2)
        {
            // Die weitere Verarbeitung abbrechen.
            return IV_ERROR_SUCCESS;
        }

        return res;
    }
    mask |= OFF_SO_IOCONTROL;
    options |= (soValue & 0x0001) ? OFF_SO_IOCONTROL : 0;

    unsigned int flags = 0;
    // Systemflags
    if ( (res = read_system_flags( device_address, flags )) != IV_ERROR_SUCCESS )
    {
        // Erst ab Version 2 verfuegbar
        if (((res >> 8) & 0xFF) == 2)
        {
            // Die weitere Verarbeitung abbrechen.
            return IV_ERROR_SUCCESS;
        }

        return res;
    }
    mask |= OFF_KEEP_SETUP_AFTER_UPDATE;
    options |= ((flags & IV_VALUE_SYSTEM_FLAG_DEL_SETUP_AND_LISTS) == 0) ? OFF_KEEP_SETUP_AFTER_UPDATE : 0;

    return res;
}

int Channel4::setWlanConfigDataWithTimeboySlot( unsigned char device_address, const std::string &file_path )
{
    int res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = setWlanConfigData( device_address, file_path );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

int Channel4::setWlanConfigData( unsigned char device_address, const std::string &file_path )
{
    int res = IV_ERROR_SUCCESS;
    HardwareInfo hardwareInfo;

    if ( (res = read_hardware_info( device_address, hardwareInfo )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    listItemBoard tagsWLAN;
    listItemBoard tagsLAN;

    tagsWLAN = hardwareInfo.moduleWLANCommunication();
    // Nur wenn weder LAN noch WLAN verbaut ist.
    tagsLAN = hardwareInfo.moduleLANCommunication();
    if ( tagsWLAN.size() == 0 && tagsLAN.size() == 0 )
    {
        return -IV_ERROR_NO_MODULE;
    }

    // WlanConfig Daten
    FileWlanConfig fileWlanConfig;
    unsigned int address = 0;
    // WLAN Konfigurationsdatei laden
    if ( !fileWlanConfig.load( file_path.c_str() ) )
    {
        switch ( fileWlanConfig.errorState() )
        {
        case FileWlanConfig::Error_state::file_open:
            return -IV_ERROR_FILE_OPEN;

        case FileWlanConfig::Error_state::file_format:
        default:
            return -IV_ERROR_FILE_FORMAT;
        }
    }

    if ( !fileWlanConfig.WlanConfigCheckCRC( -1 ) )
    {
        return -IV_ERROR_FILE_CRC;
    }

    // Ggf. gesetztes Streaming deaktivieren.
    set_stream_transfer( false );
    clear_stream_transfer();

    // Systemstopp setzen
    if ( (res = enter_write_data_segment( device_address, IV_VALUE_SEGMENT_UPDATE_BG, address, 0, 0 )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // Sind WLAN Daten vorhanden?
    if ( fileWlanConfig.sizeData() )
    {
        address = 0;
        auto data = fileWlanConfig.data();
        auto sizeData = fileWlanConfig.sizeData();
        auto crc = fileWlanConfig.checksum32bit();

        // Systemstopp setzen
        res = enter_write_data_segment( device_address, IV_VALUE_SEGMENT_WLANCONFIGS, address, sizeData, crc );
        if ( res != IV_ERROR_SUCCESS && (res & 0xFF) != 1 )
        {
            return res;
        }

        // Funktionsfehler nicht Blocktyp unbekannt?
        if ( (res & 0xFF) != 1 )
        {
            // Textdaten uebertragen
            if ( (res = write_to_flash( device_address, address, data, sizeData )) != IV_ERROR_SUCCESS )
            {
                return res;
            }

            // Systemstopp aufheben
            if ( (res = exit_write_data_segment( device_address, IV_VALUE_SEGMENT_WLANCONFIGS, crc )) != IV_ERROR_SUCCESS )
            {
                return res;
            }
        }
        else
        {
            // Funktion wird von Firmware nicht unterstuetzt.
            return -IV_ERROR_FUNCTION_UNAVAILABLE;
        }
    }
    else
    {
        // Es liegen keine Daten fuer die Uebertragung vor.
        return -IV_ERROR_NO_DATA_AVAILABLE;
    }

    // Systemstopp aufheben
    if ( (res = exit_write_data_segment( device_address, IV_VALUE_SEGMENT_UPDATE_BG, 0 )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // WLAN Daten erfolgreich uebertragen
    return res;
}

int Channel4::getWlanConfigDataWithTimeboySlot( unsigned char device_address, const std::string &file_path )
{
    int res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = getWlanConfigData( device_address, file_path );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

int Channel4::getWlanConfigData( unsigned char device_address, const std::string &file_path )
{
    int res = IV_ERROR_SUCCESS;
    HardwareInfo hardwareInfo;

    if ( (res = read_hardware_info( device_address, hardwareInfo )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    listItemBoard tagsWLAN;
    listItemBoard tagsLAN;

    tagsWLAN = hardwareInfo.moduleWLANCommunication();
    tagsLAN = hardwareInfo.moduleLANCommunication();
    // Nur wenn weder LAN noch WLAN verbaut ist.
    if ( tagsWLAN.size() == 0  && tagsLAN.size() == 0 )
    {
        return -IV_ERROR_NO_MODULE;
    }

    //
    FileWlanConfig fileWlanConfig;
    unsigned int address = 0;
    unsigned int size = 0;
    unsigned int crc = 0;

    // Segmentbegin: Bedienung stoppen
    if ( (res = enter_read_data_segment( device_address, IV_VALUE_SEGMENT_GET_UPDATE_BG, address, size, crc )) != IV_ERROR_SUCCESS )
    {
        if ( (res & 0xFF) == 1 )
        {
            // Funktionsfehler Typ: IV_VALUE_SEGMENT_GET_UPDATE_BG unbekannt.
            // Wenn dass der Fall ist, wird das Auslesen noch nicht unterstuetzt.
            res = -IV_ERROR_FUNCTION_UNAVAILABLE;
        }
        else
        {
            return res;
        }
    }
    trace( "%s(%i):INFO(1): %s [ beginGetSegment(%i): %i, %i, 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_UPDATE_BG, address, size, crc );

    fileWlanConfig.resetData();
    //////////////////////////////
    // Begin

    // Segmentbegin: Setup uebertragen
    if ( (res = enter_read_data_segment( device_address, IV_VALUE_SEGMENT_WLANCONFIGS, address, size, crc )) != IV_ERROR_SUCCESS )
    {
        return res;
    }
    trace( "%s(%i):INFO(1): %s [ beginGetSegment(%i): %i, %i, 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_SETUP, address, size, crc );


    //////////////////////////////
    // Daten
    fileWlanConfig.setSizeData( size );
    // EEPROM uebertragen
    if ( (res = read_from_flash( device_address, address, fileWlanConfig.data(), fileWlanConfig.sizeData() )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    //////////////////////////////
    // Ende

    // Segmentend: Setup uebertragen
    if ( (res = exit_read_data_segment( device_address, IV_VALUE_SEGMENT_WLANCONFIGS )) != IV_ERROR_SUCCESS )
    {
        return res;
    }
    trace( "%s(%i):INFO(1): %s [ endGetSegment(%i) ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_SETUP );

    // Segmentend: Bedienung stoppen
    if ( (res = exit_read_data_segment( device_address, IV_VALUE_SEGMENT_GET_UPDATE_BG )) != IV_ERROR_SUCCESS )
    {
        return res;
    }
    trace( "%s(%i):INFO(1): %s [ endGetSegment(%i) ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_UPDATE_BG );

    // Datenanzahl in Datei uebernehmen.
    auto crcTest = fileWlanConfig.checksum32bit();
    if ( crc != crcTest )
    {
        trace( "%s(%i):ERROR(1): %s [ CRC von Segment falsch: 0x%08X, 0x%08X ]\n", __FILE__, __LINE__, __func__, crc, crcTest );
        return -IV_ERROR_CRC_SEGMENT;
    }

    // Datei speichern
    if ( !fileWlanConfig.save( file_path.c_str() ) )
    {
        trace( "%s(%i):ERROR(1): %s [ Schreiben der Datei gescheitert. ]\n", __FILE__, __LINE__, __func__ );
        return -IV_ERROR_FILE_WRITE;
    }

    return res;
}

int Channel4::setHardwareConfigDataWithTimeboySlot( unsigned char device_address, const std::string &file_path )
{
    int res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = setHardwareConfigData( device_address, file_path );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

int Channel4::setHardwareConfigData( unsigned char device_address, const std::string &file_path )
{
    // HardwareConfig Daten
    HardwareInfo hardwareInfo;
    FileHardwareInfo &fileHardwareConfig = hardwareInfo.fileHardwareInfo();
    int res = IV_ERROR_SUCCESS;
    unsigned int address = 0;
    unsigned char *buf;
    unsigned int size = 0;
    unsigned int crc = 0;

    // Hardware Konfigurationsdatei laden
    if ( !fileHardwareConfig.load( file_path ) )
    {
        switch ( fileHardwareConfig.errorState() )
        {
        case FileHardwareInfo::Error_state::file_open:
            return -IV_ERROR_FILE_OPEN;

        case FileHardwareInfo::Error_state::file_format:
        default:
            return -IV_ERROR_FILE_FORMAT;
        }
    }

    // Handelt es sich um eine MasterIV - HWI Datei, dann Abbruch mit Fehler.
    if ( hardwareInfo.isMasterIV() )
    {
        trace( "%s(%i):ERROR(1): %s [ Es handelt sich um eine HIP-Datei für MasterIV. ]\n", __FILE__, __LINE__, __func__ );
        return -IV_ERROR_FUNCTION_UNAVAILABLE;
    }

    //	memset(version, 0, sizeof(version));
    // Bootversion lesen 
    //	if ((res = m_channel4.GetVersionBootloader(device_address, version)) != IV_ERROR_SUCCESS)
    //{
    //		return res;
    //	}

    // Ggf. gesetztes Streaming deaktivieren.
    set_stream_transfer( false );
    clear_stream_transfer();

    // Systemstopp setzen
    if ( (res = enter_write_data_segment( device_address, IV_VALUE_SEGMENT_UPDATE_BG, address, size, crc )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // Sind HardwareConfig Daten vorhanden?
    if ( fileHardwareConfig.sizeData() )
    {
        address = 0;
        buf = const_cast<unsigned char*>( fileHardwareConfig.data() );
        size = fileHardwareConfig.sizeData();
        crc = fileHardwareConfig.checksum32bit();

        // Systemstopp setzen
        res = enter_write_data_segment( device_address, IV_VALUE_SEGMENT_HARDWARECONFIGS, address, size, crc );
        if ( res != IV_ERROR_SUCCESS && (res & 0xFF) != 1 )
        {
            return res;
        }

        // Funktionsfehler nicht Blocktyp unbekannt?
        if ( (res & 0xFF) != 1 )
        {
            // Textdaten uebertragen
            if ( (res = write_to_flash( device_address, address, buf, size )) != IV_ERROR_SUCCESS )
            {
                return res;
            }

            // Systemstopp aufheben
            if ( (res = exit_write_data_segment( device_address, IV_VALUE_SEGMENT_HARDWARECONFIGS, crc )) != IV_ERROR_SUCCESS )
            {
                return res;
            }
        }
        else
        {
            // Funktion wird von Firmware nicht unterstuetzt.
            return -IV_ERROR_FUNCTION_UNAVAILABLE;
        }
    }
    else
    {
        // Es liegen keine Daten fuer die Uebertragung vor.
        return -IV_ERROR_NO_DATA_AVAILABLE;
    }

    crc = 0;
    // Systemstopp aufheben
    if ( (res = exit_write_data_segment( device_address, IV_VALUE_SEGMENT_UPDATE_BG, crc )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // Hardware Daten erfolgreich uebertragen
    return res;
}

int Channel4::getHardwareConfigDataWithTimeboySlot( unsigned char device_address, const std::string &file_path )
{
    int res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = getHardwareConfigData( device_address, file_path );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

int Channel4::getHardwareConfigData( unsigned char device_address, const std::string &file_path )
{
    HardwareInfo hardwareInfo;
    int res = IV_ERROR_SUCCESS;

    if ( (res = read_hardware_info( device_address, hardwareInfo )) == IV_ERROR_SUCCESS )
    {
        // Datei speichern
        if ( !hardwareInfo.fileHardwareInfo().save( file_path ) )
        {
            trace( "%s(%i):ERROR(1): %s [ Schreiben der Datei gescheitert. ]\n", __FILE__, __LINE__, __func__ );
            return -IV_ERROR_FILE_WRITE;
        }
    }

    return res;
}

int Channel4::writeFileDataWithTimeboySlot( unsigned char device_address, const std::string &file_path )
{
    int res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = writeFileData( device_address, file_path );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

int Channel4::writeFileData( unsigned char device_address, const std::string &file_path )
{
    // HardwareConfig Daten
    int res = IV_ERROR_SUCCESS;
    int errorApplication = IV_ERROR_SUCCESS;

    // BlockTransfer durchfuehren.
    controller_segment &controllerSegmentCurrent = get_controller_segment( device_address );

    if ( controllerSegmentCurrent.transfer_state == controller_segment::bts_inactive )
    {
        struct stat fileStat;
        std::size_t pos = file_path.find_last_of( "\\/" );
        std::string fileName;
        if ( pos == std::string::npos )
        {
            fileName = file_path;
        }
        else
        {
            fileName = file_path.substr( pos + 1 );
        }

        if ( stat( file_path.c_str(), &fileStat ) != 0 )
        {
            return -IV_ERROR_FILE_OPEN;
        }

        FILE *fp = fopen( file_path.c_str(), "rb" );
        // Datei oeffnen
        if ( fp == nullptr )
        {
            return -IV_ERROR_FILE_OPEN;
        }

        controllerSegmentCurrent.file_data.resize( fileStat.st_size );
        if ( fread( &controllerSegmentCurrent.file_data[0], sizeof( unsigned char ), fileStat.st_size, fp ) != static_cast<size_t>( fileStat.st_size ) )
        {
            fclose( fp );
            trace( "%s(%i):ERROR(1): %s [ Lesen der Datei gescheitert. ]\n", __FILE__, __LINE__, __func__ );
            res = -IV_ERROR_FILE_READ;
            goto label_end;
        }

        fclose( fp );

        if ( (res = write_system_variable( device_address, "SYSTEM.FILETRANSFER", fileName ) ) != IV_ERROR_SUCCESS )
        {
            // Das Befehlsbyte wird bei unbekannter Variabler nicht veraendert
            // und wenn der Wert nicht gesetzt werden konnte wird das hoechste bit gesetzt.
            if ( (res & 0xFF) == 3 )
            {
                // Fehler liegt bei Wert
                res = -IV_ERROR_FUNCTION_UNAVAILABLE;
            }
            else if ( (res & 0xFF) == 2 )
            {
                // Variable nicht bekannt
                res = -IV_ERROR_FUNCTION_UNAVAILABLE;
            }

            goto label_end;
        }

        // Ggf. gesetztes Streaming deaktivieren.
        set_stream_transfer( false );
        clear_stream_transfer();

        unsigned int unsusedUInt = 0;
        // Systemstopp setzen
        if ( (res = enter_write_data_segment( device_address, IV_VALUE_SEGMENT_UPDATE_BG, unsusedUInt, 0, 0 )) != IV_ERROR_SUCCESS )
        {
            goto label_end;
        }

        // Beginn eintragen.
        controllerSegmentCurrent.reset();
        controllerSegmentCurrent.transfer_state = controller_segment::bts_active;
        controllerSegmentCurrent.transfer_type = controller_segment::btt_write_file;
        if ( controllerSegmentCurrent.allowed_transfer_types == controller_segment::btt_none )
        {
            // Settings ggf. von Global-Object holen.
            controller_segment &controllerSegmentGlobal = get_controller_segment( 255 );
            if ( controllerSegmentGlobal.allowed_transfer_types != controller_segment::btt_none )
            {
                controllerSegmentCurrent.allowed_transfer_types = controllerSegmentGlobal.allowed_transfer_types;
                controllerSegmentCurrent.duration = controllerSegmentGlobal.duration;
                controllerSegmentCurrent.transfer_flags |= controller_segment::cf_settings_from_global;
            }
        }
        controllerSegmentCurrent.begin_time_point = std::chrono::system_clock::now();

        controllerSegmentCurrent.address = 0;
        controllerSegmentCurrent.crc = get_crc32( &controllerSegmentCurrent.file_data[0], static_cast<unsigned int>( controllerSegmentCurrent.file_data.size() ) );

        // Systemstopp setzen
        if ( (res = enter_write_data_segment( device_address, IV_VALUE_SEGMENT_FILETRANSFER, controllerSegmentCurrent.address, 
            static_cast<unsigned int>( controllerSegmentCurrent.file_data.size() ), controllerSegmentCurrent.crc )) != IV_ERROR_SUCCESS )
        {
            if ( (res & 0xFF) == 1 )
            {
                // Funktion wird von Firmware nicht unterstuetzt.    
                res = -IV_ERROR_FUNCTION_UNAVAILABLE;
            }

            goto label_end;
        }
    }

    if ( controllerSegmentCurrent.discard == false )
    {
        // Dateidaten uebertragen
        if ( (res = write_to_flash( device_address, controllerSegmentCurrent.address + controllerSegmentCurrent.size, 
            &controllerSegmentCurrent.file_data[controllerSegmentCurrent.size], 
            static_cast<unsigned int>( controllerSegmentCurrent.file_data.size() - controllerSegmentCurrent.size )) ) != IV_ERROR_SUCCESS )
        {
            if ( res == -IV_ERROR_BLOCK_TRANSFER_BREAK )
            {
                controllerSegmentCurrent.resume = true;
                return res;
            }

            // Bei Kommunikationsfehlern abbrechen,
            // bei Ausfuehrungsfehlern das Uebertragen abbrechen.
            if ( res < IV_ERROR_SUCCESS )
            {
                goto label_end;
            }
            else
            {
                errorApplication = res;
            }
        }
    }

    // Dateidaten komplett uebertragen.
    controllerSegmentCurrent.resume = false;

    // Systemstopp aufheben
    if ( (res = exit_write_data_segment( device_address, IV_VALUE_SEGMENT_FILETRANSFER, controllerSegmentCurrent.crc )) != IV_ERROR_SUCCESS )
    {
        goto label_end;
    }

    // Systemstopp aufheben
    if ( (res = exit_write_data_segment( device_address, IV_VALUE_SEGMENT_UPDATE_BG, 0 )) != IV_ERROR_SUCCESS )
    {
        goto label_end;
    }

label_end:
    if ( controllerSegmentCurrent.file_data.size() )
    {
        controllerSegmentCurrent.file_data.clear();
    }

    // Blocktransfer beendet.
    controllerSegmentCurrent.end_time_point = std::chrono::system_clock::now();
    controllerSegmentCurrent.transfer_state = controller_segment::bts_inactive;

    // Ggf. aufgetretenen Fehler bei setFlash zurueckgeben.
    if ( errorApplication != IV_ERROR_SUCCESS )
    {
        res = errorApplication;
    }

    return res;
}

int Channel4::readFileDataWithTimeboySlot( unsigned char device_address, const std::string &file_path )
{
    int res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = readFileData( device_address, file_path );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

/*! Lesen von Dateidaten und speichern in der Datei unter file_path.

\param device_address Geraeteaddresse des anzusprechenden Geraetes.
\param file_path Dateiname + Pfad unter dem die Datei abgespeichert wird, der Dateiname selbst wird gelesen.

\retval IV_ERROR_SUCCESS Erfolgreich ausgelesen und abgespeichert.
\retval > IV_ERROR_SUCCESS Fehler auf der Anwendungsschicht der Firmware.
\retval < IV_ERROR_SUCCESS Fehler in der Kommunikation.
*/
int Channel4::readFileData( unsigned char device_address, const std::string &file_path )
{
    int res = IV_ERROR_SUCCESS;
    int errorApplication = IV_ERROR_SUCCESS;
    unsigned int address;
    unsigned int size;
    unsigned int crc;
    unsigned int crcTest;
    unsigned int addressOffset = 0;
    address = size = crc = 0;

    // BlockTransfer durchfuehren.
    controller_segment &controllerSegmentCurrent = get_controller_segment( device_address );

    if ( controllerSegmentCurrent.transfer_state == controller_segment::bts_inactive )
    {    
        std::size_t pos = file_path.find_last_of( "\\/" );
        std::string fileName;
        if ( pos == std::string::npos )
        {
            fileName = file_path;
        }
        else
        {
            fileName = file_path.substr( pos + 1 );
        }

        //

        if ( (res = write_system_variable( device_address, "SYSTEM.FILETRANSFER", fileName ) ) != IV_ERROR_SUCCESS )
        {
            // Das Befehlsbyte wird bei unbekannter Variabler nicht veraendert
            // und wenn der Wert nicht gesetzt werden konnte wird das hoechste bit gesetzt.
            if ( (res & 0xFF) == 3 )
            {
                // Fehler liegt bei Wert
                res = -IV_ERROR_FUNCTION_UNAVAILABLE;
            }
            else if ( (res & 0xFF) == 2 )
            {
                // Variable nicht bekannt
                res = -IV_ERROR_FUNCTION_UNAVAILABLE;
            }

            return res;
        }

        unsigned int unusedUInt = 0;
        // Segmentbegin: Hintergrunduebertragung
        if ( (res = enter_read_data_segment( device_address, IV_VALUE_SEGMENT_GET_UPDATE_BG, unusedUInt, unusedUInt, unusedUInt )) != IV_ERROR_SUCCESS )
        {
            if ( (res & 0xFF) == 1 )
            {
                // Funktionsfehler Typ: IV_VALUE_SEGMENT_GET_UPDATE_BG unbekannt.
                // Wenn dass der Fall ist, wird das Auslesen noch nicht unterstuetzt.
                return -IV_ERROR_FUNCTION_UNAVAILABLE;
            }
            else
            {
                return res;
            }
        }
        trace( "%s(%i):INFO(1): %s [ beginGetSegment(%i): %i, %i, 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_UPDATE_BG, address, size, crc );

        //////////////////////////////
        // Begin

        // Segmentbegin: Dateidaten lesen
        if ( (res = enter_read_data_segment( device_address, IV_VALUE_SEGMENT_FILETRANSFER, address, size, crc )) != IV_ERROR_SUCCESS )
        {
            return res;
        }
        trace( "%s(%i):INFO(1): %s [ beginGetSegment(%i): %i, %i, 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_SETUP, address, size, crc );

        // Liegen Daten zur Datei vor?
        if ( size == 0 )
        {
            // Datei konnte nicht geoeffnet werden.
            //@TODO: Besseren Fehler fuer die Bedeutung. Zu lesende Datei existiert nicht, oder so aehnlich.
            errorApplication = -IV_ERROR_FILE_OPEN;
            goto label_end;
        }

        // Beginn eintragen.
        controllerSegmentCurrent.reset();
        controllerSegmentCurrent.transfer_state = controller_segment::bts_active;
        controllerSegmentCurrent.transfer_type = controller_segment::btt_read_file;
        if ( controllerSegmentCurrent.allowed_transfer_types == controller_segment::btt_none )
        {
            // Settings ggf. von Global-Object holen.
            controller_segment &controllerSegmentGlobal = get_controller_segment( 255 );
            if ( controllerSegmentGlobal.allowed_transfer_types != controller_segment::btt_none )
            {
                controllerSegmentCurrent.allowed_transfer_types = controllerSegmentGlobal.allowed_transfer_types;
                controllerSegmentCurrent.duration = controllerSegmentGlobal.duration;
                controllerSegmentCurrent.transfer_flags |= controller_segment::cf_settings_from_global;
            }
        }
        controllerSegmentCurrent.begin_time_point = std::chrono::system_clock::now();

        // Ermittelte Daten halten.
        controllerSegmentCurrent.file_path = file_path;
        controllerSegmentCurrent.address = address;
        controllerSegmentCurrent.crc = crc;
        // Dateigroesse halten.
        controllerSegmentCurrent.file_data.resize( size );
    }
    else
    {
        // Zwischengespeicherte Daten wiederherstellen.
        address = controllerSegmentCurrent.address;
        crc = controllerSegmentCurrent.crc;
        addressOffset = controllerSegmentCurrent.size;
    }

    if ( controllerSegmentCurrent.discard == false && errorApplication == IV_ERROR_SUCCESS )
    {
        //////////////////////////////
        // Daten

        // EEPROM uebertragen
        if ( (res = read_from_flash( device_address, address + addressOffset, &controllerSegmentCurrent.file_data[addressOffset], 
            static_cast<unsigned int>( controllerSegmentCurrent.file_data.size() - addressOffset )) ) != IV_ERROR_SUCCESS )
        {
            if ( res == -IV_ERROR_BLOCK_TRANSFER_BREAK )
            {
                controllerSegmentCurrent.resume = true;
                return res;
            }

            // Bei Kommunikationsfehlern abbrechen,
            // bei Ausfuehrungsfehlern das Uebertragen abbrechen.
            if ( res < IV_ERROR_SUCCESS )
            {
                goto label_end;
            }
            else
            {
                errorApplication = res;
            }
        }
    }

    // Dateidaten komplett uebertragen.
    controllerSegmentCurrent.resume = false;

    //////////////////////////////
    // Ende

    // Segmentend: Setup uebertragen
    if ( (res = exit_read_data_segment( device_address, IV_VALUE_SEGMENT_FILETRANSFER )) != IV_ERROR_SUCCESS )
    {
        goto label_end;
    }
    trace( "%s(%i):INFO(1): %s [ endGetSegment(%i) ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_SETUP );

    // Segmentend: Bedienung stoppen
    if ( (res = exit_read_data_segment( device_address, IV_VALUE_SEGMENT_GET_UPDATE_BG )) != IV_ERROR_SUCCESS )
    {
        goto label_end;
    }
    trace( "%s(%i):INFO(1): %s [ endGetSegment(%i) ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_UPDATE_BG );

    if (errorApplication == IV_ERROR_SUCCESS)
    {    
        crcTest = get_crc32( controllerSegmentCurrent.file_data.data(), static_cast<unsigned int>( controllerSegmentCurrent.file_data.size() ) );
        if (crc != crcTest)
        {
            trace( "%s(%i):ERROR(1): %s [ CRC von Segment falsch: 0x%08X, 0x%08X ]\n", __FILE__, __LINE__, __func__, crc, crcTest );
            res = -IV_ERROR_CRC_SEGMENT;
            goto label_end;
        }

        std::size_t pos = controllerSegmentCurrent.file_path.find_last_of( "\\/" );
        std::string fileName;
        if ( pos == std::string::npos )
        {
            fileName = controllerSegmentCurrent.file_path;
        }
        else
        {
            fileName = controllerSegmentCurrent.file_path.substr(pos + 1);
        }
        // Bei Verwendung von '*' oder '?' im Dateinamen wird ggf. die Systemvariable mit dem richtigen Namen der Datei gefüllt
        // und muss im Nachgang ausgelesen und verwendet werden.
        pos = fileName.find_first_of( "*?" );
        if ( pos != std::string::npos )
        {
            std::string tmpFileName;
            if ( (res = read_system_variable( device_address, "SYSTEM.FILETRANSFER", tmpFileName ) ) != IV_ERROR_SUCCESS )
            {
                // Das Befehlsbyte wird bei unbekannter Variabler nicht veraendert
                // und wenn der Wert nicht gesetzt werden konnte wird das hoechste bit gesetzt.
                if ( (res & 0xFF) == 3 )
                {
                    // Fehler liegt bei Wert
                    res = -IV_ERROR_FUNCTION_UNAVAILABLE;
                }
                else if ( (res & 0xFF) == 2 )
                {
                    // Variable nicht bekannt
                    res = -IV_ERROR_FUNCTION_UNAVAILABLE;
                }

                goto label_end;
            }

            if ( fileName == tmpFileName )
            {
                // Die Musterzeichen dürfen nun nicht mehr im Namen vorhanden sein. Dateien mit diesen Musterzeichen dürfen im System nicht existieren.
                // Dieses führt bei fopen zu einem Fehler.
            }
            else
            {
                pos = controllerSegmentCurrent.file_path.find_last_of( "\\/" );
                controllerSegmentCurrent.file_path.replace( pos + 1, fileName.size(), tmpFileName );
            }
        }

        // Datenanzahl in Datei uebernehmen.
        FILE *fp = fopen( controllerSegmentCurrent.file_path.c_str(), "wb" );
        // Datei oeffnen
        if ( fp == nullptr )
        {
            res = -IV_ERROR_FILE_OPEN;
            goto label_end;
        }

        if ( fwrite( controllerSegmentCurrent.file_data.data(), sizeof(unsigned char), controllerSegmentCurrent.file_data.size(), fp ) != size )
        {
            fclose( fp );
            trace( "%s(%i):ERROR(1): %s [ Schreiben der Datei gescheitert. ]\n", __FILE__, __LINE__, __func__ );
            res = -IV_ERROR_FILE_WRITE;
            goto label_end;
        }

        fclose( fp );
    }

label_end:
    // Dateidaten loeschen.
    controllerSegmentCurrent.file_path.clear();
    controllerSegmentCurrent.file_data.clear();

    // Blocktransfer beendet.
    controllerSegmentCurrent.end_time_point = std::chrono::system_clock::now();
    controllerSegmentCurrent.transfer_state = controller_segment::bts_inactive;

    // Ggf. aufgetretenen Fehler bei Verarbeitung zurueckgeben.
    if ( errorApplication != IV_ERROR_SUCCESS )
    {
        res = errorApplication;
    }

    return res;
}

int Channel4::setTextDataWithTimeboySlot( unsigned char device_address, const std::string &file_path )
{
    int res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = setTextData( device_address, file_path );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

int Channel4::setTextData( unsigned char device_address, const std::string &file_path )
{
    // Zeichensatz- und Grafikdaten
    FileDeviceText *pFileDeviceText = m_file_device.GetFileDeviceText();
    int res = IV_ERROR_SUCCESS;
    unsigned int address = 0;

    if ( pFileDeviceText->Import( file_path ) == false )
    {
        switch ( pFileDeviceText->errorState() )
        {
        case FileIntelHex::Error_state::file_open:
            return -IV_ERROR_FILE_OPEN;

        case FileIntelHex::Error_state::file_format:
        default:
            return -IV_ERROR_FILE_FORMAT;
        }
    }

    // Ggf. gesetztes Streaming deaktivieren.
    set_stream_transfer( false );
    clear_stream_transfer();

    // Systemstopp setzen
    if ( (res = enter_write_data_segment( device_address, IV_VALUE_SEGMENT_UPDATE, address, 0, 0 )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // Sind Textdaten vorhanden?
    if ( pFileDeviceText->sizeData() )
    {
        // Type User setzen.
        TEXTDATA *td = reinterpret_cast<TEXTDATA *>( pFileDeviceText->data() );
        td->type = (m_load_as_default_language) ? 2 : 3; // Textdata Default/User
        m_load_as_default_language = 0; // ggf. zuruecksetzen

        address = 0;
        // Systemstopp setzen
        res = enter_write_data_segment( device_address, IV_VALUE_SEGMENT_TEXT, address, pFileDeviceText->sizeData(), pFileDeviceText->checksum32bit() );
        if (res != IV_ERROR_SUCCESS && (res & 0xFF) != 1)
        {
            return res;
        }

        // Funktionsfehler nicht Blocktyp unbekannt?
        if ( (res & 0xFF) != 1 )
        {
            // Textdaten uebertragen
            if ( (res = write_to_flash( device_address, address, pFileDeviceText->data(), pFileDeviceText->sizeData() )) != IV_ERROR_SUCCESS )
            {
                return res;
            }

            // Systemstopp aufheben
            if ( (res = exit_write_data_segment( device_address, IV_VALUE_SEGMENT_TEXT, pFileDeviceText->checksum32bit() )) != IV_ERROR_SUCCESS )
            {
                return res;
            }
        }
        else
        {
            // Funktion wird von Firmware nicht unterstuetzt.
            return -IV_ERROR_FUNCTION_UNAVAILABLE;
        }
    }
    else
    {
        // Es liegen keine Daten fuer die Uebertragung vor.
        return -IV_ERROR_NO_DATA_AVAILABLE;
    }

    // Systemstopp aufheben
    if ( (res = exit_write_data_segment( device_address, IV_VALUE_SEGMENT_UPDATE, 0 )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // Textdaten erfolgreich uebertragen
    return res;
}

int Channel4::setTouchKeyDataWithTimeboySlot( unsigned char device_address, const std::string &file_path )
{
    int res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = setTouchKeyData( device_address, file_path );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

int Channel4::setTouchKeyData( unsigned char device_address, const std::string &file_path )
{
    int res = IV_ERROR_SUCCESS;
    HardwareInfo hardwareInfo;

    if ( (res = read_hardware_info( device_address, hardwareInfo )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    listItemBoard tags;
    // Nur wenn Touch verbaut ist.
    tags = hardwareInfo.moduleTouch();
    if ( tags.size() == 0 )
    {
        return -IV_ERROR_NO_MODULE;
    }

    // Touchkeydaten
    FileTouchKeys fileTouchKeys;;
    // Touchdatendatei laden
    if ( !fileTouchKeys.load( file_path ) )
    {
        switch ( fileTouchKeys.errorState() )
        {
        case FileTouchKeys::Error_state::file_open:
            return -IV_ERROR_FILE_OPEN;

        case FileTouchKeys::Error_state::file_format:
        default:
            return -IV_ERROR_FILE_FORMAT;
        }
    }

    // Ggf. gesetztes Streaming deaktivieren.
    set_stream_transfer( false );
    clear_stream_transfer();

    unsigned int address = 0;
    // Systemstopp setzen
    if ( (res = enter_write_data_segment( device_address, IV_VALUE_SEGMENT_UPDATE, address, 0, 0 )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // Sind TouchKeyDaten vorhanden?
    if ( fileTouchKeys.sizeData() )
    {
        address = 0;
        auto data = fileTouchKeys.data();
        auto sizeData = fileTouchKeys.sizeData();
        auto crc = fileTouchKeys.checksum32bit();

        // Systemstopp setzen
        res = enter_write_data_segment( device_address, IV_VALUE_SEGMENT_TOUCHKEYS, address, sizeData, crc );
        if ( res != IV_ERROR_SUCCESS && (res & 0xFF) != 1 )
        {
            return res;
        }

        // Funktionsfehler nicht Blocktyp unbekannt?
        if ( (res & 0xFF) != 1 )
        {
            // Textdaten uebertragen
            if ( (res = write_to_flash( device_address, address, data, sizeData )) != IV_ERROR_SUCCESS )
            {
                return res;
            }

            // Systemstopp aufheben
            if ( (res = exit_write_data_segment( device_address, IV_VALUE_SEGMENT_TOUCHKEYS, crc )) != IV_ERROR_SUCCESS )
            {
                return res;
            }
        }
        else
        {
            // Funktion wird von Firmware nicht unterstuetzt.
            return -IV_ERROR_FUNCTION_UNAVAILABLE;
        }
    }
    else
    {
        // Es liegen keine Daten fuer die Uebertragung vor.
        return -IV_ERROR_NO_DATA_AVAILABLE;
    }

    // Systemstopp aufheben
    if ( (res = exit_write_data_segment( device_address, IV_VALUE_SEGMENT_UPDATE, 0 )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // TouchKeyDaten erfolgreich uebertragen
    return res;
}

int Channel4::setDisplayColorsWithTimeboySlot( unsigned char device_address, const std::string &file_path )
{
    int res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = setDisplayColors( device_address, file_path );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

int Channel4::setDisplayColors( unsigned char device_address, const std::string &file_path )
{
    int res = IV_ERROR_SUCCESS;
    HardwareInfo hardwareInfo;

    if ( (res = read_hardware_info( device_address, hardwareInfo )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    if ( hardwareInfo.deviceType() != HardwareInfo::TimeboyIV )
    {
        return -IV_ERROR_DEVICE_WRONG;
    }

    listItemBoard tags;
    // Nur wenn Farbdisplay verbaut ist.
    tags = hardwareInfo.moduleDisplay( boost::assign::list_of(EnumModuleDisplay::Color_1_9)(EnumModuleDisplay::Color_2_0) );
    if ( tags.size() == 0 )
    {
        return -IV_ERROR_NO_MODULE;
    }

    // Farbdaten
    FileDisplayColors fileDisplayColors;
    // Touchdatendatei laden
    if ( !fileDisplayColors.load( file_path ) )
    {
        switch ( fileDisplayColors.errorState() )
        {
        case FileDisplayColors::Error_state::file_open:
            return -IV_ERROR_FILE_OPEN;

        case FileDisplayColors::Error_state::file_format:
        default:
            return -IV_ERROR_FILE_FORMAT;
        }
    }

    // Ggf. gesetztes Streaming deaktivieren.
    set_stream_transfer( false );
    clear_stream_transfer();

    unsigned int address = 0;
    // Systemstopp setzen
    if ( (res = enter_write_data_segment( device_address, IV_VALUE_SEGMENT_UPDATE, address, 0, 0 )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // Sind TouchKeyDaten vorhanden?
    if ( fileDisplayColors.sizeData() )
    {
        address = 0;
        auto data = fileDisplayColors.data();
        auto sizeData = fileDisplayColors.sizeData();
        auto crc = fileDisplayColors.checksum32bit();

        // Systemstopp setzen
        res = enter_write_data_segment( device_address, IV_VALUE_SEGMENT_DISPLAYCOLORS, address, sizeData, crc );
        if ( res != IV_ERROR_SUCCESS && (res & 0xFF) != 1 )
        {
            return res;
        }

        // Funktionsfehler nicht Blocktyp unbekannt?
        if ( (res & 0xFF) != 1 )
        {
            // Textdaten uebertragen
            if ( (res = write_to_flash( device_address, address, data, sizeData )) != IV_ERROR_SUCCESS )
            {
                return res;
            }

            // Systemstopp aufheben
            if ( (res = exit_write_data_segment( device_address, IV_VALUE_SEGMENT_DISPLAYCOLORS, crc )) != IV_ERROR_SUCCESS )
            {
                return res;
            }
        }
        else
        {
            // Funktion wird von Firmware nicht unterstuetzt.
            return -IV_ERROR_FUNCTION_UNAVAILABLE;
        }
    }
    else
    {
        // Es liegen keine Daten fuer die Uebertragung vor.
        return -IV_ERROR_NO_DATA_AVAILABLE;
    }

    // Systemstopp aufheben
    if ( (res = exit_write_data_segment( device_address, IV_VALUE_SEGMENT_UPDATE, 0 )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // DisplayColorDaten erfolgreich uebertragen
    return res;
}

int Channel4::getTablesWithTimeboySlot( unsigned char device_address, const std::string &type_name )
{
    int res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ((res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = getTables( device_address, type_name );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

// Datensatzbeschreibungen aus Geraet lesen
int Channel4::getTables( unsigned char device_address, const std::string &type_name )
{
    unsigned char tableCount, fieldCount, i;
    int err = IV_ERROR_SUCCESS;

    StructTable2 table;
    StructTable2Field field;
    StructDatensatzBeschreibung *pTable = nullptr;
    StructDatensatzFeld *pField = nullptr;

    // Struktur uebernehmen
    if (type_name == "LIST")
    {
        MaxListB = 0;
    }
    else if (type_name == "RECORD")
    {
        MaxDatB = 0;
    }
    else
    {
        return -IV_ERROR_LIST_PROCESSING;
    }

    // Alle Tabellenbeschreibungen lesen
    for ( tableCount = 0; tableCount < MAX_DATENSATZ_BESCHREIBUNG; ++tableCount )
    {
        std::string name( type_name );
        name += "." + std::to_string( tableCount );

        if ( (err = read_table_struct( device_address, name, table )) != IV_ERROR_SUCCESS )
        {
            // Nur wenn der Minorname (Index) nicht bekannt ist, 
            // liegt wirklich ein Fehler vor.
            if ( err < 0 )
            {
                // Kommunikationsfehler
                return err;
            }
            else if ( err == 2 )
            { 
                // Tabellenname unbekannt
                // Fehlercode zuruecksetzen
                err = IV_ERROR_SUCCESS;
            }
            else
            {
                // Funktionsfehler
                return err;
            }

            break;
        }

        // Struktur uebernehmen
        if ( type_name == "LIST" )
        {
            pTable = &ListenBeschreibung[tableCount];
        }
        else if ( type_name == "RECORD" )
        {
            pTable = &DatensatzBeschreibung[tableCount];
        }

        strncpy( (char*)pTable->Name, table.name, sizeof( pTable->Name ) - 1 );
        pTable->Felder = table.count;
        pTable->P = 0;
    }

    // Feldbeschreibung lesen
    for ( i = 0; i < tableCount; ++i )
    {
        // Struktur uebernehmen
        if ( type_name == "LIST" )
        {
            pTable = &ListenBeschreibung[i];
        }
        else if ( type_name == "RECORD" )
        {
            pTable = &DatensatzBeschreibung[i];
        }

        // Alle Felder lesen
        for ( fieldCount = 0; fieldCount < pTable->Felder; ++fieldCount )
        {
            std::string name( type_name );
            name += "." + std::to_string( i ) + "." + std::to_string( fieldCount );

            if ( (err = read_table_column_struct( device_address, name, field )) != IV_ERROR_SUCCESS )
            {
                if ( err < 0 )
                {
                    // Kommunikationsfehler
                    return err;
                }
                else
                {
                    // Funktionsfehler
                    return err;
                }
            }

            // Struktur uebernehmen
            if ( type_name == "LIST" )
            {
                pField = &ListenFeld[i][fieldCount];
            }
            else if ( type_name == "RECORD" )
            {
                pField = &DatensatzFeld[i][fieldCount];
            }

            pField->Typ = field.type;
            pField->Cnt = field.size;
            strncpy( (char*)pField->Name, field.name, sizeof( pField->Name ) - 1 );
        }
    }

    // Information uebernehmen
    if ( type_name == "LIST" )
    {
        MaxListB = tableCount;
    }
    else if ( type_name == "RECORD" )
    {
        MaxDatB = tableCount;
    }

    return err;
}

int Channel4::writeSetupWithTimeboySlot( unsigned char device_address, const std::string &file_path )
{
    int res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ((res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        //char val[80];
        //// Systemvariablen fuer MD5-Pruefwerte lesen.
        //// Wenn die Systemvariable definiert ist, dann wird eine andere Art der Setupuebertragung verwendent.
        //if ((res = getSystemVariable(device_address, "SYSTEM.TMP1", val)) != IV_ERROR_SUCCESS) {

        //    // Bei Kommunikationsfehler, abbruch.
        //    if (res < IV_ERROR_SUCCESS)
        //        return res;

        //    // Befehlsausfuehrung starten.
        res = writeSetupData( device_address, file_path );
        //} else {

        //    // Befehlsausfuehrung starten.
        //    res = writeSetupData4_2( device_address );
        //}

        if (res == IV_ERROR_SUCCESS && !is_password_protection() )
            res = clear_records( device_address );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

int Channel4::writeSetupData( unsigned char device_address, const std::string &file_path )
{
    int res = IV_ERROR_SUCCESS;
    int fileError;

    // BlockTransfer durchfuehren.
    controller_segment &controllerSegmentCurrent = get_controller_segment( device_address );

    if ( controllerSegmentCurrent.transfer_state == controller_segment::bts_inactive )
    {
        std::string ext( file_path.substr( file_path.size() - 4 ) );
        boost::to_lower( ext );

        if ( ext != ".aes" && ext != ".ael" )
        {
            // Die Anwendung kann ueber den gewaehlten Weg, nur Dateien vom Typ (*.aes; *.ael) oeffnen.
            return -IV_ERROR_FILE_EXTENSION;
        }

        // Reset der StudioFile-Klasse
        m_file.SetDefault();

        // Laden der Dateidaten. Keine Beachtung der Studioversion.
        if ( m_file.load( file_path, &fileError, static_cast<FileStudio::Enum_flag_ignore_version>( FileStudio::fiv_min_studio | FileStudio::fiv_should_studio ) ) == false )
        {
            switch ( fileError )
            {
            case -3:
                return -IV_ERROR_FILE_READ;

            case -7: // Mindestversion DLL wird nicht eingehalten.
                return -IV_ERROR_SETUPFILE_MIN_LIBRARY_VERSION;

            case -8: // Sollteversion DLL
                return -IV_ERROR_SETUPFILE_SHOULD_LIBRARY_VERSION;

            case -9: // Format der Setupdatei
                return -IV_ERROR_FILE_FORMAT;

            case -4: // CRC in Dateidaten
            case -10: case -11: case -12: case -13: case -14: case -15: case -16: case -17: case -18:
                return -IV_ERROR_FILE_CRC;
            }

            return -IV_ERROR_FILE_OPEN;
        }

        int res;
        std::string version;
        if ( (res = read_firmware_version( device_address, version )) != IV_ERROR_SUCCESS )
        {
            return res;
        }

        SoftwareVersion versionFirmware( version );
        unsigned char type = versionFirmware.typeDevice();

        // Pruefen ob die Firmwareversion unterstuetzt wird.
        if ( !m_file.IsSupportedFirmwareVersion( version.c_str() ) )
        {
            return -IV_ERROR_VERSION_FIRMWARE;
        }

        if ( (m_file.m_flash_header.DEVICE_TYPE - 1) != type )
        {
            // Sonderfall IPC-Extension
            switch ( type )
            {
            case IV_VALUE_TYPE_DEVICE_IPC_EXTENSION_VARIO_5_7:
            case IV_VALUE_TYPE_DEVICE_IPC_EXTENSION_VARIO_10:
            case IV_VALUE_TYPE_DEVICE_IPC_EXTENSION_Q7:
                {
                    // Alle IPC-Extensions werden als die SetupId IV_VALUE_TYPE_DEVICE_IPC_EXTENSION_VARIO_5_7 geführt.
                    if ( (m_file.m_flash_header.DEVICE_TYPE - 1) != IV_VALUE_TYPE_DEVICE_IPC_EXTENSION_VARIO_5_7 )
                    {
                        return -IV_ERROR_DEVICE_WRONG;
                    }

                    break;
                }
            case IV_VALUE_TYPE_DEVICE_TERMINAL_2_8:
            case IV_VALUE_TYPE_DEVICE_TERMINAL_3_5:
                {
                    if ( (m_file.m_flash_header.DEVICE_TYPE - 1) != IV_VALUE_TYPE_DEVICE_TERMINAL_2_8  && (m_file.m_flash_header.DEVICE_TYPE - 1) != IV_VALUE_TYPE_DEVICE_TERMINAL_3_5 )
                    {
                        return -IV_ERROR_DEVICE_WRONG;
                    }
                    break;
                }
            default:
                return -IV_ERROR_DEVICE_WRONG;
            }
        }

        // Ist ein Dateiheader verfuegbar, ist die Firmwareversion zu pruefen.
        if ( m_file.IsHeader() )
        {
            SoftwarebuildVersion fwVersion( version );
            // Achtung:
            // Bei der Auswertung ist die Variante "SHOULD" durch die Library nicht zu berücksichtigen, da sonst ein Einspielen verhindert würde.
            switch ( m_file.ValidFirmwareVersion( fwVersion ) )
            {
            case 0: // Kein Dateiheader, hier unmoeglich.
                return -IV_ERROR_FILE_FORMAT;

            case -1: // Unsupported 04.01.xx
                return -IV_ERROR_SETUPFILE_UNSUPPORTED_FIRMWARE1_VERSION;

            case -2: // Mindestversion 04.01.xx
                return -IV_ERROR_SETUPFILE_MIN_FIRMWARE1_VERSION;

                //case -3: // Sollteversion 04.01.xx
                //return -IV_ERROR_SETUPFILE_SHOULD_FIRMWARE1_VERSION;

            case -4: // Unsupported 04.02.xx
                return -IV_ERROR_SETUPFILE_UNSUPPORTED_FIRMWARE2_VERSION;

            case -5: // Mindestversion 04.02.xx
                return -IV_ERROR_SETUPFILE_MIN_FIRMWARE2_VERSION;

                //case -6: // Sollteversion 04.02.xx
                //return -IV_ERROR_SETUPFILE_SHOULD_FIRMWARE2_VERSION;

            case -7: // Firmware wird nicht unterstuetz.
                return -IV_ERROR_VERSION_FIRMWARE;

            case -8: // Unsupported 04.03.xx
                return -IV_ERROR_SETUPFILE_UNSUPPORTED_FIRMWARE3_VERSION;

            case -9: // Mindestversion 04.03.xx
                return -IV_ERROR_SETUPFILE_MIN_FIRMWARE3_VERSION;

                //case -10: // Sollteversion 04.03.xx
                //return -IV_ERROR_SETUPFILE_SHOULD_FIRMWARE3_VERSION;
            }
        }
        else
        {
            // Wenn kein Dateiheader vorliegt, wird sichergestellt, dass eine Setupdatei mit der hinterlegten
            // DatafoxStudioIV Version 04.02.xx nur auf eine Hardware >= V3 aufgespielt werden kann.

            // Boardtyp ermitteln und Pruefen ob die Setupversion eingespielt werden darf.
            if ( (res = read_board_type( device_address, type )) != IV_ERROR_SUCCESS )
            {
                if ( res  == -IV_ERROR_FUNCTION_UNAVAILABLE )
                {
                    // Baugruppe als 1.4 annehmen
                    type = 0x14;
                }
                else
                {
                    return res;
                }
            }

            if  ( type < IV_VALUE_TYPE_BOARD_3_0 )
            {
                // Neuere Setupdateien nicht in aeltere Hardwarestaende einspielen.
                //############################
                // 2010-03-04
                auto setup_version = m_file.endian_reverse( m_file.m_flash_header.SETUP_VERSION );
                if ( (setup_version & 0xFF000000) > 4 || setup_version > 0x04010000 )
                {
                    return -IV_ERROR_VERSION_SETUPFILE;
                }
            }
        }

        if ( (res = checkData( device_address )) == -1 )
        {
            return -IV_ERROR_UNREACHABLE;
        }
        else if ( res )
        {
            return -IV_ERROR_DATA_AVAILABLE;
        }
    }

    res = writeSetupData4( device_address );

    if ( res == IV_ERROR_SUCCESS && !is_password_protection() )
    {
        //res = ModuleTable::clearRecords(device_address);
    }

    return res;
}

int Channel4::writeSetupData4( unsigned char device_address )
{
    int i;
    int res = IV_ERROR_SUCCESS;

    // BlockTransfer durchfuehren.
    controller_segment &controllerSegmentCurrent = get_controller_segment( device_address );

    //if (controllerSegmentCurrent.blockTransferState == ControllerSegment::Inactive)
    //{    

    //@@: Uebergangsweise zur 04.01.01 release
    //unsigned char *buf = nullptr;
    unsigned char countTable;
    unsigned char countField;
    unsigned int sizeAccess;

    //////////////////////////////
    // Vorbereitung (Berechnung von Groessen und CRC)
    unsigned int sizeSetup = GetFlashSetupSize();

    // Vorhandene zur moeglichen Groesse pruefen
    // 	if (sizeSetup > (EEP_SIZE_SU + sizeof(StructConfig) + sizeof(dummyConfig) + FLASH_SIZE_SU)) {
    // 
    // 		return -IV_ERROR_FILE_FORMAT;
    // 	}

    unsigned int sizeEK = GetEntrance2EKFlashSize();
    // Speicher reservieren
    std::vector<unsigned char> buf( sizeEK, 0 );
    std::fill( buf.begin(), buf.end(), 0 );

    //@@: Uebergangsweise zur 04.01.01 release
    // Zeiger auf Eingabeketten ohne FLASH_KONFIG_START
    // und die Zeiger der Eingabeketten mit einem Offset versehen.
    PrepareEntrance2EK( sizeSetup, &buf[0] );

    // Groesse der Flashdaten in EepMem eintragen
    m_file.m_flash_header.FLASH_SIZE_SUS = m_file.endian_reverse( sizeSetup );

    // Anfangsadresse der Zutrittseingabeketten eintragen
    m_file.m_flash_header.IC_ACCESS = m_file.endian_reverse( sizeSetup );

    // Anfangsadresse der Zutrittskontrolllisten eintragen
    m_file.m_flash_header.LIST_ACCESS = m_file.endian_reverse( sizeSetup + sizeEK );

    // Groesse der MDE - Eingabeketten in EepMem eintragen.
    m_file.m_flash_header.MDE_EK_SIZE = GetMDEEKFlashSize();

    // Groesse der Zutrittskontroll - Eingabeketten in EepMem eintragen.
    m_file.m_flash_header.ENTRANCE_EK_SIZE = m_file.endian_reverse( sizeEK );

    // Ggf. gesetztes Streaming deaktivieren.
    set_stream_transfer( false );
    clear_stream_transfer();

    //
    unsigned int address = 0;
    unsigned int size = 0;
    unsigned int crc = 0;

    // Segmentbegin: Bedienung stoppen
    if ( (res = enter_write_data_segment( device_address, IV_VALUE_SEGMENT_UPDATE, address, size, crc )) != IV_ERROR_SUCCESS )
    {
        return res;
    }
    trace( "%s(%i):INFO(1): %s [ beginSetSegment(%i): %i, %i, 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_UPDATE, address, size, crc );

    // Pruefen ob die Uebertragung per Stream moeglich ist.
    if ( (res = write_stream( device_address, Stream_type::Update, nullptr, 0 )) == IV_ERROR_SUCCESS )
    {
        set_stream_transfer( true );
    }

    ///////////////
    // Grunddaten
    // Groesse darf nicht veraendert werden!!!
    // Berechnung 0x80 - sizeof(StructConfig)
    std::array<unsigned char, 0x1C> dummyConfig;
    std::fill( dummyConfig.begin(), dummyConfig.end(), 0 );

    // CRC berechnen
    crc = 0xFFFFFFFF;
    crc = get_crc32( m_file.m_eep_mem.data(), m_file.m_eep_mem.size(), crc, false );
    crc = get_crc32( m_file.m_config_mem.data(), m_file.m_config_mem.size(), crc, false );
    crc = get_crc32( &dummyConfig[0], dummyConfig.size(), crc, false );
    crc = get_crc32( m_file.m_flash_mem.data(), sizeSetup, crc, false );
    //@@: Uebergangsweise zur 04.01.01 release
    crc = get_crc32( &buf[0], sizeEK, crc, false );

    ///////////////
    // ZK II

    // Anzahl der Zutrittslisten ermitteln
    for ( i = 0, countTable = 0; countTable < (sizeof( m_entrance2Pointer ) / sizeof( unsigned int )); ++countTable )
    {
        if ( m_entrance2Pointer[countTable] == 0 )
        {
            break;
        }

        // Gesamte Feldanzahl fuer Groeszenberechnung sammeln.
        i += m_entrance2Table[countTable].count;
    }
    sizeAccess = sizeof(m_entrance2Pointer);
    sizeAccess += countTable * sizeof(StructTable2);
    sizeAccess += i * sizeof(StructTable2Field);

    // CRC der Zutrittslisten ermitteln
    crc = get_crc32( reinterpret_cast<unsigned char *>( m_entrance2Pointer ), sizeof( m_entrance2Pointer ), crc, false );

    for ( i = 0; i < countTable; ++i )
    {
        crc = get_crc32( reinterpret_cast<unsigned char *>( &m_entrance2Table[i] ), sizeof( StructTable2 ), crc, false );

        for ( countField = 0; countField < m_entrance2Table[i].count; ++countField )
        {
            // Beim letzten Feld der letzten Tabelle musz der CRC abgeschlossen werden.
            crc = get_crc32( reinterpret_cast<unsigned char *>( &m_entrance2TableField[i][countField] ), sizeof( StructTable2Field ), crc, 
                ((i == (countTable - 1)) && (countField == (m_entrance2Table[i].count - 1))) );
        }
    }

    //////////////////////////////
    // Begin
    //	double now = (double)clock() / (double)CLOCKS_PER_SEC;
    //	int nr = 0;
    //	Trace("Elapse (%i): %f\n", nr++, ((double)clock() / (double)CLOCKS_PER_SEC) - now);

    // Gesamtgroesze berechnen
    size = EEP_SIZE_SU + sizeof( StructConfig ) + static_cast<unsigned int>( dummyConfig.size() ) + sizeSetup + sizeEK + sizeAccess;

    // Wenn kein Streaming moeglich, dann mit inneren Bloecken uebertragen.
    if ( !is_stream_transfer() )
    {
        // Segmentbeginn: Setup uebertragen
        if ( (res = enter_write_data_segment( device_address, IV_VALUE_SEGMENT_SETUP, address, size, crc )) != IV_ERROR_SUCCESS )
        {
            // Zu grosse Setupdatei, es ist nicht genug speicher verfuegbar.
            if ( res == 3 )
            {
                res = -IV_ERROR_NOT_ENOUGH_FLASH;
            }

            return res;
        }

        trace( "%s(%i):INFO(1): %s [ beginSetSegment(%i): %i, %i, 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_SETUP, address, size, crc );
    }
    //}
    //else
    {
        // Zwischengepeicherte Daten fuer Wiederaufnahme setzen.

    }

    //////////////////////////////
    // Grunddaten

    // EEPROM uebertragen
    if ( (res = write_to_flash( device_address, address, m_file.m_eep_mem.data(), m_file.m_eep_mem.size() )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // CONFIG uebertragen (100 Byte)
    address += EEP_SIZE_SU;
    if ( (res = write_to_flash( device_address, address, m_file.m_config_mem.data(), m_file.m_config_mem.size()) ) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // CONFIG dummyConfig (28 Byte)
    address += sizeof( StructConfig );
    if ( (res = write_to_flash( device_address, address, &dummyConfig[0], static_cast<unsigned int>( dummyConfig.size() )) ) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // FLASH uebertragen
    address += static_cast<unsigned int>( dummyConfig.size() );
    if ( (res = write_to_flash( device_address, address, m_file.m_flash_mem.data(), sizeSetup )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    //@@: Uebergangsweise zur 04.01.01 release
    //////////////////////////////
    // Pointerarray + Strukturen der Zutrittslisten uebertragen

    // Eingabekette der Zutrittskontrolle
    address += sizeSetup;
    if ( (res = write_to_flash( device_address, address, &buf[0], sizeEK )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    address += sizeEK;
    // Pointerarray uebertragen
    if ( (res = write_to_flash( device_address, address, reinterpret_cast<unsigned char *>( m_entrance2Pointer ), sizeof( m_entrance2Pointer )) ) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    address += sizeof( m_entrance2Pointer );

    // Strukturen uebertragen
    for ( i = 0; i < countTable; ++i )
    {
        // Tabellenstruktur uebertragen
        if ( (res = write_to_flash( device_address, address, reinterpret_cast<unsigned char *>( &m_entrance2Table[i] ), sizeof( StructTable2 )) ) != IV_ERROR_SUCCESS )
        {
            return res;
        }

        address += sizeof( StructTable2 );

        // Feldstrukturen uebertragen
        if ( (res = write_to_flash( device_address, address, reinterpret_cast<unsigned char *>( m_entrance2TableField[i] ), m_entrance2Table[i].count * sizeof( StructTable2Field )) ) != IV_ERROR_SUCCESS )
        {
            return res;
        }

        address += m_entrance2Table[i].count * sizeof( StructTable2Field );
    }

    // Im Modus Streaming ?, dann gepufferte Daten senden.
    if ( is_stream_transfer() && m_stream_buffer.size() )
    {
        if ( (res = write_stream( device_address, Stream_type::Setup, m_stream_buffer.data(), (uint32_t)m_stream_buffer.size() ) ) != IV_ERROR_SUCCESS )
        {
            // Gepufferte Daten loeschen.
            clear_stream_transfer();
            // Zu grosse Setupdatei, es ist nicht genug speicher verfuegbar.
            if ( res == 3 )
            {
                res = -IV_ERROR_NOT_ENOUGH_FLASH;
            }

            return res;
        }

        // Gepufferte Daten loeschen.
        clear_stream_transfer();
    }

    // Wenn kein Streaming moeglich, dann mit inneren Bloecken uebertragen.
    if ( !is_stream_transfer() )
    {
        //////////////////////////////
        // Ende
        // Segmentend: Setup uebertragen
        if ( (res = exit_write_data_segment( device_address, IV_VALUE_SEGMENT_SETUP, crc )) != IV_ERROR_SUCCESS )
        {
            return res;
        }

        trace( "%s(%i):INFO(1): %s [ endSetSegment(%i): 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_SETUP, crc );
    }

    // Extendedstruktur lesen
    StructExtendedFlash extBegin = *reinterpret_cast<StructExtendedFlash *>( m_file.m_flashExtended );

    ///////////////
    // Extended - Speicher Vorbereitung

    // Zeigeroffsets anpassen (substrahieren von Offset)
    AdjustExtendedMemPointer();

    // Groesse MDE und dessen Eingabeketten
    size = extBegin.offsetFreeScope;

    // CRC berechnen
    crc = 0xFFFFFFFF;
    // IOModul - Konfiguration
    crc = get_crc32( m_file.m_flashExtended, size, crc );

    // Wenn kein Streaming moeglich, dann mit inneren Bloecken uebertragen.
    if ( !is_stream_transfer() )
    {
        //////////////////////////////
        // Begin

        // Segmentbegin: Setup uebertragen
        if ( (res = enter_write_data_segment( device_address, IV_VALUE_SEGMENT_SETUP_EXTENDED, address, size, crc )) != IV_ERROR_SUCCESS )
        {
            if ( (res & 0xFF) == 1 )
            {
                // Funktionsfehler Typ IV_VALUE_SEGMENT_SETUP_EXTENDED unbekannt.
            }
            else
            {
                return res;
            }
        }

        trace( "%s(%i):INFO(1): %s [ (%i)beginSetSegment(%i): %i, %i, 0x%08X ]\n", __FILE__, __LINE__, __func__, res, IV_VALUE_SEGMENT_SETUP_EXTENDED, address, size, crc );
    }

    if ( res == IV_ERROR_SUCCESS )
    {
        //////////////////////////////
        // Extended - Speicher Uebertragen

        if ( (res = write_to_flash( device_address, address, m_file.m_flashExtended, size )) != IV_ERROR_SUCCESS )
        {
            return res;
        }

        address += size;

        // Im Modus Streaming ?, dann gepufferte Daten senden.
        if ( is_stream_transfer() && m_stream_buffer.size() )
        {
            if ( (res = write_stream( device_address, Stream_type::Setup_extended, m_stream_buffer.data(), (uint32_t)m_stream_buffer.size() )) != IV_ERROR_SUCCESS )
            {
                // Gepufferte Daten loeschen.
                clear_stream_transfer();
                return res;
            }

            // Gepufferte Daten loeschen.
            clear_stream_transfer();
        }

        // Wenn kein Streaming moeglich, dann mit inneren Bloecken uebertragen.
        if ( !is_stream_transfer() )
        {
            //////////////////////////////
            // Ende

            // Segmentend: Setup uebertragen
            if ( (res = exit_write_data_segment( device_address, IV_VALUE_SEGMENT_SETUP_EXTENDED, crc )) != IV_ERROR_SUCCESS )
            {
                return res;
            }

            trace( "%s(%i):INFO(1): %s [ endSetSegment(%i): 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_SETUP_EXTENDED, crc );
        }
    }

    // Zeigeroffsets anpassen (addieren von Offset)
    AdjustExtendedMemPointer();

    // Pause wegen Verarbeitungsporblemen bis 4.1.5
    std::this_thread::sleep_for( std::chrono::seconds( 1 ) );

    crc = 0;
    // Segmentend: Bedienung stoppen
    if ( (res = exit_write_data_segment( device_address, IV_VALUE_SEGMENT_UPDATE, crc )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    trace( "%s(%i):INFO(1): %s [ endSetSegment(%i): 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_UPDATE, crc );


    // Pruefen und setzen des Status von Passwortschutz.
    write_setup_password( device_address, reinterpret_cast<const char *>( "ISPW" ) );
    ////////////////////////////////////////////////////////////////////
    // Der folgende Teil ist nur bei Versionen unter 04.01.07 noetig.
    // Ab dieser Version ist auch der Passwortschutz moeglich.
    // Es ist hier nicht moeglich, ein Setup mit einem gesetzten Passwort zu uebertragen 
    // und darauffolgend gleich weitere Funktionen aufrufen, weil ansonsten der
    // Schluessel fuer die Passwortverschluesselung offen gelegt werden muesste.
    if ( !is_password_protection() )
    {
        // GV setzen?
        if ( extBegin.offsetGlobalVariableValues )
        {
            StructGlobVar *globVar = reinterpret_cast<StructGlobVar *>( m_file.m_flash_mem.data() );
            char name[64];
            memset( name, 0, sizeof( name ) );

            // Struktur lesen
            StructValueGVx ValueGVx = *reinterpret_cast<StructValueGVx*>( &m_file.m_flashExtended[extBegin.offsetGlobalVariableValues] );

            for ( i = 0; i < 8; ++i )
            {
                if ( ValueGVx.ucFlags & (1 << i) )
                {
                    std::string name( "SETUP." );
                    name += globVar->SFeld[i];
                    if ( (res = write_variable( device_address, name.c_str(), (char*)&ValueGVx.szValue[i] )) != IV_ERROR_SUCCESS )
                    {
                        // Das Befehlsbyte wird bei unbekannter Variabler nicht veraendert
                        // und wenn der Wert nicht gesetzt werden konnte wird das hoechste bit gesetzt.
                        if ( (res & 0xFF) == 3 )
                        {
                            // Fehler liegt bei Wert
                            //res = -IV_ERROR_VARIABLE_VALUE;
                        }
                        else if ( (res & 0xFF) == 2 )
                        {
                            // Variable nicht bekannt
                            //res = -IV_ERROR_VARIABLE_UNDEFINED;
                        }

                        return res;
                    }
                }
            }
        }

        // Standardwerte fuer Systemvariable
        //############################
        // 2005-09-19: Erweiterte Struktur fuer Parameter
        auto setup_version = m_file.endian_reverse( m_file.m_flash_header.SETUP_VERSION );
        if ( ((setup_version & 0xFFFF0000) == 0x03010000 && (setup_version & 0x0000FFFF) >= 0x00003406) 
            || ((setup_version & 0xFFFF0000) == 0x04010000 && (setup_version & 0x0000FFFF) > 0x00000002) )
        {
            StructExtendedFlash extBegin = *(StructExtendedFlash*)m_file.m_flashExtended;

            // Liegt die Struktur vor?
            if ( extBegin.offsetExtendedStruct )
            {
                StructExtended extParam = *(StructExtended*)&m_file.m_flashExtended[extBegin.offsetExtendedStruct];

                char val[32];
                char name[64];
                // Buchungswiederholsperrdauer
                sprintf( name, "_SERVICE.DBLINPUT" );
                if ( extParam.doubleInputTime > 65499 )
                {
                    extParam.doubleInputTime = 65499;
                }
                sprintf( val, "%i", extParam.doubleInputTime );
                // 
                if ( (res = write_system_variable( device_address, name, val )) != IV_ERROR_SUCCESS )
                {
                    // Das Befehlsbyte wird bei unbekannter Variabler nicht veraendert
                    // und wenn der Wert nicht gesetzt werden konnte wird das hoechste bit gesetzt.
                    if ( (res & 0xFF) == 3 )
                    {
                        // Fehler liegt bei Wert
                        //res = -IV_ERROR_VARIABLE_VALUE;
                        res = IV_ERROR_SUCCESS;
                    }
                    else if ( (res & 0xFF) == 2 )
                    {
                        // Variable nicht bekannt
                        //res = -IV_ERROR_VARIABLE_UNDEFINED;
                        res = IV_ERROR_SUCCESS;
                    }
                }
                // Abblendedauer fuer Bestaetigungsfeld
                sprintf( name, "_SERVICE.DEMANDTIMER" );
                if ( extParam.confirmationTime > 255 )
                {
                    extParam.confirmationTime = 255;
                }
                sprintf( val, "%i", extParam.confirmationTime );
                // 
                if ( (res = write_system_variable( device_address, name, val )) != IV_ERROR_SUCCESS )
                {
                    // Das Befehlsbyte wird bei unbekannter Variabler nicht veraendert
                    // und wenn der Wert nicht gesetzt werden konnte wird das hoechste bit gesetzt.
                    if ( (res & 0xFF) == 3 )
                    {
                        // Fehler liegt bei Wert
                        //res = -IV_ERROR_VARIABLE_VALUE;
                        res = IV_ERROR_SUCCESS;
                    }
                    else if ( (res & 0xFF) == 2 )
                    {
                        // Variable nicht bekannt
                        //res = -IV_ERROR_VARIABLE_UNDEFINED;
                        res = IV_ERROR_SUCCESS;
                    }
                }

                // Tiefentladungsschutz
                sprintf( name, "POWER.LOW" );
                sprintf( val, "%i", extParam.energieOption.powerLow );
                // 
                if ( (res = write_system_variable( device_address, name, val )) != IV_ERROR_SUCCESS )
                {
                    // Das Befehlsbyte wird bei unbekannter Variabler nicht veraendert
                    // und wenn der Wert nicht gesetzt werden konnte wird das hoechste bit gesetzt.
                    if ( (res & 0xFF) == 3 )
                    {
                        // Fehler liegt bei Wert
                        //res = -IV_ERROR_VARIABLE_VALUE;
                        res = IV_ERROR_SUCCESS;
                    }
                    else if ( (res & 0xFF) == 2 )
                    {
                        // Variablenname nicht bekannt
                        //res = -IV_ERROR_VARIABLE_UNDEFINED;
                        res = IV_ERROR_SUCCESS;
                    }
                    else if ( (res & 0xFF) == 1 )
                    {
                        // Majorbestandteil unbekannt
                        //res = -IV_ERROR_VARIABLE_UNDEFINED;
                        res = IV_ERROR_SUCCESS;
                    }
                }
            }
        }
    }

    return res;
}

int Channel4::writeSetupData4_2( unsigned char device_address )
{
    int res = IV_ERROR_SUCCESS;
    unsigned char dummyConfig[0x1C]; // Groesse darf nicht veraendert werden!!!
    // Berechnung 0x80 - sizeof(StructConfig)

    //////////////////////////////
    // Vorbereitung (Berechnung von Groessen und CRC)
    unsigned int sizeSetup = GetFlashSetupSize();

    // Extendedstruktur
    StructExtendedFlash extBegin = *(StructExtendedFlash*)m_file.m_flashExtended;

    // Groesse der Flashdaten in EepMem eintragen
    m_file.m_flash_header.FLASH_SIZE_SUS = m_file.endian_reverse( sizeSetup );

    // Anfangsadresse der Zutrittseingabeketten eintragen
    m_file.m_flash_header.IC_ACCESS = m_file.endian_reverse( static_cast<unsigned int>( 0 ) );

    // Anfangsadresse der Zutrittskontrolllisten eintragen
    m_file.m_flash_header.LIST_ACCESS = m_file.endian_reverse( static_cast<unsigned int>( 0 ) );

    // Groesse der MDE - Eingabeketten in EepMem eintragen.
    m_file.m_flash_header.MDE_EK_SIZE = GetMDEEKFlashSize();

    // Groesse der Zutrittskontroll - Eingabeketten in EepMem eintragen.
    m_file.m_flash_header.ENTRANCE_EK_SIZE = m_file.endian_reverse( static_cast<unsigned int>( 0 ) );

    // Groesse der Headerstruktur in EepMem eintragen.
    m_file.m_flash_header.HEADER_SIZE = m_file.m_file_header.size;

    // Groesse des Erweiterten Speicherbereichs eintragen.
    m_file.m_flash_header.EXTENDED_SIZE = extBegin.offsetFreeScope;

    // Ggf. gesetztes Streaming deaktivieren.
    set_stream_transfer( false );
    clear_stream_transfer();

    //
    unsigned int address, size, crc;
    address = size = crc = 0;

    // Segmentbegin: Bedienung stoppen
    if ((res = enter_write_data_segment(device_address, IV_VALUE_SEGMENT_UPDATE, address, size, crc)) != IV_ERROR_SUCCESS)
    {
        return res;
    }
    trace("%s(%i):INFO(1): %s [ beginSetSegment(%i): %i, %i, 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_UPDATE, address, size, crc);

    // Es wird ab der Version 04.02.04 ein MD5 ueber den Dateiheader die Listen- und Datensatzbeschreibung als Systemvariable abgelegt.
    // Vor der Uebertragung des Setups werden diese hier geloescht, falls verfuegbar.
    bool md5Sysvars = false;

    // Systemvariablen fuer MD5-Pruefwerte loeschen.
    if ((res = write_system_variable(device_address, (char*)"SYSTEM.TMP1", (char*)"")) != IV_ERROR_SUCCESS) {

        // Wenn der Fehler nicht Major oder Minorbestandteil unbekannt ist, dann hier bereits abbrechen.
        if (res != 1 && res != 2)
            return res;
    } else if ((res = write_system_variable(device_address, (char*)"SYSTEM.TMP2", (char*)"")) != IV_ERROR_SUCCESS) {

        // Wenn der Fehler nicht Major oder Minorbestandteil unbekannt ist, dann hier bereits abbrechen.
        if (res != 1 && res != 2)
            return res;
    } else if ((res = write_system_variable(device_address, (char*)"SYSTEM.TMP3", (char*)"")) != IV_ERROR_SUCCESS) {

        // Wenn der Fehler nicht Major oder Minorbestandteil unbekannt ist, dann hier bereits abbrechen.
        if (res != 1 && res != 2)
            return res;
    } else if ((res = write_system_variable(device_address, (char*)"SYSTEM.TMP4", (char*)"")) != IV_ERROR_SUCCESS) {

        // Wenn der Fehler nicht Major oder Minorbestandteil unbekannt ist, dann hier bereits abbrechen.
        if (res != 1 && res != 2)
            return res;
    } else {

        // Die Systemvariablen wurde beide erfolgreich geleert.
        md5Sysvars = true;
    }

    // Pruefen ob die Uebertragung per Stream moeglich ist.
    if ( (res = write_stream( device_address, Stream_type::Update, nullptr, 0 )) == IV_ERROR_SUCCESS )
    {
        set_stream_transfer( true );
    }

    ///////////////
    // Grunddaten
    memset(dummyConfig, 0, sizeof(dummyConfig));

    // CRC berechnen
    crc = 0xFFFFFFFF;
    crc = get_crc32(m_file.m_eep_mem.data(), m_file.m_eep_mem.size(), crc, false);
    crc = get_crc32(m_file.m_config_mem.data(), m_file.m_config_mem.size(), crc, false);
    crc = get_crc32(dummyConfig, sizeof(dummyConfig), crc, false);
    crc = get_crc32(m_file.m_flash_mem.data(), sizeSetup, crc, false);
    crc = get_crc32(m_file.m_file_header_mem.data(), m_file.m_file_header.size, crc);

    //////////////////////////////
    // Begin
    //	double now = (double)clock() / (double)CLOCKS_PER_SEC;
    //	int nr = 0;
    //	Trace("Elapse (%i): %f\n", nr++, ((double)clock() / (double)CLOCKS_PER_SEC) - now);

    // Gesamtgroesze berechnen
    // In dem Block des Setups werden verschiedene Einzelbloecke zusammengefasst uebertragen.
    size = EEP_SIZE_SU + sizeof(StructConfig) + sizeof(dummyConfig) + sizeSetup + m_file.m_file_header.size;

    // Wenn kein Streaming moeglich, dann mit inneren Bloecken uebertragen.
    if (!is_stream_transfer()) {
        // Segmentbeginn: Setup uebertragen
        if ((res = enter_write_data_segment(device_address, IV_VALUE_SEGMENT_SETUP, address, size, crc)) != IV_ERROR_SUCCESS)
        {
            // Zu grosse Setupdatei, es ist nicht genug speicher verfuegbar.
            if (res == 3) {
                res = -IV_ERROR_NOT_ENOUGH_FLASH;
            }
            return res;
        }
        trace("%s(%i):INFO(1): %s [ beginSetSegment(%i): %i, %i, 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_SETUP, address, size, crc);
    }

    //////////////////////////////
    // Grunddaten

    // EEPROM uebertragen
    if ((res = write_to_flash(device_address, address, m_file.m_eep_mem.data(), m_file.m_eep_mem.size())) != IV_ERROR_SUCCESS)
    {
        return res;
    }

    // CONFIG uebertragen (100 Byte)
    address += EEP_SIZE_SU;
    if ((res = write_to_flash(device_address, address, m_file.m_config_mem.data(), m_file.m_config_mem.size())) != IV_ERROR_SUCCESS)
    {
        return res;
    }

    // CONFIG dummyConfig (28 Byte)
    address += sizeof(StructConfig);
    if ((res = write_to_flash(device_address, address, dummyConfig, sizeof(dummyConfig))) != IV_ERROR_SUCCESS)
    {
        return res;
    }

    // FLASH uebertragen
    address += sizeof(dummyConfig);
    if ((res = write_to_flash(device_address, address, m_file.m_flash_mem.data(), sizeSetup)) != IV_ERROR_SUCCESS)
    {
        return res;
    }

    // Header der Datei uebertragen
    address += sizeSetup;
    if ((res = write_to_flash(device_address, address, m_file.m_file_header_mem.data(), m_file.m_file_header.size)) != IV_ERROR_SUCCESS)
    {
        return res;
    }

    // Im Modus Streaming ?, dann gepufferte Daten senden.
    if (is_stream_transfer() && m_stream_buffer.size()) {
        if ((res = write_stream(device_address, Stream_type::Setup, m_stream_buffer.data(), m_stream_buffer.size() )) != IV_ERROR_SUCCESS) {
            // Gepufferte Daten loeschen.
            clear_stream_transfer();
            return res;
        }
        // Gepufferte Daten loeschen.
        clear_stream_transfer();
    }

    // Wenn kein Streaming moeglich, dann mit inneren Bloecken uebertragen.
    if (!is_stream_transfer()) {

        //////////////////////////////
        // Ende
        // Segmentend: Setup uebertragen
        if ((res = exit_write_data_segment(device_address, IV_VALUE_SEGMENT_SETUP, crc)) != IV_ERROR_SUCCESS)
        {
            return res;
        }
        trace("%s(%i):INFO(1): %s [ endSetSegment(%i): 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_SETUP, crc);
    }

    ///////////////
    // Extended - Speicher Vorbereitung

    // Zeigeroffsets anpassen (substrahieren von Offset)
    AdjustExtendedMemPointer();

    // Groesse MDE und dessen Eingabeketten
    size = extBegin.offsetFreeScope;

    // CRC berechnen
    crc = 0xFFFFFFFF;
    // IOModul - Konfiguration
    crc = get_crc32(m_file.m_flashExtended, size, crc);

    // Wenn kein Streaming moeglich, dann mit inneren Bloecken uebertragen.
    if (!is_stream_transfer()) {
        //////////////////////////////
        // Begin

        // Segmentbegin: Setup uebertragen
        if ((res = enter_write_data_segment(device_address, IV_VALUE_SEGMENT_SETUP_EXTENDED, address, size, crc)) != IV_ERROR_SUCCESS)
        {
            // Zu grosse Setupdatei, es ist nicht genug speicher verfuegbar.
            if (res == 3) {
                res = -IV_ERROR_NOT_ENOUGH_FLASH;
            }

            return res;
        }
        trace("%s(%i):INFO(1): %s [ (%i)beginSetSegment(%i): %i, %i, 0x%08X ]\n", __FILE__, __LINE__, __func__, res, IV_VALUE_SEGMENT_SETUP_EXTENDED, address, size, crc);
    }

    if (res == IV_ERROR_SUCCESS)
    {
        //////////////////////////////
        // Extended - Speicher Uebertragen

        if ((res = write_to_flash(device_address, address, m_file.m_flashExtended, size)) != IV_ERROR_SUCCESS)
        {
            return res;
        }
        address += size;

        // Im Modus Streaming ?, dann gepufferte Daten senden.
        if (is_stream_transfer() && m_stream_buffer.size()) {
            if ((res = write_stream(device_address, Stream_type::Setup_extended, m_stream_buffer.data(), static_cast<unsigned int>( m_stream_buffer.size() ))) != IV_ERROR_SUCCESS) {
                // Gepufferte Daten loeschen.
                clear_stream_transfer();
                return res;
            }
            // Gepufferte Daten loeschen.
            clear_stream_transfer();
        }

        // Wenn kein Streaming moeglich, dann mit inneren Bloecken uebertragen.
        if (!is_stream_transfer()) {
            //////////////////////////////
            // Ende

            // Segmentend: Setup uebertragen
            if ((res = exit_write_data_segment(device_address, IV_VALUE_SEGMENT_SETUP_EXTENDED, crc)) != IV_ERROR_SUCCESS)
            {
                return res;
            }
            trace("%s(%i):INFO(1): %s [ endSetSegment(%i): 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_SETUP_EXTENDED, crc);
        }
    }

    // Zeigeroffsets anpassen (addieren von Offset)
    AdjustExtendedMemPointer();

    // Pause wegen Verarbeitungsporblemen bis 4.1.5
    std::this_thread::sleep_for( std::chrono::seconds( 1 ) );

    // Werden die Systemvariablen fuer die MD5 unterstuetzt?
    if (md5Sysvars) {

        char *md5Tables = m_file.GetMD5Header(); // Dateiheader
        // Systemvariablen fuer MD5 loeschen.
        if (md5Tables && (res = write_system_variable(device_address, (char*)"SYSTEM.TMP1", md5Tables)) != IV_ERROR_SUCCESS) {

            // Wenn der Fehler nicht Major oder Minorbestandteil unbekannt ist, dann hier bereits abbrechen.
            if (res != 1 && res != 2)
                return res;
        } 

        md5Tables = m_file.GetMD5Tables(true); // Listenbeschreibungen
        if (md5Tables && (res = write_system_variable(device_address, (char*)"SYSTEM.TMP2", md5Tables)) != IV_ERROR_SUCCESS) {

            // Wenn der Fehler nicht Major oder Minorbestandteil unbekannt ist, dann hier bereits abbrechen.
            if (res != 1 && res != 2)
                return res;
        }

        md5Tables = m_file.GetMD5Tables(false); // Datensatzbeschreibungen
        if (md5Tables && (res = write_system_variable(device_address, (char*)"SYSTEM.TMP3", md5Tables)) != IV_ERROR_SUCCESS) {

            // Wenn der Fehler nicht Major oder Minorbestandteil unbekannt ist, dann hier bereits abbrechen.
            if (res != 1 && res != 2)
                return res;
        }
    }

    crc = 0;
    // Segmentend: Bedienung stoppen
    if ((res = exit_write_data_segment(device_address, IV_VALUE_SEGMENT_UPDATE, crc)) != IV_ERROR_SUCCESS)
    {
        return res;
    }
    trace("%s(%i):INFO(1): %s [ endSetSegment(%i): 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_UPDATE, crc);


    // Pruefen und setzen des Status von Passwortschutz.
    write_setup_password(device_address, (char*)"ISPW");

    return res;
}

int Channel4::readSetupWithTimeboySlot( unsigned char device_address, const std::string &file_path )
{
    int res;
    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        //char val[80];
        //// Systemvariablen fuer MD5-Pruefwerte ermitteln.
        //// Wenn die Systemvariable definiert ist und einen gueltigen MD5 beinhaltet, 
        //// dann wird eine andere Art der Setupuebertragung verwendent.
        //if ((res = getSystemVariable(device_address, "SYSTEM.TMP1", val)) == IV_ERROR_SUCCESS && IsMD5(val)) {

        //    // Befehlsausfuehrung starten.
        //    res = readSetupData4_2(device_address);
        //} else {

        //    // Bei einem Kommunikationsfehler, abbrechen.
        //    if (res < IV_ERROR_SUCCESS)
        //        return res;

        //    // Befehlsausfuehrung starten.
        res = readSetupData( device_address, file_path );
        //}


        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

int Channel4::readSetupData( unsigned char device_address, const std::string &file_path )
{
    int res = 0;

    std::string ext = boost::to_lower_copy( file_path.substr( file_path.size() - 4 ) );
    if ( ext != ".aes" && ext != ".ael" )
    {
        // Die Anwendung kann ueber den gewaehlten Weg, nur Dateien vom Typ (*.aes; *.ael) oeffnen.
        return -IV_ERROR_FILE_EXTENSION;
    }

    // Speicher initialisieren
    m_file.SetDefault();

    std::string version;
    if ( (res = read_firmware_version(device_address, version)) != IV_ERROR_SUCCESS)
    {
        return res;
    }

    // Pruefen ob die Firmwareversion unterstuetzt wird.
    if ( !m_file.IsSupportedFirmwareVersion( version.c_str() ) )
    {
        return -IV_ERROR_VERSION_FIRMWARE;
    }

    // Setupdatei nach erfolgreichem Auslesen, abspeichern.
    if ( (res = readSetupData4( device_address )) == IV_ERROR_SUCCESS )
    {
        // Diese Version der Firmware der Setupdatei als Mindestversion vorgeben.
        SoftwarebuildVersion fwVersion( version );
        m_file.NeedFirmwareVersion( fwVersion );

        int error;
        if ( !m_file.save( file_path, &error ) )
        {
            res = -IV_ERROR_FILE_OPEN;
        }
    }

    return res;
}

int Channel4::readSetupData4( unsigned char device_address )
{
    int i;
    int res = IV_ERROR_SUCCESS;
    unsigned char dummyConfig[0x1C]; // Groesse nicht veraendern !!!
    // Berechnet sich aus 0x80 - sizeof(StructConfig)

    unsigned char countTable;
    unsigned char countField;
    unsigned int sizeAccess;

    //
    unsigned int address = 0;
    unsigned int size = 0;
    unsigned int crc = 0;
    unsigned int crcTest = 0;

    // Segmentbegin: Bedienung stoppen
    if ((res = enter_read_data_segment(device_address, IV_VALUE_SEGMENT_GET_UPDATE_BG, address, size, crc)) != IV_ERROR_SUCCESS)
    {
        if ((res & 0xFF) == 1)
        {
            // Funktionsfehler Typ: IV_VALUE_SEGMENT_GET_UPDATE_BG unbekannt.
            // Wenn dass der Fall ist, wird das Auslesen noch nicht unterstuetzt.
            res = -IV_ERROR_FUNCTION_UNAVAILABLE;
        }
        else
        {
            return res;
        }
    }
    trace("%s(%i):INFO(1): %s [ beginGetSegment(%i): %i, %i, 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_UPDATE_BG, address, size, crc);

    //////////////////////////////
    // Begin

    // Segmentbegin: Setup uebertragen
    if ((res = enter_read_data_segment(device_address, IV_VALUE_SEGMENT_GET_SETUP, address, size, crc)) != IV_ERROR_SUCCESS)
    {
        return res;
    }
    trace("%s(%i):INFO(1): %s [ beginGetSegment(%i): %i, %i, 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_SETUP, address, size, crc);

    // Vorhandene zur moeglichen Groesse pruefen
    if ( size < (EEP_SIZE_SU + (sizeof( StructConfig ) + sizeof( dummyConfig ))) )
    {
        return -IV_ERROR_PROCESSING;
    }

    //////////////////////////////
    // Grunddaten

    // EEPROM uebertragen
    if ((res = read_from_flash(device_address, address, m_file.m_eep_mem.data(), m_file.m_eep_mem.size())) != IV_ERROR_SUCCESS)
    {
        return res;
    }
    address += EEP_SIZE_SU;
    // CONFIG uebertragen (100 Byte)
    if ((res = read_from_flash(device_address, address, m_file.m_config_mem.data(), m_file.m_config_mem.size())) != IV_ERROR_SUCCESS)
    {
        return res;
    }
    address += sizeof(StructConfig);
    // CONFIG dummybytes (28 Byte)
    if ((res = read_from_flash(device_address, address, dummyConfig, sizeof(dummyConfig))) != IV_ERROR_SUCCESS)
    {
        return res;
    }
    address += sizeof(dummyConfig);

    // Groesse der Flashdaten fuer Setup aus EepMem entnehmen
    unsigned int sizeSetup = m_file.endian_reverse( m_file.m_flash_header.FLASH_SIZE_SUS );

    // FLASH uebertragen
    if ((res = read_from_flash(device_address, address, m_file.m_flash_mem.data(), sizeSetup)) != IV_ERROR_SUCCESS)
    {
        return res;
    }
    address += sizeSetup;


    // Groesse der Zutrittskontroll - Eingabeketten aus EepMem entnehmen.
    unsigned int sizeEK = m_file.endian_reverse( m_file.m_flash_header.ENTRANCE_EK_SIZE );

    // Speicher reservieren
    std::unique_ptr<unsigned char[]> bufEK( new unsigned char[sizeEK] );
    std::fill( bufEK.get(), bufEK.get() + sizeEK, 0 );

    //////////////////////////////
    // Pointerarray + Strukturen der Zutrittslisten uebertragen

    // Eingabekette der Zutrittskontrolle
    if ((res = read_from_flash(device_address, address, bufEK.get(), sizeEK)) != IV_ERROR_SUCCESS)
    {
        return res;
    }
    address += sizeEK;

    // Pointerarray lesen
    if ((res = read_from_flash(device_address, address, (unsigned char*)m_entrance2Pointer, sizeof(m_entrance2Pointer))) != IV_ERROR_SUCCESS)
    {
        return res;
    }
    address += sizeof(m_entrance2Pointer);

    ///////////////
    // ZK II

    // Anzahl der Zutrittslisten ermitteln
    for (i=0, countTable=0; countTable < (sizeof(m_entrance2Pointer) / sizeof(unsigned int)); countTable++)
    {
        if (m_entrance2Pointer[countTable] == 0)
        {
            break;
        }

        // Gesamte Feldanzahl fuer Groeszenberechnung sammeln.
        i += m_entrance2Table[countTable].count;
    }
    sizeAccess = sizeof(m_entrance2Pointer);
    sizeAccess += countTable * sizeof(StructTable2);
    sizeAccess += i * sizeof(StructTable2Field);

    // Strukturen lesen
    for (i=0; i < countTable; i++)
    {
        // Tabellenstruktur uebertragen
        if ((res = read_from_flash(device_address, address, (unsigned char*)&m_entrance2Table[i], sizeof(StructTable2))) != IV_ERROR_SUCCESS)
        {
            return res;
        }
        address += sizeof(StructTable2);

        // Feldstrukturen lesen
        if ((res = read_from_flash(device_address, address, (unsigned char*)m_entrance2TableField[i], m_entrance2Table[i].count * sizeof(StructTable2Field))) != IV_ERROR_SUCCESS)
        {
            return res;
        }
        address += m_entrance2Table[i].count * sizeof(StructTable2Field);
    }


    ///////////////
    // Grunddaten

    // CRC berechnen
    crcTest = 0xFFFFFFFF;
    crcTest = get_crc32(m_file.m_eep_mem.data(), m_file.m_eep_mem.size(), crcTest, false);
    crcTest = get_crc32(m_file.m_config_mem.data(), m_file.m_config_mem.size(), crcTest, false);
    crcTest = get_crc32(dummyConfig, sizeof(dummyConfig), crcTest, false);
    crcTest = get_crc32(m_file.m_flash_mem.data(), sizeSetup, crcTest, false);
    crcTest = get_crc32(bufEK.get(), sizeEK, crcTest, false);

    // Speicher wieder freigeben.
    bufEK.reset( nullptr );

    // CRC der Zutrittslisten ermitteln
    crcTest = get_crc32((unsigned char*)m_entrance2Pointer, sizeof(m_entrance2Pointer), crcTest, false);

    for (i=0; i < countTable; i++)
    {
        crcTest = get_crc32((unsigned char*)&m_entrance2Table[i], sizeof(StructTable2), crcTest, false);

        for (countField=0; countField < m_entrance2Table[i].count; countField++)
        {
            // Beim letzten Feld der letzten Tabelle musz der CRC abgeschlossen werden.
            crcTest = get_crc32((unsigned char*)&m_entrance2TableField[i][countField], sizeof(StructTable2Field), crcTest, ((i == (countTable - 1)) && (countField == (m_entrance2Table[i].count - 1))));
        }
    }

    //////////////////////////////
    // Ende

    if (crc != crcTest)
    {
        trace("%s(%i):ERROR(1): %s [ CRC von Segment falsch: 0x%08X, 0x%08X ]\n", __FILE__, __LINE__, __func__, crc, crcTest);
        return -IV_ERROR_CRC_SEGMENT;
    }

    // Segmentend: Setup uebertragen
    if ((res = exit_read_data_segment(device_address, IV_VALUE_SEGMENT_GET_SETUP)) != IV_ERROR_SUCCESS)
    {
        return res;
    }
    trace("%s(%i):INFO(1): %s [ endGetSegment(%i) ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_SETUP);

    //////////////////////////////
    // Begin
    address = size = crc = 0;
    // Segmentbegin: Setup uebertragen
    if ((res = enter_read_data_segment(device_address, IV_VALUE_SEGMENT_GET_SETUP_EXTENDED, address, size, crc)) != IV_ERROR_SUCCESS)
    {
        return res;
    }
    trace("%s(%i):INFO(1): %s [ beginGetSegment(%i): %i, %i, 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_SETUP_EXTENDED, address, size, crc);

    //////////////////////////////
    // Extended - Speicher Lesen

    if ((res = read_from_flash(device_address, address, m_file.m_flashExtended, size)) != IV_ERROR_SUCCESS)
    {
        return res;
    }
    address += size;

    ///////////////
    // Extended - Speicher pruefen

    // Extended - Speicher
    crcTest = 0xFFFFFFFF;
    crcTest = get_crc32(m_file.m_flashExtended, size, crcTest);

    if (crc != crcTest)
    {
        trace("%s(%i):ERROR(1): %s [ CRC von Segment falsch: 0x%08X, 0x%08X ]\n", __FILE__, __LINE__, __func__, crc, crcTest);
        return -IV_ERROR_CRC_SEGMENT;
    }

    //////////////////////////////
    // Ende

    // Segmentend: Setup uebertragen
    if ((res = exit_read_data_segment(device_address, IV_VALUE_SEGMENT_GET_SETUP_EXTENDED)) != IV_ERROR_SUCCESS)
    {
        return res;
    }
    trace("%s(%i):INFO(1): %s [ endGetSegment(%i) ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_SETUP_EXTENDED);


    // Die Eingabekettenzeiger anpassen, da diese fuer die FW vor dem Einspielen angepasst wurden.
    // Zeigeroffsets anpassen (addieren von Offset)
    AdjustExtendedMemPointer();

    crc = 0;
    // Segmentend: Bedienung stoppen
    if ((res = exit_read_data_segment(device_address, IV_VALUE_SEGMENT_GET_UPDATE_BG)) != IV_ERROR_SUCCESS)
    {
        return res;
    }
    trace("%s(%i):INFO(1): %s [ endGetSegment(%i) ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_UPDATE_BG);

    return res;
}

int Channel4::readSetupData4_2( unsigned char device_address )
{
    int res = IV_ERROR_SUCCESS;
    unsigned char dummyConfig[0x1C]; // Groesse nicht veraendern !!!
    // Berechnet sich aus 0x80 - sizeof(StructConfig)

    unsigned int sizeSetup = 0, sizeHeader = 0, sizeExtended = 0;

    //
    unsigned int address, size, crc, crcTest = 0;
    address = size = crc = 0;

    // Segmentbegin: Bedienung stoppen
    if ((res = enter_read_data_segment(device_address, IV_VALUE_SEGMENT_GET_UPDATE_BG, address, size, crc)) != IV_ERROR_SUCCESS)
    {
        if ((res & 0xFF) == 1)
        {
            // Funktionsfehler Typ: IV_VALUE_SEGMENT_GET_UPDATE_BG unbekannt.
            // Wenn dass der Fall ist, wird das Auslesen noch nicht unterstuetzt.
            res = -IV_ERROR_FUNCTION_UNAVAILABLE;
        }
        else
        {
            return res;
        }
    }
    trace("%s(%i):INFO(1): %s [ beginGetSegment(%i): %i, %i, 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_UPDATE_BG, address, size, crc);

    //////////////////////////////
    // Begin

    // Segmentbegin: Setup uebertragen
    if ((res = enter_read_data_segment(device_address, IV_VALUE_SEGMENT_GET_SETUP, address, size, crc)) != IV_ERROR_SUCCESS)
    {
        return res;
    }
    trace("%s(%i):INFO(1): %s [ beginGetSegment(%i): %i, %i, 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_SETUP, address, size, crc);

    // Gelieferte zur Mindestgroesse pruefen
    if (size < (EEP_SIZE_SU + sizeof(StructConfig) + sizeof(dummyConfig)))
    {
        return -IV_ERROR_PROCESSING;
    }

    //////////////////////////////
    // Grunddaten

    // EEPROM uebertragen
    if ((res = read_from_flash(device_address, address, m_file.m_eep_mem.data(), m_file.m_eep_mem.size())) != IV_ERROR_SUCCESS)
    {
        return res;
    }
    address += EEP_SIZE_SU;
    // CONFIG uebertragen (100 Byte)
    if ((res = read_from_flash(device_address, address, m_file.m_config_mem.data(), m_file.m_config_mem.size())) != IV_ERROR_SUCCESS)
    {
        return res;
    }
    address += sizeof(StructConfig);
    // CONFIG dummybytes (28 Byte)
    if ((res = read_from_flash(device_address, address, dummyConfig, sizeof(dummyConfig))) != IV_ERROR_SUCCESS)
    {
        return res;
    }
    address += sizeof(dummyConfig);

    // Groesse der Flashdaten fuer Setup aus EepMem entnehmen
    sizeSetup = m_file.endian_reverse( m_file.m_flash_header.FLASH_SIZE_SUS );

    // FLASH uebertragen
    if ((res = read_from_flash(device_address, address, m_file.m_flash_mem.data(), sizeSetup)) != IV_ERROR_SUCCESS)
    {
        return res;
    }
    address += sizeSetup;

    // Groesse der Headerdaten aus EepMem entnehmen.
    sizeHeader = m_file.m_flash_header.HEADER_SIZE;
    // Headerdaten uebertragen
    if ((res = read_from_flash(device_address, address, m_file.m_file_header_mem.data(), sizeHeader)) != IV_ERROR_SUCCESS)
    {
        return res;
    }

    // Pruefen ob die Headerangaben passen.
    if (!m_file.IsHeader())
        return -IV_ERROR_SETUPFILE_MIN_LIBRARY_VERSION;

    SoftwarebuildVersion minVersion(m_file.m_file_header.minLibraryVersion);
    // Pruefen ob die Mindestversion der benoetigten DLL erreicht bleibt.
    if (libraryVersion < minVersion) {

        // Mindestversion DLL wird nicht eingehalten.
        return -IV_ERROR_SETUPFILE_MIN_LIBRARY_VERSION;
    }

    ///////////////
    // Grunddaten
    if (size != (EEP_SIZE_SU + sizeof(StructConfig) + sizeof(dummyConfig) + sizeSetup + sizeHeader)) {

        trace("%s(%i):ERROR(1): %s [ Size von Segment falsch: %u, %u ]\n", __FILE__, __LINE__, __func__, size, 
            (EEP_SIZE_SU + sizeof(StructConfig) + sizeof(dummyConfig) + sizeSetup + sizeHeader));
        return -IV_ERROR_SETUPFILE_MIN_LIBRARY_VERSION;;
    }

    // CRC berechnen
    crcTest = 0xFFFFFFFF;
    crcTest = get_crc32(m_file.m_eep_mem.data(), m_file.m_eep_mem.size(), crcTest, false);
    crcTest = get_crc32(m_file.m_config_mem.data(), m_file.m_config_mem.size(), crcTest, false);
    crcTest = get_crc32(dummyConfig, sizeof(dummyConfig), crcTest, false);
    crcTest = get_crc32(m_file.m_flash_mem.data(), sizeSetup, crcTest, false);
    crcTest = get_crc32(m_file.m_file_header_mem.data(), sizeHeader, crcTest);

    //////////////////////////////
    // Ende

    if (crc != crcTest)
    {
        trace("%s(%i):ERROR(1): %s [ CRC von Segment falsch: 0x%08X, 0x%08X ]\n", __FILE__, __LINE__, __func__, crc, crcTest);
        return -IV_ERROR_CRC_SEGMENT;
    }

    // Segmentend: Setup uebertragen
    if ((res = exit_read_data_segment(device_address, IV_VALUE_SEGMENT_GET_SETUP)) != IV_ERROR_SUCCESS)
    {
        return res;
    }
    trace("%s(%i):INFO(1): %s [ endGetSegment(%i) ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_SETUP);

    //////////////////////////////
    // Begin
    address = size = crc = 0;
    // Segmentbegin: Setup uebertragen
    if ((res = enter_read_data_segment(device_address, IV_VALUE_SEGMENT_GET_SETUP_EXTENDED, address, size, crc)) != IV_ERROR_SUCCESS)
    {
        return res;
    }
    trace("%s(%i):INFO(1): %s [ beginGetSegment(%i): %i, %i, 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_SETUP_EXTENDED, address, size, crc);

    //////////////////////////////
    // Extended - Speicher Lesen

    // Groesse des Extended Speichers aus eeprom ermitteln.
    sizeExtended = m_file.m_flash_header.EXTENDED_SIZE;

    if ((res = read_from_flash(device_address, address, m_file.m_flashExtended, sizeExtended)) != IV_ERROR_SUCCESS)
    {
        return res;
    }
    address += sizeExtended;

    ///////////////
    // Extended - Speicher pruefen
    if (size != sizeExtended) {

        trace("%s(%i):ERROR(1): %s [ Size von Segment falsch: %u, %u ]\n", __FILE__, __LINE__, __func__, size, sizeExtended);
        return -IV_ERROR_SETUPFILE_MIN_LIBRARY_VERSION;
    }

    // Extended - Speicher
    crcTest = 0xFFFFFFFF;
    crcTest = get_crc32(m_file.m_flashExtended, size, crcTest);

    if (crc != crcTest)
    {
        trace("%s(%i):ERROR(1): %s [ CRC von Segment falsch: 0x%08X, 0x%08X ]\n", __FILE__, __LINE__, __func__, crc, crcTest);
        return -IV_ERROR_CRC_SEGMENT;
    }

    //////////////////////////////
    // Ende

    // Segmentend: Setup uebertragen
    if ((res = exit_read_data_segment(device_address, IV_VALUE_SEGMENT_GET_SETUP_EXTENDED)) != IV_ERROR_SUCCESS)
    {
        return res;
    }
    trace("%s(%i):INFO(1): %s [ endGetSegment(%i) ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_SETUP_EXTENDED);


    // Die Eingabekettenzeiger anpassen, da diese fuer die FW vor dem Einspielen angepasst wurden.
    // Zeigeroffsets anpassen (addieren von Offset)
    AdjustExtendedMemPointer();

    crc = 0;
    // Segmentend: Bedienung stoppen
    if ((res = exit_read_data_segment(device_address, IV_VALUE_SEGMENT_GET_UPDATE_BG)) != IV_ERROR_SUCCESS)
    {
        return res;
    }
    trace("%s(%i):INFO(1): %s [ endGetSegment(%i) ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_UPDATE_BG);

    return res;
}

int Channel4::writeListWithTimeboySlot( unsigned char device_address )
{
    int res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = writeListData( device_address );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

int Channel4::writeListData( unsigned char device_address )
{
    // BlockTransfer durchfuehren.
    controller_segment &controllerSegmentCurrent = get_controller_segment(device_address);

    if (controllerSegmentCurrent.transfer_state == controller_segment::bts_inactive)
    {    
        int i;
        // Listengesamtgroesse berechnen und ggf. sortieren.
        unsigned int listSize = MAX_DATENSATZ_BESCHREIBUNG * sizeof(unsigned int);
        unsigned int listPointer[MAX_DATENSATZ_BESCHREIBUNG];
        for (i = 0; i < MAX_DATENSATZ_BESCHREIBUNG; ++i)
        {
            if (m_listSize[i])
            {			
                listPointer[i] = listSize + FLASH_LIST_START;
                ChangeOrderLong((unsigned char*)&listPointer[i]);

                listSize += m_listSize[i];
            }
            else
            {
                listPointer[i] = 0;
            }
        }

        if (listSize == (MAX_DATENSATZ_BESCHREIBUNG * sizeof(unsigned int)))
        {
            return -IV_ERROR_LIST_PROCESSING;
            return 0; // keine Listen vorhanden
        }
    }

    int res = writeListData4( device_address );

    if ((res > 0) && (((res >> 8) & 0xFF) == 3))
        return -IV_ERROR_NOT_ENOUGH_FLASH; // Es steht nicht genuegend freier Flash Speicher zur Verfuegung.

    return res;
}

// Fuer neue Listentechnik
int Channel4::writeListData4( unsigned char device_address )
{
    unsigned char tableNum;
    int res = IV_ERROR_SUCCESS;

    StructTable2 table;
    char name[32];

    //
    unsigned short segmentUpdateVariation = IV_VALUE_SEGMENT_UPDATE_BG;
    unsigned int address, size, crc;
    unsigned int addressOffset = 0;
    address = size = crc = 0;

    // BlockTransfer durchfuehren.
    controller_segment &controllerSegmentCurrent = get_controller_segment(device_address);

    if (controllerSegmentCurrent.transfer_state == controller_segment::bts_inactive)
    {
        // Ggf. gesetztes Streaming deaktivieren.
        set_stream_transfer( false );
        clear_stream_transfer();

        // Systemstopp Listen uebertragen
        while ((res = enter_write_data_segment(device_address, segmentUpdateVariation, address, size, crc)) != IV_ERROR_SUCCESS)
        {
            // Wenn das Hintergrunduebertragen nicht unterstuetzt wird, dann die Standardmethode verwenden.
            if (((res & 0xFF) == 1) && (segmentUpdateVariation == IV_VALUE_SEGMENT_UPDATE_BG))
            {
                segmentUpdateVariation = IV_VALUE_SEGMENT_UPDATE;
                continue;
            }

            return res;
        }

        // Beginn eintragen.
        controllerSegmentCurrent.reset();
        controllerSegmentCurrent.transfer_state = controller_segment::bts_active;
        controllerSegmentCurrent.transfer_type = controller_segment::btt_list;
        if (controllerSegmentCurrent.allowed_transfer_types == controller_segment::btt_none)
        {
            // Settings ggf. von Global-Object holen.
            controller_segment &controllerSegmentGlobal = get_controller_segment(255);
            if (controllerSegmentGlobal.allowed_transfer_types != controller_segment::btt_none)
            {
                controllerSegmentCurrent.allowed_transfer_types = controllerSegmentGlobal.allowed_transfer_types;
                controllerSegmentCurrent.duration = controllerSegmentGlobal.duration;
                controllerSegmentCurrent.transfer_flags |= controller_segment::cf_settings_from_global;
            }
        }
        controllerSegmentCurrent.begin_time_point = std::chrono::system_clock::now();
        // UpdateVariation halten.
        controllerSegmentCurrent.update_variation = segmentUpdateVariation;

        // Pruefen ob die Uebertragung per Stream moeglich ist.
        if ((res = write_stream(device_address, Stream_type::Update_background, nullptr, 0)) == IV_ERROR_SUCCESS)
        {
            set_stream_transfer(true);
        }
    }

    // Listendaten Tabellenweise uebertragen
    for (tableNum=controllerSegmentCurrent.number; tableNum < MAX_DATENSATZ_BESCHREIBUNG; ++tableNum, ++controllerSegmentCurrent.number)
    {
        // Nicht importierte Listen uebergehen
        if (m_listSize[tableNum] == 0)
            continue;

        // Bei wiederaufnahme eines Blocktransfers ist dieses bereits durchgeführt worden.
        if (controllerSegmentCurrent.resume == false)
        {
            // Bei Streaming werden die inneren Bloecke wegelassen.
            if (is_stream_transfer() == false)
            {
                // Maximale Segmentanzahl nicht ueberschreiten
                if (tableNum && !(tableNum % 16))
                {
                    crc = 0;
                    // Systemstopp Listen uebertragen
                    if ((res = exit_write_data_segment(device_address, segmentUpdateVariation, crc)) != IV_ERROR_SUCCESS)
                    {
                        return res;
                    }

                    // Systemstopp Listen uebertragen
                    if ((res = enter_write_data_segment(device_address, segmentUpdateVariation, address, size, crc)) != IV_ERROR_SUCCESS)
                    {
                        return res;
                    }
                }
            }

            memset(name, 0, sizeof(name));
            snprintf(name, sizeof( name), "LIST.%i", tableNum);
            // Zur Liste Informationen ermitteln
            if ((res = read_table_struct(device_address, name, table)) != IV_ERROR_SUCCESS)
            {
                //TODO: Liste unbekannt Funktionsfehler 2 auswerten.
                return res;
            }
            strncpy( name, table.name, sizeof( name ) - 1 );
            snprintf( table.name, sizeof( table.name ), "LIST.%s", name );

            // Muss die Liste auf Sortierung geprueft und ggf. sortiert werden.
            if (m_listSortCheck[tableNum] == false)
            {
                // Listen ohne Datensaetze nicht sortieren.
                // ACHTUNG: Ansonsten Division durch 0 bei Pruefung der Datensatzgroesse !!!
                if ( *reinterpret_cast<unsigned int*>(&m_pucListMem[tableNum][0]) )
                {
                    int fieldOffs = 0;

                    if (table.flags & TF_SORT && table.sort < table.count)
                    {
                        char name[64];
                        int fieldNum;
                        StructTable2Field field;
                        memset(&field, 0, sizeof(StructTable2Field));
                        m_ucListRecordSize[tableNum] = 0;
                        // Alle Felder fuer Offset lesen
                        for (fieldNum=0; fieldNum < table.count; fieldNum++)
                        {
                            if (fieldNum < (table.sort + 1))
                            {
                                // Datensatzoffset fuer Sortierfeld aufsummieren.
                                fieldOffs += field.size;
                            }

                            snprintf(name, sizeof( name ), "LIST.%i.%i", tableNum, fieldNum);

                            if ((res = read_table_column_struct(device_address, name, field)) != IV_ERROR_SUCCESS)
                            {
                                if (res < 0)
                                {
                                    // Kommunikationsfehler
                                    return res;
                                }
                                else
                                {
                                    // Funktionsfehler
                                    return res;
                                }
                            }

                            // Datensatzgroesse fuer Swap-Routine berechnen
                            m_ucListRecordSize[tableNum] += field.size;
                        }

                        // Die Listensatzgröße muss mit der importierten übereinstimmen.
                        if ( ((m_listSize[tableNum] - sizeof(unsigned int)) / *(unsigned int*)&m_pucListMem[tableNum][0]) != m_ucListRecordSize[tableNum] )
                        {
                            trace("%s(%i):ERROR(1): %s [ Listensatzgroesse (%i) stimmt nicht mit der im Setup definierten ueberein. ]\n", __FILE__, __LINE__, __func__, tableNum);
                            return -IV_ERROR_LIST_PROCESSING;
                        }

                        // Liste sortieren.
                        SortList(m_pucListMem[tableNum], m_ucListRecordSize[tableNum], fieldOffs);

                        // Bei allen Listen ist 0 die sortierte Spalte
                        // Flag dafuer dass die Liste sortiert wurde setzen.
                        table.flags |= TF_SORTED;
                    }
                }

                // Sortiert oder zumindest geprueft.
                m_listSortCheck[tableNum] = true;
            }

            // CRC berechnen
            crc = 0xFFFFFFFF;
            crc = get_crc32((unsigned char*)&table, sizeof(StructTable2), crc, false);
            crc = get_crc32(m_pucListMem[tableNum], m_listSize[tableNum], crc);

            // Groesze berechnen
            size = sizeof(StructTable2) + m_listSize[tableNum];

            // Bei Streaming werden die inneren Bloecke wegelassen.
            if (!is_stream_transfer())
            {
                // Systemstopp Liste uebertragen
                if ((res = enter_write_data_segment(device_address, IV_VALUE_SEGMENT_LIST, address, size, crc)) != IV_ERROR_SUCCESS)
                {
                    return res;
                }

                // Zusatzdaten
                controllerSegmentCurrent.address = address;
                controllerSegmentCurrent.crc = crc;
            }

            // Daten uebertragen
            if ((res = write_to_flash(device_address, address, (unsigned char*)&table, sizeof(StructTable2))) != IV_ERROR_SUCCESS)
            {
                // Speicherplatz reicht nicht aus.
                if (res == 3)
                {
                    res = -IV_ERROR_NOT_ENOUGH_FLASH;
                }

                return res;
            }

            // Daten uebertragen
            address += sizeof(StructTable2);

            // Startadresse halten
            controllerSegmentCurrent.address = address;
            // Bereits uebertragene Datenmenge der Tabellenstruktur auf 0 setzen.
            controllerSegmentCurrent.size = 0;
            addressOffset = 0;
        }
        else
        {
            // Zwischengespeicherte Daten fuer Wiederaufnahme setzen.
            segmentUpdateVariation = controllerSegmentCurrent.update_variation;
            address = controllerSegmentCurrent.address;
            crc = controllerSegmentCurrent.crc;
            addressOffset = controllerSegmentCurrent.size;
        }

        // Abbruch der Blockuebertragung.
        if (controllerSegmentCurrent.discard == false)
        {        
            if ((res = write_to_flash(device_address, address + addressOffset, m_pucListMem[tableNum] + addressOffset, m_listSize[tableNum] - addressOffset)) != IV_ERROR_SUCCESS)
            {
                // Speicherplatz reicht nicht aus.
                if (res == 3)
                {
                    res = -IV_ERROR_NOT_ENOUGH_FLASH;
                }

                if (res == -IV_ERROR_BLOCK_TRANSFER_BREAK)
                {
                    controllerSegmentCurrent.resume = true;
                }

                return res;
            }
        }

        // Listendaten komplett uebertragen.
        controllerSegmentCurrent.resume = false;

        // Im Modus Streaming ?, dann gepufferte Daten senden.
        if (is_stream_transfer() && m_stream_buffer.size()) {
            if ((res = write_stream(device_address, Stream_type::Setup_list, m_stream_buffer.data(), static_cast<unsigned int>( m_stream_buffer.size() ))) != IV_ERROR_SUCCESS) {
                // Gepufferte Daten loeschen.
                clear_stream_transfer();
                return res;
            }
            // Gepufferte Daten loeschen.
            clear_stream_transfer();
        }

        // Bei Streaming werden die inneren Bloecke wegelassen.
        if (!is_stream_transfer())
        {
            // Systemstopp Liste uebertragen
            if ((res = exit_write_data_segment(device_address, IV_VALUE_SEGMENT_LIST, crc)) != IV_ERROR_SUCCESS)
            {
                return res;
            }
        }

        // Abbruch der Blockuebertragung.
        if (controllerSegmentCurrent.discard)
            break;
    }

    crc = 0;
    // Systemstopp Liste uebertragen
    if ((res = exit_write_data_segment(device_address, segmentUpdateVariation, crc)) != IV_ERROR_SUCCESS)
    {
        return res;
    }

    // BlockTransfer Beendet.
    controllerSegmentCurrent.end_time_point = std::chrono::system_clock::now();
    controllerSegmentCurrent.transfer_state = controller_segment::bts_inactive;

    return res;
}

int Channel4::writeEntrance2ListWithTimeboySlot( unsigned char device_address, unsigned char list_number )
{
    int res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        // Listendaten zum AEIV uebertragen
        res = writeEntrance2ListData4( device_address, list_number );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

// ListenBuffer in AE laden
int Channel4::writeEntrance2ListData4( unsigned char device_address, unsigned char list_number )
{
    // BlockTransfer durchfuehren.
    controller_segment &controllerSegmentCurrent = get_controller_segment(device_address);

    unsigned short segmentUpdateVariation = IV_VALUE_SEGMENT_UPDATE_BG;
    int res = IV_ERROR_SUCCESS;
    unsigned int address, size, crc;
    address = size = crc = 0;

    if (controllerSegmentCurrent.transfer_state == controller_segment::bts_inactive)
    {    
        // Alle Listen laden?
        if (list_number == 0xFF)
        {
            // Mindestens eine Liste geladen?
            int i;
            for (i=0; i < 8; i++)
            {
                if (m_ulEntrance2ListSize[i] != 0)
                {
                    break;
                }
            }
            if (i == 8)
            {
                return -IV_ERROR_LIST_PROCESSING; // keine Listen vorhanden
            }
        }
        else
        {
            if (m_ulEntrance2ListSize[list_number] == 0)
            {
                return -IV_ERROR_LIST_PROCESSING; // keine Listen vorhanden
            }
        }

        // Ggf. gesetztes Streaming deaktivieren.
        set_stream_transfer( false );
        clear_stream_transfer();

        // Systemstopp Setup uebertragen
        while ((res = enter_write_data_segment(device_address, segmentUpdateVariation, address, size, crc)) != IV_ERROR_SUCCESS)
        {
            // Wenn das Hintergrunduebertragen nicht unterstuetzt wird, dann die Standardmethode verwenden.
            if (((res & 0xFF) == 1) && (segmentUpdateVariation == IV_VALUE_SEGMENT_UPDATE_BG))
            {
                segmentUpdateVariation = IV_VALUE_SEGMENT_UPDATE;
                continue;
            }

            return res;
        }

        // Beginn eintragen.
        controllerSegmentCurrent.reset();
        controllerSegmentCurrent.transfer_state = controller_segment::bts_active;
        controllerSegmentCurrent.transfer_type = controller_segment::btt_access_control_list;
        if (controllerSegmentCurrent.allowed_transfer_types == controller_segment::btt_none)
        {
            // Settings ggf. von Global-Object holen.
            controller_segment &controllerSegmentGlobal = get_controller_segment(255);
            if (controllerSegmentGlobal.allowed_transfer_types != controller_segment::btt_none)
            {
                controllerSegmentCurrent.allowed_transfer_types = controllerSegmentGlobal.allowed_transfer_types;
                controllerSegmentCurrent.duration = controllerSegmentGlobal.duration;
                controllerSegmentCurrent.transfer_flags |= controller_segment::cf_settings_from_global;
            }
        }
        controllerSegmentCurrent.begin_time_point = std::chrono::system_clock::now();
        // UpdateVariation halten.
        controllerSegmentCurrent.update_variation = segmentUpdateVariation;

        if (list_number == 0xFF)
        {
            controllerSegmentCurrent.number = 0;
            controllerSegmentCurrent.max_number = 8;
        }
        else
        {
            controllerSegmentCurrent.number = list_number;
            controllerSegmentCurrent.max_number = list_number + 1;
        }

        // Pruefen ob die Uebertragung per Stream moeglich ist.
        if ((res = write_stream(device_address, Stream_type::Update_background, nullptr, 0)) == IV_ERROR_SUCCESS)
        {
            set_stream_transfer(true);
        }
    }
    else
    {
        // Zwischengespeicherte Daten fuer Wiederaufnahme setzen.
        segmentUpdateVariation = controllerSegmentCurrent.update_variation;
    }

    for (; controllerSegmentCurrent.number < controllerSegmentCurrent.max_number; ++controllerSegmentCurrent.number)
    {
        // nur geladene Liste bei 0xFF uebertragen
        if (m_ulEntrance2ListSize[controllerSegmentCurrent.number] != 0)
        {
            // Liste uebertragen
            if ((res = writeEntrance2List4(device_address, controllerSegmentCurrent.number)) != IV_ERROR_SUCCESS)
            {
                return res;
            }
        }

        // Abbruch der Blockuebertragung.
        if (controllerSegmentCurrent.discard)
            break;
    }

    crc = 0;
    // Systemstopp Setup uebertragen
    if ((res = exit_write_data_segment(device_address, segmentUpdateVariation, crc)) != IV_ERROR_SUCCESS)
    {
        return res;
    }

    // BlockTransfer Beendet.
    controllerSegmentCurrent.end_time_point = std::chrono::system_clock::now();
    controllerSegmentCurrent.transfer_state = controller_segment::bts_inactive;

    if ((res > 0) && (((res >> 8) & 0xFF) == 3))
    {
        return -IV_ERROR_NOT_ENOUGH_FLASH; // Es steht nicht genuegend freier Flash Speicher zur Verfuegung.
    }

    return res;
}

// ListenBuffer in AE laden
int Channel4::writeEntrance2List4( unsigned char device_address, unsigned char list_number )
{
    // BlockTransfer durchfuehren.
    controller_segment &controllerSegmentCurrent = get_controller_segment(device_address);

    int res = IV_ERROR_SUCCESS;
    StructTable2 table;
    char name[32];

    //
    unsigned short art = IV_VALUE_SEGMENT_LIST_ACCESS;
    unsigned int address, size, crc;
    address = size = crc = 0;
    unsigned int addressOffset = 0;

    // Bei wiederaufnahme eines Blocktransfers ist dieses bereits durchgeführt worden.
    if (controllerSegmentCurrent.resume == false)
    {       
        memset(name, 0, sizeof(name));
        snprintf(name, sizeof( name ), "ACCESS.%i", list_number);
        // Zur Liste Informationen ermitteln
        if ((res = read_table_struct(device_address, name, table)) != IV_ERROR_SUCCESS)
        {
            //TODO: Action2 unbekannt Funktionsfehler 2 auswerten.
            return res;
        }
        strncpy( name, table.name, sizeof( name ) - 1 );
        snprintf( table.name, sizeof( table.name ), "ACCESS.%s", name );

        // Muss die Liste auf Sortierung geprueft und ggf. sortiert werden.
        if (m_entrance2SortCheck[list_number] == false)
        {	
            // Listen ohne Datensaetze nicht sortieren.
            // ACHTUNG: Ansonsten Division durch 0 bei Pruefung der Datensatzgroesse !!!
            if ( m_ulEntrance2ListSize[list_number] )
            {
                int fieldOffs = 0;

                if (table.flags & TF_SORT && table.sort < table.count)
                {			
                    //char name[64];
                    int fieldNum;
                    StructTable2Field field;
                    memset(&field, 0, sizeof(StructTable2Field));
                    // Alle Felder fuer Offset lesen
                    for (fieldNum=0; fieldNum < table.count; fieldNum++)
                    {				
                        if (fieldNum < (table.sort + 1))
                        {					
                            // Datensatzoffset fuer Sortierfeld aufsummieren.
                            fieldOffs += field.size;
                        }
                        else // Da die Datensatzgroesse bekannt ist, brauchen die restlichen Felder nicht ermittelt zu werden. 
                        { 
                            break; 
                        }

                        field = m_entrance2TableField[list_number][fieldNum];

                        // Datensatzgroesse wird fuer Swap-Routine benoetigt!
                    }

                    // Liste sortieren.
                    SortList(m_pucEntrance2ListMem[list_number], m_ucEntrance2RecordSize[list_number], fieldOffs);

                    // Bei allen Listen ist 0 die sortierte Spalte
                    // Flag dafuer dass die Liste sortiert wurde setzen.
                    table.flags |= TF_SORTED;
                }
            }

            // Sortiert oder zumindes geprueft.
            m_entrance2SortCheck[list_number] = true;
        }

        // CRC berechnen
        crc = 0xFFFFFFFF;
        crc = get_crc32((unsigned char*)&table, sizeof(StructTable2), crc, false);
        crc = get_crc32(m_pucEntrance2ListMem[list_number], m_ulEntrance2ListSize[list_number], crc);

        // Groesze berechnen
        size = sizeof(StructTable2) + m_ulEntrance2ListSize[list_number];

        // Bei Streaming werden die inneren Bloecke wegelassen.
        if (!is_stream_transfer()) {

            // Systemstopp List uebertragen
            while ((res = enter_write_data_segment(device_address, art, address, size, crc)) != IV_ERROR_SUCCESS)
            {
                // Typ unbekannt?, dann als "normale" Liste versuchen..
                if (((res & 0xFF) == 1) && (art == IV_VALUE_SEGMENT_LIST_ACCESS))
                {
                    art = IV_VALUE_SEGMENT_LIST;
                    continue;
                }

                return res;
            }

            // Zusatzdaten
            controllerSegmentCurrent.address = address;
            controllerSegmentCurrent.crc = crc;
        }

        // Daten uebertragen
        if ((res = write_to_flash(device_address, address, (unsigned char*)&table, sizeof(StructTable2))) != IV_ERROR_SUCCESS)
        {
            // Speicherplatz reicht nicht aus.
            if (res == 3) { res = -IV_ERROR_NOT_ENOUGH_FLASH; }
            return res;
        }

        // Daten uebertragen
        address += sizeof(StructTable2);

        // Startadresse halten
        controllerSegmentCurrent.address = address;
        // Bereits uebertragene Datenmenge der Tabellenstruktur auf 0 setzen.
        controllerSegmentCurrent.size = 0;
        addressOffset = 0;
    }
    else
    {
        // Zwischengespeicherte Daten fuer Wiederaufnahme setzen.
        address = controllerSegmentCurrent.address;
        crc = controllerSegmentCurrent.crc;
        addressOffset = controllerSegmentCurrent.size;
    }

    // Abbruch der Blockuebertragung.
    if (controllerSegmentCurrent.discard == false)
    {
        if ((res = write_to_flash(device_address, address + addressOffset, m_pucEntrance2ListMem[list_number] + addressOffset, m_ulEntrance2ListSize[list_number] - addressOffset)) != IV_ERROR_SUCCESS)
        {
            // Speicherplatz reicht nicht aus.
            if (res == 3)
            {
                res = -IV_ERROR_NOT_ENOUGH_FLASH;
            }

            if (res == -IV_ERROR_BLOCK_TRANSFER_BREAK)
            {
                controllerSegmentCurrent.resume = true;
            }

            return res;
        }
    }

    // Listendaten komplett uebertragen.
    controllerSegmentCurrent.resume = false;

    // Im Modus Streaming ?, dann gepufferte Daten senden.
    if (is_stream_transfer() && m_stream_buffer.size()) {
        if ((res = write_stream(device_address, Stream_type::List_access, m_stream_buffer.data(), static_cast<unsigned int>( m_stream_buffer.size() ))) != IV_ERROR_SUCCESS) {
            // Gepufferte Daten loeschen.
            clear_stream_transfer();
            return res;
        }
        // Gepufferte Daten loeschen.
        clear_stream_transfer();
    }

    // Bei Streaming werden die inneren Bloecke wegelassen.
    if (!is_stream_transfer()) {

        // Systemstopp Setup uebertragen
        if ((res = exit_write_data_segment(device_address, art, crc)) != IV_ERROR_SUCCESS)
        {
            return res;
        }
    }

    return res;
}

// ListenBuffer in AE laden
int Channel4::writeTimeboyListData4( unsigned char device_address, unsigned char group_id, unsigned char list_number, int reserve )
{
    // Alle Listen laden?
    if (group_id == 0xFF)
    {
        int groupID, listNum, found = 0;
        for (groupID=0; groupID < MAX_DATENSATZ_BESCHREIBUNG; groupID++)
        {
            if (list_number == 0xFF)
            {
                // Mindestens eine Liste geladen?
                for (listNum=0; listNum < MAX_DATENSATZ_BESCHREIBUNG; listNum++)
                {
                    if (m_ulTimeboyListSize[groupID][listNum] != 0)
                    {
                        found = 1;
                        break;
                    }
                }
                if (found)
                {
                    break;
                }
            }
            else
            {
                // Mindestens eine Liste geladen?
                if (m_ulTimeboyListSize[groupID][list_number] == 0)
                {
                    found = 1;
                    break;
                }
            }
        }
        if (found == 0)
        {
            return -IV_ERROR_LIST_PROCESSING; // keine Listen vorhanden
        }
    }
    else
    {
        if (list_number == 0xFF)
        {
            int listNum, found = 0;
            // Mindestens eine Liste geladen?
            for (listNum=0; listNum < MAX_DATENSATZ_BESCHREIBUNG; listNum++)
            {
                if (m_ulTimeboyListSize[group_id][listNum] != 0)
                {
                    found = 1;
                    break;
                }
            }
            if (found == 0)
            {
                return -IV_ERROR_LIST_PROCESSING; // keine Listen vorhanden
            }
        }
        else
        {
            if (m_ulTimeboyListSize[group_id][list_number] == 0)
            {
                return -IV_ERROR_LIST_PROCESSING; // keine Listen vorhanden
            }
        }
    }

    // Ggf. gesetztes Streaming deaktivieren.
    set_stream_transfer( false );
    clear_stream_transfer();

    unsigned short segmentUpdateVariation = IV_VALUE_SEGMENT_UPDATE_BG;
    int res = IV_ERROR_SUCCESS;
    unsigned int address, size, crc;
    address = size = crc = 0;

    // Systemstopp Listen uebertragen
    while ((res = enter_write_data_segment(device_address, segmentUpdateVariation, address, size, crc)) != IV_ERROR_SUCCESS)
    {
        // Wenn das Hintergrunduebertragen nicht unterstuetzt wird, dann die Standardmethode verwenden.
        if (((res & 0xFF) == 1) && (segmentUpdateVariation == IV_VALUE_SEGMENT_UPDATE_BG))
        {
            segmentUpdateVariation = IV_VALUE_SEGMENT_UPDATE;
            continue;
        }

        return res;
    }

    // Pruefen ob die Uebertragung per Stream moeglich ist.
    if ((res = write_stream(device_address, Stream_type::Update_background, nullptr, 0)) == IV_ERROR_SUCCESS)
    {
        set_stream_transfer(true);
    }

    // Alle ids abarbeiten?
    if (group_id == 0xFF)
    {
        int tableNum, tableCount = 0;
        int groupID;
        for (groupID=0; groupID < MAX_DATENSATZ_BESCHREIBUNG; groupID++)
        {
            if (list_number == 0xFF)
            {
                for (tableNum=0; tableNum < MAX_DATENSATZ_BESCHREIBUNG; tableNum++)
                {
                    // nur geladene Liste bei 0xFF uebertragen
                    if (m_ulTimeboyListSize[groupID][tableNum] != 0)
                    {
                        // Maximale Segmentanzahl nicht ueberschreiten
                        if (tableCount && !(tableCount % 16))
                        {
                            crc = 0;
                            // Systemstopp Listen uebertragen
                            if ((res = exit_write_data_segment(device_address, segmentUpdateVariation, crc)) != IV_ERROR_SUCCESS)
                            {
                                return res;
                            }

                            // Systemstopp Listen uebertragen
                            if ((res = enter_write_data_segment(device_address, segmentUpdateVariation, address, size, crc)) != IV_ERROR_SUCCESS)
                            {
                                return res;
                            }
                        }

                        // Liste uebertragen
                        if ((res = writeTimeboyList4(device_address, groupID, tableNum, reserve)) != IV_ERROR_SUCCESS)
                        {
                            return res;
                        }

                        tableCount++;
                    }
                }
            }
            else
            {
                // nur geladene Liste bei 0xFF uebertragen
                if (m_ulTimeboyListSize[groupID][list_number] != 0)
                {
                    // Maximale Segmentanzahl nicht ueberschreiten
                    if (tableCount && !(tableCount % 16))
                    {
                        crc = 0;
                        // Systemstopp Listen uebertragen
                        if ((res = exit_write_data_segment(device_address, segmentUpdateVariation, crc)) != IV_ERROR_SUCCESS)
                        {
                            return res;
                        }

                        // Systemstopp Listen uebertragen
                        if ((res = enter_write_data_segment(device_address, segmentUpdateVariation, address, size, crc)) != IV_ERROR_SUCCESS)
                        {
                            return res;
                        }
                    }

                    // Liste uebertragen
                    if ((res = writeTimeboyList4(device_address, groupID, list_number, reserve)) != IV_ERROR_SUCCESS)
                    {
                        return res;
                    }

                    tableCount++;
                }
            }
        }
    }
    else
    {
        if (list_number == 0xFF)
        {
            int tableNum, tableCount = 0;
            for (tableNum=0; tableNum < MAX_DATENSATZ_BESCHREIBUNG; tableNum++)
            {
                // nur geladene Liste bei 0xFF uebertragen
                if (m_ulTimeboyListSize[group_id][tableNum] != 0)
                {
                    // Maximale Segmentanzahl nicht ueberschreiten
                    if (tableCount && !(tableCount % 16))
                    {
                        crc = 0;
                        // Systemstopp Listen uebertragen
                        if ((res = exit_write_data_segment(device_address, segmentUpdateVariation, crc)) != IV_ERROR_SUCCESS)
                        {
                            return res;
                        }

                        // Systemstopp Listen uebertragen
                        if ((res = enter_write_data_segment(device_address, segmentUpdateVariation, address, size, crc)) != IV_ERROR_SUCCESS)
                        {
                            return res;
                        }
                    }

                    // Liste uebertragen
                    if ((res = writeTimeboyList4(device_address, group_id, tableNum, reserve)) != IV_ERROR_SUCCESS)
                    {
                        return res;
                    }

                    tableCount++;
                }
            }
        }
        else
        {
            // Liste uebertragen
            if ((res = writeTimeboyList4(device_address, group_id, list_number, reserve)) != IV_ERROR_SUCCESS)
            {
                return res;
            }
        }
    }

    crc = 0;
    // Systemstopp Listen uebertragen
    if ((res = exit_write_data_segment(device_address, segmentUpdateVariation, crc)) != IV_ERROR_SUCCESS)
    {
        return res;
    }

    return res;
}

int Channel4::writeTimeboyListWithTimeboySlot( unsigned char device_address, unsigned char group_id, unsigned char list_number, int reserve )
{
    int res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        // Listendaten zum AEIV uebertragen
        res = writeTimeboyListData4( device_address, group_id, list_number, reserve );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

// ListenBuffer in AE laden
int Channel4::writeTimeboyList4( unsigned char device_address, unsigned char group_id, unsigned char list_number, int reserve )
{
    (void)reserve;
    int res = IV_ERROR_SUCCESS;
    StructTable2 table;
    memset(&table, 0, sizeof(StructTable2));
    table.num = list_number;
    snprintf(table.name, sizeof( table.name ), "TBLIST.%i.%i", group_id, list_number);
    table.flags = m_timeboyFlags[group_id][list_number];

    //
    unsigned short art = IV_VALUE_SEGMENT_LIST_TIMEBOY;
    unsigned int address, size, crc;
    address = size = crc = 0;
    /*
    // Systemstopp Setup uebertragen
    if ((res = SegmentBegin(device_address, SS_UPDATE, address, size, crc)) != IV_ERROR_SUCCESS)
    {
    return res;
    }
    */
    // CRC berechnen
    crc = 0xFFFFFFFF;
    crc = get_crc32((unsigned char*)&table, sizeof(StructTable2), crc, false);
    crc = get_crc32(m_pucTimeboyListMem[group_id][list_number], m_ulTimeboyListSize[group_id][list_number], crc);

    // Groesze berechnen
    size = sizeof(StructTable2) + m_ulTimeboyListSize[group_id][list_number];

    // Bei Streaming werden die inneren Bloecke wegelassen.
    if (!is_stream_transfer()) {

        // Systemstopp List uebertragen
        while ((res = enter_write_data_segment(device_address, art, address, size, crc)) != IV_ERROR_SUCCESS)
        {
            // Typ unbekannt?, dann als "normale" Liste versuchen..
            if (((res & 0xFF) == 1) && (art == IV_VALUE_SEGMENT_LIST_ACCESS))
            {
                art = IV_VALUE_SEGMENT_LIST;
                continue;
            }

            return res;
        }
    }

    // Daten uebertragen
    if ((res = write_to_flash(device_address, address, (unsigned char*)&table, sizeof(StructTable2))) != IV_ERROR_SUCCESS)
    {
        // Speicherplatz reicht nicht aus.
        if (res == 3) { res = -IV_ERROR_NOT_ENOUGH_FLASH; }
        return res;
    }
    // Daten uebertragen
    address += sizeof(StructTable2);
    if ((res = write_to_flash(device_address, address, m_pucTimeboyListMem[group_id][list_number], m_ulTimeboyListSize[group_id][list_number])) != IV_ERROR_SUCCESS)
    {
        // Speicherplatz reicht nicht aus.
        if (res == 3) { res = -IV_ERROR_NOT_ENOUGH_FLASH; }
        return res;
    }

    // Im Modus Streaming ?, dann gepufferte Daten senden.
    if (is_stream_transfer() && m_stream_buffer.size()) {
        if ((res = write_stream(device_address, Stream_type::List_timeboy, m_stream_buffer.data(), static_cast<unsigned int>( m_stream_buffer.size() ))) != IV_ERROR_SUCCESS) {
            // Gepufferte Daten loeschen.
            clear_stream_transfer();
            return res;
        }
        // Gepufferte Daten loeschen.
        clear_stream_transfer();
    }

    // Bei Streaming werden die inneren Bloecke wegelassen.
    if (!is_stream_transfer()) {

        // Systemstopp Setup uebertragen
        if ((res = exit_write_data_segment(device_address, art, crc)) != IV_ERROR_SUCCESS)
        {
            return res;
        }
    }

    if ((res > 0) && (((res >> 8) & 0xFF) == 3))
        return -IV_ERROR_NOT_ENOUGH_FLASH; // Es steht nicht genuegend freier Flash Speicher zur Verfuegung.

    /*
    crc = 0;
    // Systemstopp Setup uebertragen
    if ((res = SegmentEnd(device_address, SS_UPDATE, crc)) != IV_ERROR_SUCCESS)
    {
    return res;
    }
    */
    return res;
}

int Channel4::writeDeviceWithTimeboySlot( unsigned char device_address, std::string &file_path )
{
    int res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = writeDevice( device_address, file_path );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

int Channel4::writeDevice( unsigned char device_address, std::string &file_path )
{
    // Hardware- und Softwareinfo der ausgewaehlten Datei.
    StructHardwareInfo hwInfo;

    int iRes = IV_ERROR_SUCCESS;

    // Puffer fuer die Hardwareinfo in Textform
    m_hardware_info_as_text.clear();
    // Strukturen loeschen.
    memset(&hwInfo, 0, sizeof(StructHardwareInfo));


    //#########################################################################
    // Pruefen ob gewaehlte Datei in das Geraet eingespielt werden kann.
    // Dateityp ermitteln
    struct stat st;
    if (stat(file_path.c_str(), &st) == -1) 
    {	
        return -IV_ERROR_FILE_OPEN;
    }

    // Angabe ist eine Datei
    if (S_ISREG(st.st_mode)) 
    {	
        // Nur ".hex" gueltig.
        size_t pos = file_path.size() - 4;
#if defined(_MSC_VER) || defined(__MINGW32__)
        if (file_path[pos] != '.' || 
            ((!(file_path[pos + 1] == 'h' || file_path[pos + 1] == 'H') ||
            !(file_path[pos + 2] == 'e' || file_path[pos + 2] == 'E') ||
            !(file_path[pos + 3] == 'x' || file_path[pos + 3] == 'X')) &&
            (!(file_path[pos + 1] == 'd' || file_path[pos + 1] == 'D') ||
            !(file_path[pos + 2] == 'f' || file_path[pos + 2] == 'F') ||
            !(file_path[pos + 3] == 'z' || file_path[pos + 3] == 'Z')) &&
            (!(file_path[pos + 1] == 'z' || file_path[pos + 1] == 'Z') ||
            !(file_path[pos + 2] == 'i' || file_path[pos + 2] == 'I') ||
            !(file_path[pos + 3] == 'p' || file_path[pos + 3] == 'P'))) ) 
        {	
            return -IV_ERROR_FILE_EXTENSION;
        }
#else 
        if (!S_ISREG(st.st_mode) || file_path[pos] != '.' || 
            !(file_path[pos + 1] == 'h' || file_path[pos + 1] == 'H') ||
            !(file_path[pos + 2] == 'e' || file_path[pos + 2] == 'E') ||
            !(file_path[pos + 3] == 'x' || file_path[pos + 3] == 'X') ) 
        {
            return -IV_ERROR_FILE_EXTENSION;
        }
#endif	
        // Handelt es sich um eine Hexdatei
        if (file_path[pos + 1] == 'h' || file_path[pos + 1] == 'H') 
        {	
            // Dateidaten laden
            m_fileDevice.resetData();
            if (!m_fileDevice.load(file_path)) 
            {	
                int err = IV_ERROR_SUCCESS;
                if (m_fileDevice.errorState() == FileIntelHex::Error_state::version 
                    || m_fileDevice.errorState() == FileIntelHex::Error_state::device_unknown)
                {
                    err = -IV_ERROR_VERSION_DEVICEFILE;
                } 
                else if (m_fileDevice.errorState() == FileIntelHex::Error_state::file_open)
                {
                    err = -IV_ERROR_FILE_OPEN;
                } 
                else 
                {
                    err = -IV_ERROR_DEVICEFILE_LOAD;
                }

                m_fileDevice.resetData();
                return err;
            }

            // Ist die Softwareinformation in der Geraetedatei verfuegbar?
            if (m_fileDevice.GetHeaderSoftwareInfo().size) 
            {	
                HardwareInfo hardwareInfo;
                // Auslesen der Hardwareinfo
                if ((iRes = read_hardware_info(device_address, hardwareInfo)) != IV_ERROR_SUCCESS) 
                {
                    return iRes;
                }

                // Uebergangsweise auf die Vorgaengerversion wechseln.
                hwInfo = hardwareInfo.exportStructHardwareInfo();

                // Liegt keine Hardwareinfo vor?
                // Verwendung der ZK-Option ueber Vorhandensein der Readerliste ermitteln.
                if (hwInfo.version == 0) 
                {	
                    int handle;
                    std::string reserved;
                    if (open_table(device_address, "ACCESS.READER", 0, handle, reserved) == IV_ERROR_SUCCESS &&
                        close_table(device_address, handle) == IV_ERROR_SUCCESS) 
                    {	
                        // ZK-Option in Verwendung.
                        hwInfo.administrationFlags |= HIAF_HAS_ZK;
                    }
                }

                // Pruefen ob die Datei in die vorliegende Hardware eingespielt werden kann.
                if (0 == m_fileDevice.CompareInfos(hwInfo, m_fileDevice.GetHeaderSoftwareInfo())) 
                {	
                    // Hardwareinfo in Text aufloesen
                    hardware_info_to_string(hwInfo, m_hardware_info_as_text);
                    m_fileDevice.setErrorTextExtended( m_fileDevice.errorText() );
                    return -IV_ERROR_DEVICEFILE_MATCH;
                }

            } 
            else 
            {	
                unsigned char deviceType = 255, boardType = 255;
                // Geraetetyp ermitteln
                if ((iRes = read_device_type(device_address, deviceType)) != IV_ERROR_SUCCESS || (iRes = read_board_type(device_address, boardType)) != IV_ERROR_SUCCESS) 
                {	
                    if (iRes == -IV_ERROR_FUNCTION_UNAVAILABLE)
                    {
                        // Der Geraetetyp wird manuell gesetzt.
                        deviceType = 0;
                    }
                    else
                    {
                        return iRes;
                    }
                }

                //#######################################################################
                // Wenn die SoftwareInfo fehlt kann die Datei keine Mulitple sein!!!

                switch (deviceType) 
                {
                case IV_VALUE_TYPE_DEVICE_PZE: // PZE-MasterIV
                    if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_PZE) 
                    {
                        // Die gewaehlte Geraetedatei ist NICHT fuer das PZE-MasterIV.
                        return -IV_ERROR_DEVICEFILE_WRONG;
                    }
                    break;

                case IV_VALUE_TYPE_DEVICE_ZK: // ZK-MasterIV
                    if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_ZK) 
                    {
                        // Die gewaehlte Geraetedatei ist NICHT fuer das ZK-MasterIV.
                        return -IV_ERROR_DEVICEFILE_WRONG;
                    }
                    break;

                case IV_VALUE_TYPE_DEVICE_BDE: // BDE-MasterIV
                    if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_BDE) 
                    {
                        // Die gewaehlte Geraetedatei ist NICHT fuer das BDE-MasterIV.
                        return -IV_ERROR_DEVICEFILE_WRONG;
                    }
                    break;

                case IV_VALUE_TYPE_DEVICE_MOBIL: // Mobil-MasterIV
                    if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_MOBIL) 
                    {
                        // Die gewaehlte Geraetedatei ist NICHT fuer das Mobil-MasterIV.
                        return -IV_ERROR_DEVICEFILE_WRONG;
                    }
                    break;

                case IV_VALUE_TYPE_DEVICE_MDE: // MDE-BoxIV
                    if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_MDE) 
                    {
                        // Die gewaehlte Geraetedatei ist NICHT fuer das MDE-BoxIV
                        return -IV_ERROR_DEVICEFILE_WRONG;
                    }
                    break;

                case IV_VALUE_TYPE_DEVICE_AE: // AE-MasterIV
                    if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_AE) 
                    {
                        // Die gewaehlte Geraetedatei ist NICHT fuer das AE-MasterIV.
                        return -IV_ERROR_DEVICEFILE_WRONG;
                    }
                    break;

                case IV_VALUE_TYPE_DEVICE_TIMEBOY: // TimeboyIV
                    if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_TIMEBOY) 
                    {
                        // Die gewaehlte Geraetedatei ist NICHT fuer das TimeboyIV.
                        return -IV_ERROR_DEVICEFILE_WRONG;
                    }
                    break;

                case IV_VALUE_TYPE_DEVICE_EXKLUSIV: // Exklusiv-LineIV
                    if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_EXKLUSIV) 
                    {
                        // Die gewaehlte Geraetedatei ist NICHT fuer das Exklusiv-LineIV.
                        return -IV_ERROR_DEVICEFILE_WRONG;
                    }
                    break;

                case IV_VALUE_TYPE_DEVICE_FLEX: // Flex-MasterIV
                    if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_FLEX) 
                    {
                        // Die gewaehlte Geraetedatei ist NICHT fuer das Flex-MasterIV.
                        return -IV_ERROR_DEVICEFILE_WRONG;
                    }
                    break;

                case IV_VALUE_TYPE_DEVICE_TIMEBOY_MOBIL_PZE: // Timeboy-Mobil-PZE
                    if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_TIMEBOY_MOBIL_PZE) 
                    {
                        // Die gewaehlte Geraetedatei ist NICHT fuer das Timeboy-Mobil-PZE.
                        return -IV_ERROR_DEVICEFILE_WRONG;
                    }
                    break;

                case IV_VALUE_TYPE_DEVICE_TERMINAL_2_8: // Terminal 2.8
                    if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_TERMINAL_2_8) 
                    {
                        // Die gewaehlte Geraetedatei ist NICHT fuer das Terminal 2.8.
                        return -IV_ERROR_DEVICEFILE_WRONG;
                    }
                    break;

                case IV_VALUE_TYPE_DEVICE_TERMINAL_4_3: // Terminal 4.3
                    if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_TERMINAL_4_3) 
                    {
                        // Die gewaehlte Geraetedatei ist NICHT fuer das Terminal 4.3.
                        return -IV_ERROR_DEVICEFILE_WRONG;
                    }
                    break;

                case IV_VALUE_TYPE_DEVICE_IPC_EXTENSION_VARIO_5_7: // IPC-Extension
                    if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_IPC_EXTENSION_VARIO_5_7) 
                    {
                        // Die gewaehlte Geraetedatei ist NICHT fuer das IPC-Extension.
                        return -IV_ERROR_DEVICEFILE_WRONG;
                    }
                    break;

                case IV_VALUE_TYPE_DEVICE_IPC_EXTENSION_VARIO_10: // IPC-Extension
                    if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_IPC_EXTENSION_VARIO_10) 
                    {
                        // Die gewaehlte Geraetedatei ist NICHT fuer das IPC-Extension.
                        return -IV_ERROR_DEVICEFILE_WRONG;
                    }
                    break;

                case IV_VALUE_TYPE_DEVICE_IPC_EXTENSION_Q7: // IPC-Extension
                    if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_IPC_EXTENSION_Q7) 
                    {
                        // Die gewaehlte Geraetedatei ist NICHT fuer das IPC-Extension.
                        return -IV_ERROR_DEVICEFILE_WRONG;
                    }
                    break;

                case IV_VALUE_TYPE_DEVICE_IO_BOX: // IO-Box
                    if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_IO_BOX) 
                    {
                        // Die gewaehlte Geraetedatei ist NICHT fuer das IO-Box.
                        return -IV_ERROR_DEVICEFILE_WRONG;
                    }
                    break;

                case IV_VALUE_TYPE_DEVICE_ZK_BOX: // ZK-Box
                    if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_ZK_BOX) 
                    {
                        // Die gewaehlte Geraetedatei ist NICHT fuer das ZK-Box.
                        return -IV_ERROR_DEVICEFILE_WRONG;
                    }
                    break;

                case IV_VALUE_TYPE_DEVICE_MOBIL_BOX: // Mobil-Box
                    if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_MOBIL_BOX) 
                    {
                        // Die gewaehlte Geraetedatei ist NICHT fuer das Mobil-Box.
                        return -IV_ERROR_DEVICEFILE_WRONG;
                    }
                    break;

                default:
                    if ( m_fileDevice.GetHeaderVersionFile() == IV_DFFILEDEVICE_VERSION_SOFTWARE_INFO 
                        && (m_fileDevice.GetHeaderSoftwareInfo().deviceExtended & (1 << deviceType)) != 0 )
                    {
                        // Die Daten sind für den vorliegenden Gerätetyp geeignet.
                    }
                    else
                    {
                        if ( m_fileDevice.GetHeaderVersionFile() == IV_DFFILEDEVICE_VERSION_SOFTWARE_INFO )
                            // Die gewaehlte Geraetedatei ist NICHT fuer diese Gerätevariante.
                                return -IV_ERROR_DEVICEFILE_WRONG;
                        else
                            // Die Setupanwendung ist nicht aktuell genug fuer diesen Geraetetyp.
                            return -IV_ERROR_DEVICE_UNKNOWN;
                    }
                }

                std::string version;
                if (boardType < IV_VALUE_TYPE_BOARD_3_0)
                {
                    // Handelt es sich um einen Boardtyp < 3 muss die Firmwareversion 04.00.xx oder 04.01.xx sein.
                    if (m_fileDevice.GetHeaderVersionFirmware(version)) 
                    {
                        if ( version.compare( 0, 6, "04.00." ) != 0 && version.compare( 0, 6, "04.01." ) != 0 )
                        {
                            return -IV_ERROR_DEVICEFILE_MATCH;
                        }
                    }
                }
                else
                {
                    // Handelt es sich um einen Boardtyp >= 3 darf die Firmwareversion nicht 04.00.xx oder 04.01.xx sein.
                    if (m_fileDevice.GetHeaderVersionFirmware(version)) 
                    {
                        if (version.compare( 0, 6, "04.00." ) == 0 || version.compare( 0, 6, "04.01." ) == 0 )
                        {
                            return -IV_ERROR_DEVICEFILE_MATCH;
                        }
                    }
                }

            }

        } 
        else 
        { // Es handelt sich um eine Zipdatei

            HardwareInfo hardwareInfo;
            // Auslesen der Hardwareinfo
            if ((iRes = read_hardware_info(device_address, hardwareInfo)) != IV_ERROR_SUCCESS) 
            {
                return iRes;
            }

            // Uebergangsweise auf die Vorgaengerversion wechseln.
            hwInfo = hardwareInfo.exportStructHardwareInfo();

            // Liegt keine Hardwareinfo vor?
            // Verwendung der ZK-Option ueber Vorhandensein der Readerliste ermitteln.
            if (hwInfo.version == 0) 
            {
                int handle;
                std::string reserved;
                if (open_table(device_address, "ACCESS.READER", 0, handle, reserved) == IV_ERROR_SUCCESS 
                    && close_table(device_address, handle) == IV_ERROR_SUCCESS) 
                {
                    // ZK-Option in Verwendung.
                    hwInfo.administrationFlags |= HIAF_HAS_ZK;
                }
            }

            // Pruefen ob die Datei in die vorliegende Hardware eingespielt werden kann.
            if (0 == m_fileDevice.SearchInfosZip(file_path, hwInfo)) 
            {
                int err = IV_ERROR_SUCCESS;
                // Hardwareinfo in Text aufloesen
                hardware_info_to_string(hwInfo, m_hardware_info_as_text);
                if (m_fileDevice.errorState() == FileIntelHex::Error_state::version 
                    || m_fileDevice.errorState() == FileIntelHex::Error_state::device_unknown)
                {
                    err = -IV_ERROR_VERSION_DEVICEFILE;
                } 
                else if (m_fileDevice.errorState() == FileIntelHex::Error_state::no_file_match)
                {
                    err = -IV_ERROR_DEVICEFILE_MATCH;
                } 
                else 
                {
                    err = -IV_ERROR_DEVICEFILE_LOAD;
                }
                return err;
            }
        }	
    } 
    else if (S_ISDIR(st.st_mode)) 
    {		
#if defined(_MSC_VER) || defined(__MINGW32__)
        // Auswahl aus Verzeichnis nur unter Linux gueltig!
        return -IV_ERROR_FUNCTION_PARAMETER;
#else
        HardwareInfo hardwareInfo;
        // Auslesen der Hardwareinfo
        if ((iRes = read_hardware_info(device_address, hardwareInfo)) != IV_ERROR_SUCCESS) 
        {
            return iRes;
        }

        // Uebergangsweise auf die Vorgaengerversion wechseln.
        hwInfo = hardwareInfo.exportStructHardwareInfo();

        // Liegt keine Hardwareinfo vor?
        // Verwendung der ZK-Option ueber Vorhandensein der Readerliste ermitteln.
        if (hwInfo.version == 0) 
        {
            int handle;
            std::string tmp;
            if (open_table(device_address, (const char*)"ACCESS.READER", 0, handle, tmp) == IV_ERROR_SUCCESS &&
                close_table(device_address, handle) == IV_ERROR_SUCCESS) 
            {
                // ZK-Option in Verwendung.
                hwInfo.administrationFlags |= HIAF_HAS_ZK;
            }
        }

        // Pruefen ob eine passende Datei in die vorliegende Hardware eingespielt werden kann.
        if (0 == m_fileDevice.SearchInfosDir(file_path, hwInfo)) 
        {
            int err = IV_ERROR_SUCCESS;
            // Hardwareinfo in Text aufloesen
            hardware_info_to_string(hwInfo, m_hardware_info_as_text);
            if (m_fileDevice.errorState() == FileIntelHex::Error_state::version
                || m_fileDevice.errorState() == FileIntelHex::Error_state::device_unknown)
            {
                err = -IV_ERROR_VERSION_DEVICEFILE;
            }
            else if (m_fileDevice.errorState() == FileIntelHex::Error_state::no_file_match)
            {
                err = -IV_ERROR_DEVICEFILE_MATCH;
            } 
            else 
            {
                err = -IV_ERROR_DEVICEFILE_LOAD;
            }
            return err;
        }
#endif
    } 
    else 
    {
        return -IV_ERROR_FILE_OPEN;
    }		

    //######################################################################
    // 1. Ermittlung ob auf dem Geraet eine Applicationfirmware laeuft.
    // 2. Wenn Punkt 1 == TRUE, dann wird geprueft ob Datensaetze verfuegbar sind.
    // 2. Wenn Punkt 2 == TRUE, wird die Aktion abgebrochen.
    std::string version;
    int modus = 0;
    // Geraetemodus (Application/Bootloader) ermitteln
    if ((iRes = Module_table::read_firmware_version(device_address, version)) != IV_ERROR_SUCCESS) 
    {
        if (IV_RESULT_MODULE(iRes) == 2)
        {
            // Bootloadermodus
            modus = 1;
        }
        else
        {
            return iRes;
        }
    }

    // 2013_04_17:
    // Es muss sichergestellt werden, dass eine Firmwareversion < 04.02.04.20 nur auf Timeboy Geraete mit Ladeschaltung aufgespielt werden kann.
    // Wenn dieses geschieht, dann geht der Timeboy aus und rebooted ständig.
    if ( m_fileDevice.GetHeaderVersionFirmware() >= "04.02.01.00" && m_fileDevice.GetHeaderVersionFirmware() < "04.02.04.20" )
    {
        // Im Bootloadermodus koennen wir das Flag der Ladeschaltung nicht auswerten, dadurch lassen wir das aufspielen zu.
        if ( modus == 0 && SoftwarebuildVersion( version ) >= "04.02.04.20" )
        {
            if ( (hwInfo.typeDevice == IV_VALUE_TYPE_DEVICE_TIMEBOY || hwInfo.typeDevice == IV_VALUE_TYPE_DEVICE_TIMEBOY_MOBIL_PZE)
                && !hwInfo.componentCharge ) // Bei Timeboy's ohne Ladeschaltung ist das aufspielen unzulaessig.
            {
                return -IV_ERROR_DEVICEFILE_MATCH;
            }
        }
    }

    iRes = 0;
    // Abfrage nur wenn Geraet im Applicationmodus
    // Abfrage ob Daten vorhanden
    if (!modus && (iRes = checkData(device_address)) == -1) 
    {
        return -IV_ERROR_UNREACHABLE;
    } 
    else if (iRes) 
    {
        return -IV_ERROR_DATA_AVAILABLE;
    }

    // Befehlsausfuehrung starten.
    iRes = writeDevice4( device_address );

    if (iRes == IV_ERROR_SUCCESS)
    {
        // Ggf. unbekannte Dockingstation
        get_protocol().reset_timeboy_station();
    }

    return iRes;
}

int Channel4::writeDevice4( unsigned char device_address )
{
    // Zeichensatz- und Grafikdaten
    FileDisplayFonts *pFileDisplayFonts = m_fileDevice.GetFileDisplayFonts();
    FileDisplayGrafics *pFileDisplayGrafics = m_fileDevice.GetFileDisplayGrafics();
    FileDeviceText *pFileDeviceText = m_fileDevice.GetFileDeviceText();
    FileIntelHex *pFileRawFirmware = m_fileDevice.GetFileRawFirmware();
    int res = IV_ERROR_SUCCESS;
    unsigned int address = 0;
    unsigned char *buffer;
    unsigned int size;
    unsigned int crc = 0;
    std::string version;

    // Bootversion lesen 
    if ((res = read_bootloader_version(device_address, version)) != IV_ERROR_SUCCESS)
    {
        return res;
    }

    // Ggf. gesetztes Streaming deaktivieren.
    set_stream_transfer( false );
    clear_stream_transfer();

    // Systemstopp setzen
    if ((res = enter_write_data_segment(device_address, IV_VALUE_SEGMENT_UPDATE, address, 0, 0)) != IV_ERROR_SUCCESS)
    {
        return res;
    }

    // Pruefen ob die Uebertragung per Stream moeglich ist.
    if ((res = write_stream(device_address, Stream_type::Update, nullptr, 0)) == IV_ERROR_SUCCESS)
    {
        set_stream_transfer(true);
    }

    // Sind Textdaten vorhanden?
    if (pFileDeviceText->sizeData())
    {
        // Texte als ersten Block uebertragen
        // TEXTE

        address = 0;
        buffer = pFileDeviceText->data();
        size = pFileDeviceText->sizeData();
        crc = pFileDeviceText->checksum32bit();

        // Bei Streaming werden die inneren Bloecke wegelassen.
        if (!is_stream_transfer()) {
            // Systemstopp setzen
            res = enter_write_data_segment(device_address, IV_VALUE_SEGMENT_TEXT, address, size, crc);
            if (res != IV_ERROR_SUCCESS && (res & 0xFF) != 1)
            {
                return res;
            }
        }

        // Funktionsfehler nicht Blocktyp unbekannt?
        if ((res & 0xFF) != 1)
        {
            // Textdaten uebertragen
            if ((res = write_to_flash(device_address, address, buffer, size)) != IV_ERROR_SUCCESS)
            {
                return res;
            }

            // Im Modus Streaming ?, dann gepufferte Daten senden.
            if (is_stream_transfer() && m_stream_buffer.size()) {
                if ((res = write_stream(device_address, Stream_type::Text, m_stream_buffer.data(), static_cast<unsigned int>( m_stream_buffer.size() ))) != IV_ERROR_SUCCESS) {
                    // Gepufferte Daten loeschen.
                    clear_stream_transfer();
                    return res;
                }
                // Gepufferte Daten loeschen.
                clear_stream_transfer();
            }

            // Bei Streaming werden die inneren Bloecke wegelassen.
            if (!is_stream_transfer()) {
                // Systemstopp aufheben
                if ((res = exit_write_data_segment(device_address, IV_VALUE_SEGMENT_TEXT, crc)) != IV_ERROR_SUCCESS)
                {
                    return res;
                }
            }
        }
        else
        {
            // Texte als Zeichensatz mit uebertragen.
            int err;
            if (!(err = pFileDisplayFonts->AppendSegmentText(buffer, size)))
            {
                if (FileIntelHex::Error_state::alloc_mem == pFileDisplayFonts->errorState())
                {
                    return -IV_ERROR_ALLOC_MEM;
                }

                return -IV_ERROR_FILE_PROCESSING;
            }

            // Ruecksetzen des Flags fuer "gueltige Default-Textdaten vorhanden"
            unsigned char v = 0;
            if ((res = write_to_eeprom(device_address, 0x0ECB, &v, 1)) != IV_ERROR_SUCCESS)
            {
                return res;
            }
        }
    }

    // ZEICHENSAETZE
    if ( pFileDisplayFonts->sizeData() )
    {
        address = 0;
        buffer = pFileDisplayFonts->data();
        size = pFileDisplayFonts->sizeData();
        crc = pFileDisplayFonts->checksum32bit();

        // Bei Streaming werden die inneren Bloecke wegelassen.
        if (!is_stream_transfer()) {
            // Systemstopp setzen
            if ((res = enter_write_data_segment(device_address, IV_VALUE_SEGMENT_FONTS, address, size, crc)) != IV_ERROR_SUCCESS)
            {
                return res;
            }
        }

        // Zeichensatzdaten uebertragen
        if ((res = write_to_flash(device_address, address, buffer, size)) != IV_ERROR_SUCCESS)
        {
            return res;
        }

        // Im Modus Streaming ?, dann gepufferte Daten senden.
        if (is_stream_transfer() && m_stream_buffer.size()) {
            if ((res = write_stream(device_address, Stream_type::Fonts, m_stream_buffer.data(), static_cast<unsigned int>( m_stream_buffer.size() ))) != IV_ERROR_SUCCESS) {
                // Gepufferte Daten loeschen.
                clear_stream_transfer();
                return res;
            }
            // Gepufferte Daten loeschen.
            clear_stream_transfer();
        }

        // Bei Streaming werden die inneren Bloecke wegelassen.
        if (!is_stream_transfer()) {
            // Systemstopp aufheben
            if ((res = exit_write_data_segment(device_address, IV_VALUE_SEGMENT_FONTS, crc)) != IV_ERROR_SUCCESS)
            {
                return res;
            }
        }
    }

    // GRAFIKEN
    if ( pFileDisplayGrafics->sizeData() )
    {
        address = 0;
        buffer = pFileDisplayGrafics->data();
        size = pFileDisplayGrafics->sizeData();
        crc = pFileDisplayGrafics->checksum32bit();

        // Bei Streaming werden die inneren Bloecke wegelassen.
        if (!is_stream_transfer()) {
            // Systemstopp setzen
            if ((res = enter_write_data_segment(device_address, IV_VALUE_SEGMENT_GRAFICS, address, size, crc)) != IV_ERROR_SUCCESS)
            {
                return res;
            }
        }

        // Grafikdaten uebertragen
        if ((res = write_to_flash(device_address, address, buffer, size)) != IV_ERROR_SUCCESS)
        {
            return res;
        }

        // Im Modus Streaming ?, dann gepufferte Daten senden.
        if (is_stream_transfer() && m_stream_buffer.size()) {
            if ((res = write_stream(device_address, Stream_type::Graphics, m_stream_buffer.data(), static_cast<unsigned int>( m_stream_buffer.size() ))) != IV_ERROR_SUCCESS) {
                // Gepufferte Daten loeschen.
                clear_stream_transfer();
                return res;
            }
            // Gepufferte Daten loeschen.
            clear_stream_transfer();
        }

        // Bei Streaming werden die inneren Bloecke wegelassen.
        if (!is_stream_transfer()) {
            // Systemstopp aufheben
            if ((res = exit_write_data_segment(device_address, IV_VALUE_SEGMENT_GRAFICS, crc)) != IV_ERROR_SUCCESS) {

                return res;
            }
        }
    }

    // FIRMWARE

    address = 0;
    buffer = pFileRawFirmware->data();
    size = pFileRawFirmware->sizeData();
    crc = pFileRawFirmware->checksum32bit();

    // Bei Streaming werden die inneren Bloecke wegelassen.
    if (!is_stream_transfer()) {
        // Systemstopp setzen
        if ((res = enter_write_data_segment(device_address, IV_VALUE_SEGMENT_FIRMWARE, address, size, crc)) != IV_ERROR_SUCCESS)
        {
            return res;
        }
    }

    // Firmwaredaten uebertragen
    if ((res = write_to_flash(device_address, address, buffer, size)) != IV_ERROR_SUCCESS)
    {
        return res;
    }

    // Im Modus Streaming ?, dann gepufferte Daten senden.
    if (is_stream_transfer() && m_stream_buffer.size()) {
        if ((res = write_stream(device_address, Stream_type::Firmware, m_stream_buffer.data(), static_cast<unsigned int>( m_stream_buffer.size() ))) != IV_ERROR_SUCCESS) {
            // Gepufferte Daten loeschen.
            clear_stream_transfer();
            return res;
        }
        // Gepufferte Daten loeschen.
        clear_stream_transfer();
    }

    // Bei Streaming werden die inneren Bloecke wegelassen.
    if (!is_stream_transfer()) {
        // Systemstopp aufheben
        if ((res = exit_write_data_segment(device_address, IV_VALUE_SEGMENT_FIRMWARE, crc)) != IV_ERROR_SUCCESS)
        {
            return res;
        }
    }

    crc = 0;
    // Systemstopp aufheben
    if ((res = exit_write_data_segment(device_address, IV_VALUE_SEGMENT_UPDATE, crc)) != IV_ERROR_SUCCESS)
    {
        return res;
    }

    // Neuaushandlung bei verschluesselter Kommunikation
    Aes_crypt_session().reset_session_key();

    // Firmwareversion des Geraetes.
    m_version_firmware.clear();

    return res;
}

int Channel4::writeModuleWithTimeboySlot( unsigned char device_address, int type, const std::string &file_path )
{
    int res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = set_timeboy_slot_state( device_address, 1)) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = writeModule( device_address, type, file_path );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

int Channel4::writeModule( unsigned char device_address, int type, const std::string &file_path )
{
    int res;
    unsigned int address = 0;
    unsigned char *buf = nullptr;
    unsigned int size = 0;
    unsigned int crc = 0;

    std::string version;
    HardwareInfo hardwareInfo;
    listItemBoard tagsBiokey;

    //	memset(version, 0, sizeof(version));
    // Bootversion lesen 
    //	if ((res = getVersionBootloader(device_address, version)) != IV_ERROR_SUCCESS) {

    //		return res;
    //	}

    // Dateigroesse ermitteln
    // Datei oeffnen
    struct stat st;
    FILE *fp = nullptr;

    if (stat(file_path.c_str(), &st) != 0 || st.st_size == 0 || (fp = fopen(file_path.c_str(), "rb")) == nullptr) {

        res = -IV_ERROR_FILE_OPEN;
        goto error_label;
    }

    // Dateigroesse muss groesser 4kByte sein.
    if (st.st_size < 4096)
    {
        res = -IV_ERROR_FILE_FORMAT;
        goto error_label;
    }

    // Speicher reservieren
    buf = new unsigned char[st.st_size + 1];
    if (buf == nullptr) {

        res = -IV_ERROR_FILE_OPEN;
        goto error_label;
    }

    // Daten laden
    if ((int)fread(buf, sizeof(unsigned char), st.st_size, fp) != st.st_size) {

        res = -IV_ERROR_FILE_READ;
        goto error_label;
    }

    // Zugriff ueber Struktur initialisieren
    m_upload_module_signatur = (upload_module_signatur*)&buf[st.st_size - sizeof(upload_module_signatur)];
    // Gesamt-CRC pruefen
    crc = get_crc32(buf, st.st_size - sizeof(unsigned int));
    if (crc != m_upload_module_signatur->crc32)
        //if (crc = crc32(buf, st.st_size))
    {
        res = -IV_ERROR_FILE_CRC;
        goto error_label;
    }

    // Pruefung auf unterstutzte Version
    if (m_upload_module_signatur->version != 1)
    {
        res = -IV_ERROR_FILE_FORMAT;
        goto error_label;
    }

    // HardwareInfo abrufen.
    if ((res = read_hardware_info(device_address, hardwareInfo)) != IV_ERROR_SUCCESS)
    {
        if (res > 0) {

            if ((res & 0xFF) == 28) {

                // Passwort fuer Kommunikation wird benoetigt.
                res = -IV_ERROR_NEED_PASSWORD;
            } else if ((res & 0xFF) == 29) {

                // Nach 3 maliger Passwortfalscheingabe ist Geraet fuer 60 Sekunden gesperrt.
                res = -IV_ERROR_PASSWORD_LOCK;
            } else {

                // Funktionsfehler
                res = -IV_ERROR_PROCESSING;
            }
        }

        // Fehlerhafte durchfuehrung zurueckmelden.
        goto error_label;
    }

    tagsBiokey = hardwareInfo.moduleBiokey();
    if (tagsBiokey.size() == 0)
    {
        return -IV_ERROR_NO_MODULE;
    }
    else if (tagsBiokey.size() > 1)
    {
        return -IV_ERROR_FUNCTION_UNAVAILABLE;
    }
    else
    {
        // 1 = Idencom (BioKey 3000 - Modul)
        // 2 = Idencom (BioKey 4000 - Modul)
        // 3 = Idencom (BioKey 4020 - Modul)
        unsigned char val = IV_VALUE_TYPE_FINGERPRINT_NONE;
        switch (tagsBiokey.front()->hc().id)
        {
        case EnumModuleBiokey::Biokey3000:
            val = IV_VALUE_TYPE_FINGERPRINT_BIOKEY_3000;
            break;

        case EnumModuleBiokey::Biokey4000:
            val = IV_VALUE_TYPE_FINGERPRINT_BIOKEY_4000;
            break;

        case EnumModuleBiokey::Biokey4020:
            val = IV_VALUE_TYPE_FINGERPRINT_BIOKEY_4020;
            break;

        default:
            break;
        }

        // Pruefung auf zu aktualisierendes Modul
        if (m_upload_module_signatur->type != val)
        {
            // In der Datei liegt ein Update fuer eine andere Modulversion vor
            res = -IV_ERROR_FILE_DATA_FORMAT;
            goto error_label;
        }
    }

    // Ggf. gesetztes Streaming deaktivieren.
    set_stream_transfer( false );
    clear_stream_transfer();

    // Systemstopp setzen
    if ((res = enter_write_data_segment(device_address, IV_VALUE_SEGMENT_UPDATE, address, size, crc)) != IV_ERROR_SUCCESS) {

        if (res > 0) {

            // Wenn der Modulfehler -Funktion unbekannt- ist,
            // dann ist die Firmwareversion nicht kompatible zum Befehl.
            if (((res >> 8) & 0xFF) == 2) {

                // Hintergrundladen noch nicht implementiert
                res = -IV_ERROR_VERSION_FIRMWARE;
            } else if ((res & 0xFF) == 28) {

                // Passwort f?r Kommunikation wird ben?tigt.
                res = -IV_ERROR_NEED_PASSWORD;
            } else if ((res & 0xFF) == 29) {

                // Nach 3 maliger Passwortfalscheingabe ist Ger?t f?r 60 Sekunden gesperrt.
                res = -IV_ERROR_PASSWORD_LOCK;
            } else {

                // Funktionsfehler
                res = -IV_ERROR_PROCESSING;
            }
        }

        // Fehlerhafte durchfuehrung zurueckmelden.
        goto error_label;
    }


    // FIRMWARE des Biokey laden

    address = 0;
    size   = st.st_size - sizeof(upload_module_signatur);
    crc = get_crc32(buf, size);

    // Systemstopp setzen
    if ((res = enter_write_data_segment(device_address, IV_VALUE_SEGMENT_FINGERPRINT, address, size, crc)) != IV_ERROR_SUCCESS) {

        if (res > 0) {

            // Wenn der Modulfehler -Funktion unbekannt- ist,
            // dann ist die Firmwareversion nicht kompatible zum Befehl.
            if (((res >> 8) & 0xFF) == 2) {

                // Fingerprintmodul noch nicht implementiert
                res = -IV_ERROR_VERSION_FIRMWARE;
            } else if ((res & 0xFF) == 28) {

                // Passwort f?r Kommunikation wird ben?tigt.
                res = -IV_ERROR_NEED_PASSWORD;
            } else if ((res & 0xFF) == 29) {

                // Nach 3 maliger Passwortfalscheingabe ist Ger?t f?r 60 Sekunden gesperrt.
                res = -IV_ERROR_PASSWORD_LOCK;
            } else {

                // Funktionsfehler
                res = -IV_ERROR_PROCESSING;
            }
        }

        // Fehlerhafte durchfuehrung zurueckmelden.
        goto error_label;
    }

    // Firmwaredaten uebertragen
    if ((res = write_to_flash(device_address, address, buf, size)) != IV_ERROR_SUCCESS) {

        if (res > 0) {

            // Funktionsfehler
            res = -IV_ERROR_PROCESSING;
        }

        // Fehlerhafte durchfuehrung zurueckmelden.
        goto error_label;
    }

    // Systemstopp aufheben
    if ((res = exit_write_data_segment(device_address, IV_VALUE_SEGMENT_FINGERPRINT, crc)) != IV_ERROR_SUCCESS) {

        if (res > 0) {

            // Wenn der Modulfehler -Funktion unbekannt- ist,
            // dann ist die Firmwareversion nicht kompatible zum Befehl.
            if (((res >> 8) & 0xFF) == 2) {

                // Hintergrundladen noch nicht implementiert
                res = -IV_ERROR_VERSION_FIRMWARE;
            } else if ((res & 0xFF) == 28) {

                // Passwort f?r Kommunikation wird ben?tigt.
                res = -IV_ERROR_NEED_PASSWORD;
            } else if ((res & 0xFF) == 29) {

                // Nach 3 maliger Passwortfalscheingabe ist Ger?t f?r 60 Sekunden gesperrt.
                res = -IV_ERROR_PASSWORD_LOCK;
            } else {

                // Funktionsfehler
                res = -IV_ERROR_PROCESSING;
            }
        }

        // Fehlerhafte durchfuehrung zurueckmelden.
        goto error_label;
    }


    crc = 0;
    // Systemstopp aufheben
    if ((res = exit_write_data_segment(device_address, IV_VALUE_SEGMENT_UPDATE, crc)) != IV_ERROR_SUCCESS) {

        if (res > 0) {

            // Funktionsfehler
            res = -IV_ERROR_PROCESSING;
        }

        // Fehlerhafte durchfuehrung zurueckmelden.
        goto error_label;
    }

    // Ausfuehrung erfolgreich abgeschlossen

error_label:

    if (buf) delete[] buf;
    if (fp) fclose(fp);

    return res;
}

int Channel4::writeMemToFile( unsigned char device_address, int type, int address, int size, const std::string &file_path )
{
    StructMemFileHeader header;
    int res = IV_ERROR_SUCCESS;
    char buf[0x400];
    int baseAddress = 0, offsAddress = 0;

    FILE *fp = fopen(file_path.c_str(), "wb");
    if (fp == nullptr)
    {
        return -IV_ERROR_FILE_OPEN;
    }

    // Dateiheader aufsetzen
    memset(&header, 0, sizeof(header));
    header.structSize = sizeof(header);
    header.type = type;
    header.baseAddress = baseAddress;
    header.memAddress = address;
    header.memSize = size;
    // Dateiheader schreiben
    if (fwrite(&header, sizeof(header), 1, fp) != sizeof(header))
    {
        return -IV_ERROR_FAIL;
    }

    switch (type)
    {
    case 0:
        break;

    case 1:
        if ((res = read_from_eeprom(device_address, address, (unsigned char*)buf, size)) != IV_ERROR_SUCCESS)
        {
            goto label_out;
        }
        break;
    case 2:
        if ((res = read_from_flash(device_address, address, (unsigned char*)buf, size)) != IV_ERROR_SUCCESS)
        {
            goto label_out;
        }
        break;

    case 3:
        break;
    }

label_out:
    if (fp) fclose(fp);
    return res;
}

int Channel4::writeFileToMem( unsigned char device_address, const std::string &file_path )
{
    StructMemFileHeader header;
    int res = IV_ERROR_SUCCESS, chunk = 0;
    int offsAddress = 0;
    char buf[0x400];

    FILE *fp = fopen(file_path.c_str(), "rb");
    if (fp == nullptr)
    {
        return -IV_ERROR_FILE_OPEN;
    }

    // Dateiheader lesen
    if (fread(&header, sizeof(header), 1, fp) != sizeof(header))
    {	
        res = -IV_ERROR_FAIL;
        goto label_out;
    }
    switch (header.type)
    {
    case 0://MB_ACCESS_SRAM:
        break;

    case 1://MB_ACCESS_EEPROM:
        if ((res = write_to_eeprom(device_address, header.memAddress, (unsigned char*)buf, chunk)) != IV_ERROR_SUCCESS)
        {
            goto label_out;
        }
        break;

    case 2://MB_ACCESS_FLASH:
        if ((res = write_to_flash(device_address, header.memAddress, (unsigned char*)buf, chunk)) != IV_ERROR_SUCCESS)
        {
            goto label_out;
        }
        break;

    case 3://MB_ACCESS_PROGRAM:
        break;

    case 5://MB_ACCESS_FLASH_PAGE:
        break;

    case 6://MB_ACCESS_MMC:
        break;
    }

label_out:
    if (fp) fclose(fp);
    return res;
}

// Abfrage ob Datensaetze im AEIII+ vorliegen.
// Erst ab Datensatzweisem lesen.
int Channel4::checkData( unsigned char device_address )
{
    int iRes = IV_ERROR_SUCCESS;
    unsigned char buf[0x255];
    std::size_t count = 0;
    unsigned char status = 0;

    // Befehlsausfuehrung starten.
    iRes = read_record( device_address, buf, count, status );

    if (iRes == IV_ERROR_SUCCESS)
    {
        // OK
        iRes = status;
    }
    else
    {
        iRes = -1;
    }

    return iRes;
}

// ########################################################
// Überlagerungen aus Channel4ModuleApplication
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++

int Channel4::read_record( uint8_t device_address, uint8_t *buffer, std::size_t &count, uint8_t &status )
{
    int res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = set_timeboy_slot_state( device_address, 1)) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = Module_table::read_record( device_address, buffer, count, status );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

int Channel4::quit_record( uint8_t device_address, uint8_t &status )
{
    int res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        // Datensatz quittieren
        res = Module_table::quit_record( device_address, status );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

int Channel4::clear_records( unsigned char device_address )
{
    int res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ((res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = Module_table::clear_records( device_address );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

int Channel4::restore_records( unsigned char device_address )
{
    int res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ((res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = Module_table::restore_records( device_address );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

int Channel4::write_memo_to_display( unsigned char device_address, unsigned char maschine, unsigned char time, unsigned char font, unsigned char audio, short flags, const std::string &text )
{
    int res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ((res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = Module_table::write_memo_to_display( device_address, maschine, time, font, audio, flags, text );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

int Channel4::write_message_to_display( unsigned char device_address, unsigned char time, unsigned char action, unsigned char audio, const std::string &text )
{
    int res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ((res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = Module_table::write_message_to_display( device_address, time, action, audio, text );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

int Channel4::write_variable( unsigned char device_address, const std::string &name, const std::string &val )
{
    int res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ((res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = Module_table::write_variable( device_address, name, val );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

int Channel4::read_variable( unsigned char device_address, const std::string &name, std::string &val )
{
    int res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ((res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = Module_table::read_variable( device_address, name, val );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

int Channel4::getVersionFirmwareFromFile( const std::string &file_path, std::string &version )
{
    // Lokale Membervariable reseten
    m_fileDevice.resetData();

    if ( !m_fileDevice.GetVersionFirmware(file_path, version) )
    {
        switch ( m_fileDevice.errorState() )
        {
        case FileDeviceZip::Error_state::file_open:
            return -IV_ERROR_FILE_OPEN;

        case FileDeviceZip::Error_state::head_crc:
            return -IV_ERROR_CRC_PACKET;

        case FileDeviceZip::Error_state::version:
        case FileDeviceZip::Error_state::device_unknown:
        case FileDeviceZip::Error_state::firmware_version_length:
        case FileDeviceZip::Error_state::function_unavailable:
            return -IV_ERROR_FUNCTION_UNAVAILABLE;
        }

        return -IV_ERROR_FUNCTION_UNAVAILABLE;
    }

    return IV_ERROR_SUCCESS;
}

int Channel4::write_font_type( unsigned char device_address, unsigned char type )
{
    int res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ((res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = Module_table::write_font_type( device_address, type );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

// ########################################################
// Überlagerungen aus Channel4ModuleSystem
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++

int Channel4::read_system_record( uint8_t device_address, uint8_t *buffer, std::size_t &count, uint8_t &status )
{
    int res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = Module_table::read_system_record( device_address, buffer, count, status );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

int Channel4::quit_system_record( uint8_t device_address, uint8_t &status )
{
    int res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ((res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = Module_table::quit_system_record( device_address, status );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

int Channel4::restore_system_records( unsigned char device_address )
{
    int res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ((res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = Module_table::restore_system_records( device_address );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

int Channel4::read_date_time( unsigned char device_address, char *time )
{
    int res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ((res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = Module_table::read_date_time( device_address, time );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

int Channel4::write_date_time( unsigned char device_address, char *time )
{
    int res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ((res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = Module_table::write_date_time( device_address, time );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

int Channel4::read_firmware_version( unsigned char device_address, std::string &version )
{
    int res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ((res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = Module_table::read_firmware_version( device_address, version );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

int Channel4::system_reset( unsigned char device_address, unsigned short modus )
{
    int res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ((res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = Module_table::system_reset( device_address, modus );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

int Channel4::read_serial_number( unsigned char device_address, unsigned int &serial )
{
    int res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ((res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = Module_table::read_serial_number( device_address, serial );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

int Channel4::write_relay_state( unsigned char device_address, const std::string &name, unsigned char bus_number, unsigned char relay_number, unsigned char status, unsigned short time )
{
    int res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ((res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = Module_table::write_relay_state( device_address, name, bus_number, relay_number, status, time );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

int Channel4::read_relay_state( unsigned char device_address, const std::string &name, unsigned char bus_number, unsigned char relay_number, unsigned char &status, unsigned short &time )
{
    int res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ((res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = Module_table::read_relay_state( device_address, name, bus_number, relay_number, status, time );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

int Channel4::write_system_variable( unsigned char device_address, const std::string &name, const std::string &val )
{
    int res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ((res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = Module_table::write_system_variable( device_address, name, val );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

int Channel4::read_system_variable( unsigned char device_address, const std::string &name, std::string &val )
{
    int res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ((res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = Module_table::read_system_variable( device_address, name, val );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

int Channel4::write_virtual_key( unsigned char device_address, unsigned char key_code, unsigned char flags )
{
    int res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ((res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = Module_table::write_virtual_key( device_address, key_code, flags );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

int Channel4::write_led_state( unsigned char device_address, unsigned slot, unsigned char id, unsigned char state )
{
    return Module_system::write_led_state( device_address, slot, id, state );
}

int Channel4::read_led_state( unsigned char device_address, unsigned slot, unsigned char id, unsigned char &state )
{
    return Module_system::read_led_state( device_address, slot, id, state );
}

// ##################################################################
/**
* @brief Ermittelt eine Sammlung von Informationen zum Gerät.
*
*/
int Channel4::read_device_info( unsigned char device_address, DeviceInfo &info )
{
    int res = IV_ERROR_SUCCESS;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        unsigned int flags = 0;
        unsigned int serial;

        std::string version;
        std::string info_text;
        std::array<char, IV_PACKET_LENGTH_DATA> buffer;

        // Initalisieren der DeviceInfo.
        info.clear();

        // Ggf. bei verschlüsselter Kommunikation, Aushandlung durchführen.
        if ( (res = ping( device_address, nullptr, nullptr )) != IV_ERROR_SUCCESS )
            return res;

        // Abfrage der Bootloaderversion
        if ( (res = read_bootloader_version( device_address, version )) != IV_ERROR_SUCCESS )
            return res;

        // Der Infoklasse übergeben.
        info.setVersionBootloader( version );

        // HIP-Datei lesen (importiert ggf. "alte" Hardwareinfo in die Datei).
        HardwareInfo hardwareInfo;
        if ( (res = read_hardware_info( device_address, hardwareInfo )) != IV_ERROR_SUCCESS )
        {
            switch ( res ) 
            {
            case 28: // Funktionsfehler: Setup-Passwort wird gefordert.
                break;

            case 29: // Funktionsfehler: Setup-Passwort wird gefordert, Gerät wegen zu vielen Fehlanfragen gesperrt.
                break;

            default:
                return res;
            }
        }

        // der Infoklasse übergeben.
        info.setHardwareInfo( hardwareInfo );
        hardwareInfo.reset();

        version.clear();
        // Abfrage der Firmwareversion
        if ( (res = read_firmware_version( device_address, version )) != IV_ERROR_SUCCESS )
        {
            // War die Funktion unbekannt?
            if ( IV_RESULT_MODULE( res ) == 2 )
            {
				res = IV_ERROR_SUCCESS;
                // Firmwareversion nicht verfuegbar.
                // Geraet befindet sich im Bootloadermodus.
                info.setBootloaderMode( true );
            }
            else
            {
                return res;
            }
        }
        else
        {
            // Der Infoklasse übergeben.
            info.setVersionFirmware( version );
        }

        SoftwareInfo softwareInfo;
        if ( (res = read_software_info( device_address, softwareInfo )) != IV_ERROR_SUCCESS )
        {
            // War die Funktion unbekannt?
            if ( IV_RESULT_MODULE( res ) != 2 )
			{
                return res;
        }

			res = IV_ERROR_SUCCESS;
        }
        else
        {
            // Der Infoklasse übergeben.
            info.setSoftwareInfo( softwareInfo );
            softwareInfo.clear();
        }

        FirmwareOptions firmwareOptions;
        if ( (res = read_firmware_options( device_address, firmwareOptions )) != IV_ERROR_SUCCESS )
        {
            // War die Funktion unbekannt?
            if ( IV_RESULT_MODULE( res ) != 2 )
			{
                return res;
        }

			res = IV_ERROR_SUCCESS;
        }
        else
        {
            // Der Infoklasse übergeben.
            info.setFirmwareOptions( firmwareOptions );
            firmwareOptions.clear();
        }

        // Abfrage der Seriennummer.
        if ( (res = read_serial_number( device_address, serial )) != IV_ERROR_SUCCESS )
        {
            // War die Funktion unbekannt?
            if ( IV_RESULT_MODULE( res ) != 2 )
			{
                return res;
        }

			res = IV_ERROR_SUCCESS;
        }
        else
        {
            // Der Infoklasse übergeben.
            info.setSerialNumber( serial );
        }
        buffer.fill( '\0' );
        // Abfrage des Setuppasswortkeys
        if ( (res = read_setup_password_key( device_address, buffer.data() )) != IV_ERROR_SUCCESS )
        {
            // War die Funktion unbekannt?
            if ( IV_RESULT_MODULE( res ) == 2 )
            {
				res = IV_ERROR_SUCCESS;
				// Passwordkey nicht verfuegbar.
                buffer.fill( '\0' );
            }
            else
            {
                return res;
            }
        }
        else
        {
            // Der Infoklasse übergeben.
            info.setPasswordKey( buffer.data() );
        }


        // Zusätzliche Informationen zu verbauten Modulen verfügbar?
        std::vector<std::string> names;

        if ( info.hardwareInfo().moduleBiokey().size() )
            names.push_back( "FINGERPRINT.MODULE" );

        // Die Info liegt lediglich bei folgenden USB-Hosts vor.
        if ( info.hardwareInfo().moduleUSBHostCommunication( boost::assign::list_of( EnumModuleUSBHostCommunication::Vinculum ) ).size() )
            names.push_back( "USBHOST.MODULE" );

        if ( info.hardwareInfo().moduleGPS().size() )
            names.push_back( "GPS.MODULE" );

        if ( info.hardwareInfo().moduleLANCommunication().size() || info.hardwareInfo().moduleWLANCommunication().size() )
            names.push_back( "COM.TCPIP" );

        for ( const std::string &name : names )
        {     
            info_text.clear();
            if ( (res = read_hardware_module_info( device_address, name.c_str(), "", info_text )) != IV_ERROR_SUCCESS )
            {
                // War die Funktion im Modul unbekannt?
                if ( res > 0 && IV_RESULT_MODULE( res ) == 2 )
                {
                    res = IV_ERROR_SUCCESS;
                    // Modulinfo nicht verfuegbar.
                    info_text = "NA";
                }
                else if ( res > 0 && (IV_RESULT_FUNCTION( res ) == 1 || IV_RESULT_FUNCTION( res ) == 2) )
                {
                    res = IV_ERROR_SUCCESS;
                    // Systemvariable ist nicht verfügbar.
                    info_text = "NA";
                }
                else if ( res > 0 && (IV_RESULT_FUNCTION( res ) == 6 || IV_RESULT_FUNCTION( res ) == 7 || IV_RESULT_FUNCTION( res ) == 8) )
                {
                    std::string mac;
                    if ( (res = read_system_variable( device_address, "com.mac", mac )) != IV_ERROR_SUCCESS || mac.size() != 12 )
                    {                    
                        res = IV_ERROR_SUCCESS;
                        // Abfrage steht nicht zur Verfügung.
                        buffer.fill( '\0' );
                        std::strncpy( buffer.data(), "NA", buffer.size() - 1 );
                    }
                    else
                    {
                        // Bindestriche einfügen
                        int i = static_cast<unsigned int>( mac.size() ) - 2;
                        for ( ; i >= 2; i -= 2 )
                            mac.insert( i, "-" );

                        std::strncpy( buffer.data(), mac.c_str(), buffer.size() - 1 );
                    }
                }
                else if ( res == 28 || res == 29 )
                {
                    // 28 == Funktionsfehler: Setup-Passwort wird gefordert.
                    // 29 == Funktionsfehler: Setup-Passwort wird gefordert, Gerät wegen zu vielen Fehlanfragen gesperrt.

                    res = IV_ERROR_SUCCESS;
                    // Abfrage steht nicht zur Verfügung.
                    buffer.fill( '\0' );
                    std::strncpy( buffer.data(), "NA", buffer.size() - 1 );

                    break;
                }
                else
                {
                    return res;
                }
            }
            else
            {
                std::strncpy( buffer.data(), info_text.c_str(), buffer.size() - 1 );
            }

            if ( name == "FINGERPRINT.MODULE" )
                info.setVersionFingerprint( buffer.data() );

            else if ( name == "USBHOST.MODULE" )
                info.setVersionUSBHost( buffer.data() );

            else if ( name == "GPS.MODULE" )
                info.setVersionGPS( buffer.data() );

            else if ( name == "COM.TCPIP" )
                info.setVersionTCPIP( buffer.data() );
        }

        // Zusätzlich folgende Systemvariablen abrufen.
        names.clear();

        // Gesetzte Mac-Adresse
        if ( info.hardwareInfo().moduleLANCommunication().size() )
            names.push_back( "COM.MAC" );

        // Zutrittskontrolle nur bei verbautem RS485 aktiv.
        if ( info.hardwareInfo().moduleRS485Communication().size() )
            names.push_back( "ACCESS.MASTERID" );

        if ( info.hardwareInfo().deviceType() == HardwareInfo::TimeboyIV || info.hardwareInfo().deviceType() == HardwareInfo::TimeboyMobilPZE )
            names.push_back( "COM.GROUPID" );

        if (info.hardwareInfo().deviceType() == HardwareInfo::ZKKnoten )
        {
            names.push_back("board.application");
            names.push_back("board.place[0].output[1].mode");
            names.push_back("board.place[0].serial[1].use");
            names.push_back("board.place[0].serial[2].use");
        }

        if ( info.hardwareInfo().deviceType() == HardwareInfo::IPCExtensionC )
        {
            // Bei Embedded Q7
            names.push_back( "system.powerswitch.timestartos" );
            names.push_back( "system.powerswitch.offtimelong" );
            names.push_back( "system.powerswitch.offtimeshort" );
            names.push_back( "system.powerswitch.ontime" );
        }

        for ( const std::string &name : names )
        {
            std::string val;
            if ( (res = read_system_variable( device_address, name.c_str(), val )) != IV_ERROR_SUCCESS )
            {
                // War die Funktion im Modul unbekannt?
                if ( res > 0 && IV_RESULT_MODULE( res ) == 2 )
                {
                    res = IV_ERROR_SUCCESS;
                    // Modulinfo nicht verfuegbar.
                    val = "NA";
                }
                else if ( res > 0 && (IV_RESULT_FUNCTION( res ) == 1 || IV_RESULT_FUNCTION( res ) == 2) )
                {
                    res = IV_ERROR_SUCCESS;
                    // Major- oder Minorname nicht verfügbar.
                    val = "NA";
                }
                else if ( res == -28 || res == -29 )
                {
                    // 28 == Funktionsfehler: Setup-Passwort wird gefordert.
                    // 29 == Funktionsfehler: Setup-Passwort wird gefordert, Gerät wegen zu vielen Fehlanfragen gesperrt.

                    res = IV_ERROR_SUCCESS;
                    // Abfrage steht nicht zur Verfügung.
                    val = "NA";
                    break;
                }
                else
                {
                    return res;
                }
            }

            if ( name == "COM.MAC" )
            {
                if ( val.size() == 12 )
                {
                    // Bindestriche einfügen
                    for ( auto i = val.size() - 2; i >= 2; i -= 2 )
                        val.insert( i, "-" );
                }

                info.setMacAddress( val );
            }
            else if ( name == "ACCESS.MASTERID" )
                info.setAccessControlMasterId( val );

            else if ( name == "COM.GROUPID" )
                info.setTimeboyGroupId( val );

            else if ( name == "board.application" )
                info.setBoardApplication( val );

            else if ( name == "board.place[0].output[1].mode" )
                info.setBoardPlace_0_Output_1_Mode( val );

            else if ( name == "board.place[0].serial[1].use" )
                info.setBoardPlace_0_Serial_1_Use( val );

            else if ( name == "board.place[0].serial[2].use" )
                info.setBoardPlace_0_Serial_2_Use( val );

            else if ( name == "system.powerswitch.timestartos" )
                info.setTimeStartOS( val );

            else if ( name == "system.powerswitch.offtimelong" )
                info.setOffTimeLong( val );

            else if ( name == "system.powerswitch.offtimeshort" )
                info.setOffTimeShort( val );

            else if ( name == "system.powerswitch.ontime" )
                info.setOnTime( val );
        }

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

int Channel4::read_hardware_info( unsigned char device_address, HardwareInfo &info )
{
    int res = IV_ERROR_SUCCESS;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        unsigned char val = 0;

        // Initalisieren.
        info.reset();

        // Lesen der Hardwareconfig-Nr
        if ( (res = read_hardware_info_version( device_address, val )) != IV_ERROR_SUCCESS )
            return res;

        // Gelesene Version zur Information halten.
        info.setVersion( val );

        // Ab Version 4 kommt eine geänderte HWI zum Einsatz.
        if ( val < 4 )
        {
            unsigned char typeDevice;
            unsigned char typeBoard;
            // Lesen des Gerätetypen
            if ( (res = read_device_type( device_address, typeDevice )) != IV_ERROR_SUCCESS )
                return res;

            // Lesen der Boardversion
            if ( (res = read_board_type( device_address, typeBoard )) != IV_ERROR_SUCCESS )
                return res;

            std::array<unsigned char, sizeof(StructHardwareInfo)> dshwi;
            dshwi.fill( '\0' );
            StructHardwareInfo &shw = reinterpret_cast<StructHardwareInfo &>( dshwi );
            shw.version = val;
            shw.typeDevice = typeDevice;
            shw.typeBoard = typeBoard;

            // Prüfen ob es sich um ein MasterIV handelt.
            if ( info.isMasterIV( shw ) )
            {
                // Lesen der Vorgaenger-Hardwareinfo
                if ( (res = Module_table::read_hardware_info( device_address, shw )) != IV_ERROR_SUCCESS )
                    return res;

                // Vorgängerversion importieren
                if ( !info.importStructHardwareInfo( shw ) )
                    return -IV_ERROR_FUNCTION_UNAVAILABLE;

                return res;
            }
            else
            {
                // Es handelt sich um eine Zwischenvariante bei der Gerätetyp und Boardrevision noch von der
                // Vorgängerversion genutzt wurden.
            }
        }

        //
        unsigned int address, size, crc;
        address = size = crc = 0;

        // Segmentbegin: Bedienung stoppen
        if ((res = enter_read_data_segment(device_address, IV_VALUE_SEGMENT_GET_UPDATE_BG, address, size, crc)) != IV_ERROR_SUCCESS)
        {
            if ((res & 0xFF) == 1)
            {
                // Funktionsfehler Typ: IV_VALUE_SEGMENT_GET_UPDATE_BG unbekannt.
                // Wenn dass der Fall ist, wird das Auslesen noch nicht unterstuetzt.
                res = -IV_ERROR_FUNCTION_UNAVAILABLE;
            }
            else
            {
                return res;
            }
        }
        trace("%s(%i):INFO(1): %s [ beginGetSegment(%i): %i, %i, 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_UPDATE_BG, address, size, crc);

        // Daten aus dem Gerät in die Datei der HardwareInfo lesen.
        info.reset();
        FileHardwareInfo &fileHardwareInfo = info.fileHardwareInfo();

        //////////////////////////////
        // Begin

        // Segmentbegin: Config uebertragen
        if ((res = enter_read_data_segment(device_address, IV_VALUE_SEGMENT_HARDWARECONFIGS, address, size, crc)) != IV_ERROR_SUCCESS)
        {
            return res;
        }
        trace("%s(%i):INFO(1): %s [ beginGetSegment(%i): %i, %i, 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_SETUP, address, size, crc);

        if ( size == 0 )
        {
            // Es liegen keine Daten vor.
            res = -IV_ERROR_NO_DATA_AVAILABLE;
            goto label_no_data;
        }
        //////////////////////////////
        // Daten
        info.fileHardwareInfo().setSizeData( size );

        // HardwareInfo lesen
        if ((res = read_from_flash(device_address, address, fileHardwareInfo.data(), fileHardwareInfo.sizeData())) != IV_ERROR_SUCCESS)
        {
            return res;
        }

        //////////////////////////////
        // Ende
label_no_data:
        // Segmentend: Setup uebertragen
        if ((res = exit_read_data_segment(device_address, IV_VALUE_SEGMENT_HARDWARECONFIGS)) != IV_ERROR_SUCCESS)
        {
            return res;
        }
        trace("%s(%i):INFO(1): %s [ endGetSegment(%i) ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_SETUP);

        // Segmentend: Bedienung stoppen
        if ((res = exit_read_data_segment(device_address, IV_VALUE_SEGMENT_GET_UPDATE_BG)) != IV_ERROR_SUCCESS)
        {
            return res;
        }
        trace("%s(%i):INFO(1): %s [ endGetSegment(%i) ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_UPDATE_BG);

        if ( size )
        {
            // Datenanzahl in Datei uebernehmen.
            unsigned int crcTest = fileHardwareInfo.checksum32bit();
            if (crc != crcTest)
            {
                trace("%s(%i):ERROR(1): %s [ CRC von Segment falsch: 0x%08X, 0x%08X ]\n", __FILE__, __LINE__, __func__, crc, crcTest);
                return -IV_ERROR_CRC_SEGMENT;
            }

            // Baum aus den Daten erzeugen.
            if ( !fileHardwareInfo.readSerialize() )
            {
                trace("%s(%i):ERROR(1): %s [ Lesen der Datei gescheitert. ]\n", __FILE__, __LINE__, __func__);
                return -IV_ERROR_FILE_WRITE;
            }
        }

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

// ##################################################################
/**
* @brief Ermittelt eine Sammlung von Informationen zum Gerät.
*
*/
int Channel4::read_software_info( unsigned char device_address, SoftwareInfo &info )
{
    int res = IV_ERROR_SUCCESS;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        StructSoftwareInfo swInfo;
        int size = sizeof(StructSoftwareInfo);
        memset( &swInfo, 0, sizeof(StructSoftwareInfo) );

        // Initalisieren
        info.clear();

        // Abfrage der Softwareinfo
        if ( (res = Module_table::read_software_info( device_address, (unsigned char*)&swInfo, size )) != IV_ERROR_SUCCESS )
        {
            // War die Funktion im Modul unbekannt?
            if ( (res > 0) && (((res >> 8) & 0xFF) == 2) )
            {
                // Softwareinformation nicht verfuegbar.
            }
            else
            {
                return res;
            }
        }

        // Der Infoklasse übergeben.
        info.setStructSoftwareInfo( swInfo );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

// ##################################################################
/**
* @brief Ermittelt eine Sammlung von Informationen zum Gerät.
*
*/
int Channel4::read_firmware_options( unsigned char device_address, FirmwareOptions &options )
{
    int res = IV_ERROR_SUCCESS;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {
        bool haveFunction;
        unsigned int flags = 0;
        unsigned char type;
        unsigned char available;

        // Initalisieren
        options.clear();

        // Zusaetzlich das Flag fuer "Setup und Listen beibehalten" auswerten.
        if ( (res = read_system_flags( device_address, flags )) != IV_ERROR_SUCCESS )
        {
            // War die Funktion im Modul unbekannt?
            if ( (res > 0) && (((res >> 8) & 0xFF) == 2) )
            {
                // Versuchen direkt an der Adresse zu lesen und zu setzen
                if ( read_from_eeprom( device_address, 0x0ECC, (unsigned char*)&flags, sizeof(unsigned int) ) == IV_ERROR_SUCCESS )
                {
                    // Der Infoklasse übergeben.
                    options.setSystemFlags( flags );
                }
            }
        }
        else
        {
            // Der Infoklasse übergeben.
            options.setSystemFlags( flags );
        }

#undef CHECK_COULD_HAVE_FUNCTION
#define CHECK_COULD_HAVE_FUNCTION( x ) \
    haveFunction = true; \
    if ( (res = x) != IV_ERROR_SUCCESS ) \
        { \
        if ( (res > 0)  && (((res >> 8) & 0xFF) == 2) ) \
        { \
        res = IV_ERROR_SUCCESS; \
        /* Funktion kann ggf. aufgrund eines älteren Firmwarestandes undefiniert sein. */ \
        haveFunction = false; \
        } \
    else \
        { \
        /* Davon ausgehen, dass das Geraet unerreichbar ist.*/ \
        return res; \
        } \
        }

        // Typ ermitteln
        CHECK_COULD_HAVE_FUNCTION( read_external_transponder_reader_type( device_address, type ) );
        if ( haveFunction )
            options.setTypeExternalReader( type );

        // Option ermitteln
        CHECK_COULD_HAVE_FUNCTION( read_io_module_option( device_address, available ) );
        if ( haveFunction )
            options.setOptionIOModule( available );

        // Option ermitteln
        CHECK_COULD_HAVE_FUNCTION( read_timeboy_adaption_option( device_address, available ) );
        if ( haveFunction )
            options.setOptionTBAdaption( available );

        // Option ermitteln
        CHECK_COULD_HAVE_FUNCTION( read_com_server_option(device_address, available) );
        if ( haveFunction )
            options.setOptionComServer( available );

        // Option ermitteln
        CHECK_COULD_HAVE_FUNCTION( read_transponder_write_option( device_address, available ) );
        if ( haveFunction )
            options.setOptionTPWrite( available );

        // Option ermitteln
        CHECK_COULD_HAVE_FUNCTION( read_http_option( device_address, available ) );
        if ( haveFunction )
            options.setOptionHTTP( available );

        // Option ermitteln
        CHECK_COULD_HAVE_FUNCTION( read_localization_option(device_address, available) );
        if ( haveFunction )
            options.setOptionLanguage( available );

        // Option ermitteln
        CHECK_COULD_HAVE_FUNCTION( read_fingerprint_scanner_update_option( device_address, available ) );
        if ( haveFunction )
            options.setOptionBiokeyUpdate( available );

        unsigned short so_mask, so_options;
        // SetupOption: Signalverarbeitung
        CHECK_COULD_HAVE_FUNCTION( read_setup_options( device_address, so_mask, so_options ) );
        if ( haveFunction )
            options.setSetupOptions( so_mask, so_options );

        unsigned char po_mask, po_options;
        // ErweiterteOption: Data on Card
        CHECK_COULD_HAVE_FUNCTION( read_buy_options( device_address, po_mask, po_options ) );
        if ( haveFunction )
            options.setPayableOptions( po_mask, po_options );

        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

#undef CHECK_COULD_HAVE_FUNCTION

    return res;
}

// ##################################################################
/**
* @brief Ermittelt eine Sammlung von Informationen zum Gerät.
*
*/
int Channel4::write_firmware_options( unsigned char device_address, const FirmwareOptions &options )
{
    int res = IV_ERROR_SUCCESS;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = set_timeboy_slot_state( device_address, 1 )) == IV_ERROR_SUCCESS )
    {



        // Versuchen die rote LED abzuschalten.
        set_timeboy_slot_state( device_address, 0, res );
    }

    return res;
}

int Channel4::enter_transparency_mode( unsigned char device_address, unsigned short id, unsigned short flags, unsigned int read_buffer_size, unsigned int write_buffer_size, int timeout )
{
    return Module_system::enter_transparency_mode( device_address, id, flags, read_buffer_size, write_buffer_size, timeout );
}

int Channel4::exit_transparency_mode( unsigned char device_address )
{
    return Module_system::exit_transparency_mode( device_address );
}

bool Channel4::is_transparency_mode()
{
    return Module_system::is_transparency_mode();
}
